/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashNoise: a noise generator
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdlib.h>

#define __MASH_PLUGIN__
#include "mash.h"

/* Optimisation dependencies: none */
static MashBase * mash_noise_optimise (MASH * mash, MashBase * noise);

static MashBase *
mash_noise_init (MASH * mash, MashBase * base, CDSet * parameters)
{
  mash_noise_optimise (mash, base);
  return base;
}

static MashBase *
mash_noise_clone (MASH * mash, MashBase * base)
{
  MashBase * new_noise = mash_base_new (mash);
  mash_noise_optimise (mash, new_noise);
  return new_noise;
}

static int
mash_noise_destroy (MASH * mash, MashBase * base)
{
  free (base);
  return 0;
}

/* An MashChunkFunc for creating noise */
static MashCount
mash_noise_write_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
		      MashCount count, int channelname, void * data)
{
  MashPCM * d;
  MashPCM r, rmmash = (MashPCM)RAND_MAX;
  MashCount i;

  mash_dprintf ("[mash_noise_write_chunk] (%p, +%ld) @ %ld\n", data, count,
	      offset);

  d = &chunk->data[offset];
  for (i = 0; i < count; i++) {
    r = (MashPCM) random ();
    r *= 2.0;
    r /= rmmash;
    *d++ = r - 1.0;
  }

  return count;
}

static MashCount
mash_noise_process (MASH * mash, MashBase * base, MashCount count,
		  MashStream * input, MashStream * output)
{
  return mash_stream_chunkfuncify (mash, output, count,
				 mash_noise_write_chunk, base);
}

static MashCount
mash_noise_length (MASH * mash, MashBase * base)
{
  return MASH_COUNT_INFINITE;
}

static struct _MashMethods _mash_noise_methods = {
  mash_noise_clone,
  mash_noise_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_noise_process,
  mash_noise_length,
  NULL, /* seek */
  NULL, /* flush */
};

static MashBase *
mash_noise_optimise (MASH * mash, MashBase * noise)
{
  mash_base_set_methods (mash, noise, &_mash_noise_methods);
  return noise;
}

static struct _MashMetaText noise_metatext = {
  "mashstd::noise",
  "Generators::Noise",
  "White noise generator",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/mash/plugins/noise.html",
  MASH_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _MashPlugin noise_plugin = {
  &noise_metatext,
  MASH_FLAGS_NONE,
  CD_EMPTY_SET, /* new scheme */
  mash_noise_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

CDList *
mash_load (MASH * mash)
{
  CDList * plugins = cd_list_new (mash);

  plugins = cd_list_prepend (mash, plugins, CD_POINTER(&noise_plugin));

  return plugins;
}
