/*
 * libmash -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * mash.h -- Base types and public interfaces to libmash
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifndef __MASH_H__
#define __MASH_H__

#include <limits.h>

#include "ctxdata.h"

#define MashNone NULL

typedef int MashError;

typedef float MashPCM;

typedef long MashCount;
#define MASH_COUNT_MAX (LONG_MAX - 1L)
#define MASH_COUNT_MIN LONG_MIN
#define MASH_COUNT_INFINITE MASH_COUNT_MAX

typedef void MashOpaque;

#if defined (__MASH__)
typedef struct _MashBase MashBase;
#else
typedef MashOpaque MashBase;
#endif

typedef CDScalar MashParameter;

typedef int MashFlags;
typedef double MashSamplerate;
typedef double MashTempo;
typedef union _MashTime MashTime;


/* Errors */
#define MASH_ERROR_OK              0
#define MASH_ERROR_INVALID         1
#define MASH_ERROR_NOENTITY        2
#define MASH_ERROR_EXISTS          3
#define MASH_ERROR_SILENCE         4
#define MASH_ERROR_NOOP            5
#define MASH_ERROR_SYSTEM          6

typedef enum {
  MASH_CHANNEL_LEFT,
  MASH_CHANNEL_RIGHT,
  MASH_CHANNEL_CENTRE,
  MASH_CHANNEL_REAR,
  MASH_CHANNEL_REAR_LEFT,
  MASH_CHANNEL_REAR_RIGHT,
  MASH_CHANNEL_REAR_CENTRE,
  MASH_CHANNEL_LFE /* Low Frequency Effects */,
} MashChannelName;

typedef enum {
  MASH_TIME_INVALID,
  MASH_TIME_SAMPLES,
  MASH_TIME_SECONDS,
  MASH_TIME_BEAT24S
} MashTimeType;

/* Envelope types */
typedef enum {
  MASH_ENVELOPE_LINEAR,
  MASH_ENVELOPE_SPLINE
} MashEnvelopeType;

union _MashTime {
  long TIME;
  MashCount samples;
  float seconds;
  int beat24s;
};

typedef enum {
  MASH_TYPE_BOOL = 0,
  MASH_TYPE_INT,
  MASH_TYPE_FLOAT,
  MASH_TYPE_STRING,
  MASH_TYPE_BASE,
} MashParameterType;

typedef enum {
  MASH_CONSTRAINT_TYPE_NONE = 0,
  MASH_CONSTRAINT_TYPE_LIST,
  MASH_CONSTRAINT_TYPE_RANGE,
} MashConstraintType;


#define MASH_RANGE_LOWER_BOUND_VALID (1<<0)
#define MASH_RANGE_UPPER_BOUND_VALID (1<<1)
#define MASH_RANGE_STEP_VALID        (1<<2)
#define MASH_RANGE_ALL_VALID (MASH_RANGE_LOWER_BOUND_VALID | \
                              MASH_RANGE_UPPER_BOUND_VALID | \
                              MASH_RANGE_STEP_VALID)

#define MASH_HINT_DEFAULT  (0)
#define MASH_HINT_LOG      (1<<0)
#define MASH_HINT_TIME     (1<<1)
#define MASH_HINT_FILENAME (1<<2)

#define MASH_CONSTRAINT_EMPTY ((MashConstraint){NULL})

#define MASH_SAMPLES(x) ((MashTime){(MashCount)(x)})
#define MASH_SECONDS(x) ((MashTime){(float)(x)})
#define MASH_BEAT24S(x) ((MashTime){(int)(x)})


#if (defined(__MASH__) || defined(__MASH_PLUGIN__))
#include <mash_plugin.h>
#else
typedef MashOpaque MASH;
typedef MashOpaque MashPoint;
typedef MashOpaque MashEnvelope;
typedef MashOpaque MashChunk;
typedef MashOpaque MashChannel;
typedef MashOpaque MashStream;
typedef MashOpaque MashDeck;
typedef MashOpaque MashTrack;
typedef MashOpaque MashLayer;
typedef MashOpaque MashSound;
typedef MashOpaque MashMetaAuthor;
typedef MashOpaque MashMetaText;
typedef MashOpaque MashPlugin;
typedef MashOpaque MashSquareTone;
typedef MashOpaque MashMonitor;
#endif


#if defined(__cplusplus)
extern "C" {
#endif

extern CDSet * MASH_MONO;
extern CDSet * MASH_STEREO;


/* MASH */
MASH * mash_init (void);
MASH * mash_init_clone (MASH * mash);
void mash_purge (MASH * mash);
MashError mash_last_error (MASH * mash);
char * mash_error_string (MASH * mash, MashError error);
MashSamplerate mash_set_samplerate (MASH * mash, MashSamplerate samplerate);
MashSamplerate mash_get_samplerate (MASH * mash);
MashTempo mash_set_tempo (MASH * mash, MashTempo tempo);
MashTempo mash_get_tempo (MASH * mash);
CDSet * mash_set_channels (MASH * mash, CDSet * channelset);
CDSet * mash_get_channels (MASH * mash);

#if 0
  /* XXX */
/* Sources: Plugins, Samples etc. */
CDList * mash_open (MASH mash, char * filename);
void mash_close (MASH mash, MashPlugin plugin);
#endif

/* Base objects */

MashPlugin * mash_find_plugin (MASH * mash, char * identifier);

MashBase * mash_new (MASH * mash, MashPlugin * plugin, CDSet * parameters);
CDSet * mash_suggest (MASH * mash, MashPlugin * plugin, CDSet * parameters);

int mash_get_init_parameter_key (MASH * mash, MashPlugin * plugin, char * name);
int mash_get_parameter_key (MASH * mash, MashBase * base, char * name);

MashParameter mash_set_parameter (MASH * mash, MashBase * base, int key,
				  MashParameter parameter);
MashParameter mash_get_parameter (MASH * mash, MashBase * base, int key);
MashParameterType mash_get_parameter_type (MASH * mash, MashBase * base, int key);


MashBase * mash_clone_subclass (MASH * mash, MashBase * base);
int mash_destroy (MASH * mash, MashBase * base);
int mash_destroy_list (MASH * mash, CDList * list);
MashBase * mash_prepare (MASH * mash, MashBase * base);
MashCount mash_process (MASH * mash, MashBase * base, MashCount count,
		    MashStream * input, MashStream * output);
MashCount mash_length (MASH * mash, MashBase * base);
MashCount mash_seek (MASH * mash, MashBase * base, MashCount offset, int whence);
MashCount mash_tell (MASH * mash, MashBase * base);
int mash_flush (MASH * mash, MashBase * base);


/* Decks */
MashDeck * mash_deck_new (MASH * mash);

MashCount mash_deck_set_mixlength (MASH * mash, MashDeck * deck, MashCount mixlength);
MashCount mash_deck_get_mixlength (MASH * mash, MashDeck * deck);

CDList * mash_deck_get_tracks (MASH * mash, MashDeck * deck);

/* Tracks */
MashTrack * mash_add_track (MASH * mash, MashDeck * deck);
MashPCM mash_track_set_gain (MASH * mash, MashTrack * track, MashPCM gain);
MashPCM mash_track_get_gain (MASH * mash, MashTrack * track);
MashCount mash_track_set_mixlength (MASH * mash, MashTrack * track,
				MashCount mixlength);
MashCount mash_track_get_mixlength (MASH * mash, MashTrack * track);
void mash_remove_track (MASH * mash, MashTrack * track);

MashDeck * mash_track_get_deck (MASH * mash, MashTrack * track);


/* Layers */
MashLayer * mash_add_layer_ontop (MASH * mash, MashTrack * track,
			      MashTimeType timetype);
MashLayer * mash_add_layer_above (MASH * mash, MashLayer * above,
			      MashTimeType timetype);
MashLayer * mash_move_layer_ontop (MASH * mash, MashLayer * layer, MashTrack * track);
MashLayer * mash_move_layer_above (MASH * mash, MashLayer * layer, MashLayer * above);
MashLayer * mash_raise_layer (MASH * mash, MashLayer * layer);
MashLayer * mash_lower_layer (MASH * mash, MashLayer * layer);
MashTrack * mash_layer_get_track (MASH * mash, MashLayer * layer);
MashDeck * mash_layer_get_deck (MASH * mash, MashLayer * layer);
MashTimeType mash_layer_set_timetype (MASH * mash, MashLayer * layer,
				  MashTimeType new_type);
MashTimeType mash_layer_get_timetype (MASH * mash, MashLayer * layer);
MashSound * mash_layer_get_sound_before (MASH * mash, MashLayer * layer,
				     MashTime time);
MashSound * mash_layer_get_sound_at (MASH * mash, MashLayer * layer, MashTime time);
MashSound * mash_layer_get_sound_after (MASH * mash, MashLayer * layer, MashTime time);
MashLayer * mash_get_layer_below (MASH * mash, MashLayer * layer);
MashLayer * mash_get_layer_above (MASH * mash, MashLayer * layer);

/* Sounds */
MashSound * mash_add_sound (MASH * mash, MashBase * source, MashLayer * layer,
		      MashTime start_time, MashTime duration);
MashSound * mash_remove_sound (MASH * mash, MashSound * sound);

MashBase * mash_sound_set_source (MASH * mash, MashSound * sound, MashBase * source);
MashBase * mash_sound_get_source (MASH * mash, MashSound * sound);
MashLayer * mash_sound_get_layer (MASH * mash, MashSound * sound);
MashTrack * mash_sound_get_track (MASH * mash, MashSound * sound);
MashDeck * mash_sound_get_deck (MASH * mash, MashSound * sound);

MashTime mash_sound_move (MASH * mash, MashSound * sound, MashTime time);
MashSound * mash_sound_get_prev (MASH * mash, MashSound * sound);
MashSound * mash_sound_get_next (MASH * mash, MashSound * sound);

MashTime mash_sound_set_start_time (MASH * mash, MashSound * sound, MashTime time);
MashTime mash_sound_get_start_time (MASH * mash, MashSound * sound);
MashTime mash_sound_set_duration (MASH * mash, MashSound * sound, MashTime time);
MashTime mash_sound_get_duration (MASH * mash, MashSound * sound);

MashBase * mash_sound_set_rate_envelope (MASH * mash, MashSound * sound,
				     MashBase * rate_envelope);
MashBase * mash_sound_get_rate_envelope (MASH * mash, MashSound * sound);
MashBase * mash_sound_set_gain_envelope (MASH * mash, MashSound * sound,
				     MashBase * gain_envelope);
MashBase * mash_sound_get_gain_envelope (MASH * mash, MashSound * sound);
MashBase * mash_sound_set_blend_envelope (MASH * mash, MashSound * sound,
				      MashBase * blend_envelope);
MashBase * mash_sound_get_blend_envelope (MASH * mash, MashSound * sound);


/* Envelopes */
MashEnvelope * mash_envelope_new (MASH * mash, MashEnvelopeType type);
MashEnvelopeType mash_envelope_set_type (MASH * mash, MashEnvelope * envelope,
				     MashEnvelopeType type);
MashEnvelopeType mash_envelope_get_type (MASH * mash, MashEnvelope * envelope);
MashTimeType mash_envelope_set_timetype (MASH * mash, MashEnvelope * envelope,
				     MashTimeType timetype);
MashTimeType mash_envelope_get_timetype (MASH * mash, MashEnvelope * envelope);
MashPCM mash_envelope_get_value (MASH * mash, MashEnvelope * envelope, MashTime time);
MashTime mash_envelope_get_duration (MASH * mash, MashEnvelope * envelope);
MashPCM mash_envelope_get_integral (MASH * mash, MashEnvelope * envelope,
				MashTime t1, MashTime t2);
MashPoint * mash_envelope_add_point (MASH * mash, MashEnvelope * envelope,
				 MashTime time, MashPCM value);
MashEnvelope * mash_envelope_remove_point (MASH * mash, MashEnvelope * envelope,
				       MashPoint * point);
MashEnvelope * mash_envelope_scale (MASH * mash, MashEnvelope * envelope,
				MashPCM gain);
MashEnvelope * mash_envelope_shift (MASH * mash, MashEnvelope * envelope,
				MashTime delta);


/* Streams */
MashStream * mash_stream_new (MASH * mash);
MashStream * mash_stream_new_contiguous (MASH * mash, MashCount length);
MashCount mash_stream_nr_channels (MASH * mash, MashStream * stream);
MashChannel * mash_stream_find_channel (MASH * mash, MashStream * stream, int name);
MashChannel * mash_stream_add_channel (MASH * mash, MashStream * stream, int name);
MashStream * mash_stream_remove_channel (MASH * mash, MashStream * stream, int name);
MashStream * mash_stream_add_chunks (MASH * mash, MashStream * stream,
				 MashCount offset, MashCount length);

MashCount mash_stream_write0 (MASH * mash, MashStream * stream, MashCount count);
MashCount mash_stream_write (MASH * mash, MashStream * stream, MashCount count,
			     MashStream * data);
MashCount mash_stream_copy (MASH * mash, MashStream * src, MashStream * dest,
			    MashCount count);
MashCount mash_stream_gain (MASH * mash, MashStream * stream, MashCount count,
			    MashPCM gain);
MashCount mash_stream_mix (MASH * mash, MashStream * src, MashStream * dest,
			   MashCount count);
MashCount mash_stream_mult (MASH * mash, MashStream * src, MashStream * dest,
			    MashCount count);
MashCount mash_streams_mix (MASH * mash, CDList * streams, MashStream * dest,
			    MashCount count);
MashCount mash_stream_fade (MASH * mash, MashStream * src, MashStream * dest,
			    MashCount count);
MashCount mash_stream_blend (MASH * mash, MashStream * src, MashStream * dest,
			     MashStream * blend, MashCount count);

MashCount mash_stream_interleave_2 (MASH * mash, MashStream * stream,
				int name1, int name2,
				MashPCM * dest, MashCount count);
MashCount mash_stream_deinterleave_2 (MASH * mash, MashStream * stream,
				  int name1, int name2,
				  MashPCM * src, MashCount count);

/* Channels */
MashChannel * mash_add_channel (MASH * mash, MashStream * stream,
			    MashChannelName name);
void mash_free_channel (MASH * mash, MashChannel * channel);
void mash_remove_chunk (MASH * mash, MashChunk * chunk);

/* Chunks */
MashChunk * mash_add_chunk (MASH * mash, MashChannel * channel, MashCount length,
			MashCount offset);
int mash_chunk_later (MASH * mash, MashChunk * u1, MashChunk * u2);
MashCount mash_chunk_clear (MASH * mash, MashChunk * chunk);

/* SquareTone */
MashSquareTone * mash_squaretone_new (MASH * mash, float frequency);
float mash_squaretone_set_frequency (MASH * mash, MashSquareTone * squaretone,
				   float frequency);
float mash_squaretone_get_frequency (MASH * mash, MashSquareTone * squaretone);

/* Monitor */
MashMonitor * mash_monitor_new (MASH * mash);


/* Time */
MashTime mash_time_convert (MASH * mash, MashTime time, MashTimeType old_type,
			MashTimeType new_type);
MashTime mash_time_zero (MashTimeType type);
MashTime mash_time_invalid (MashTimeType type);
int mash_time_is_invalid (MashTimeType type, MashTime time);
MashTime mash_time_add (MashTimeType type, MashTime t1, MashTime t2);
MashTime mash_time_sub (MashTimeType type, MashTime t1, MashTime t2);
MashTime mash_time_min (MashTimeType type, MashTime t1, MashTime t2);
MashTime mash_time_max (MashTimeType type, MashTime t1, MashTime t2);
int mash_time_eq (MashTimeType type, MashTime t1, MashTime t2);
int mash_time_gt (MashTimeType type, MashTime t1, MashTime t2);
int mash_time_lt (MashTimeType type, MashTime t1, MashTime t2);
int mash_time_ge (MashTimeType type, MashTime t1, MashTime t2);
int mash_time_le (MashTimeType type, MashTime t1, MashTime t2);


MashCount mash_set_mixlength (MASH * mash, MashCount mixlength);
MashCount mash_get_mixlength (MASH * mash);


int mash_is_writeable (MASH * mash, MashBase * base);
int mash_is_seekable (MASH * mash, MashBase * base);
int mash_is_cacheable (MASH * mash, MashBase * base);
int mash_is_causal (MASH * mash, MashBase * base);

char * mash_set_name (MASH * mash, MashBase * base, char * name);
char * mash_get_name (MASH * mash, MashBase * base);

  /* XXX: Move meta out of mash.h ??? */

char * mash_meta_text_get_identifier (MASH * mash, MashMetaText * mt);
char * mash_meta_text_set_identifier (MASH * mash, MashMetaText * mt,
				    char * identifier);
char * mash_meta_text_get_category (MASH * mash, MashMetaText * mt);
char * mash_meta_text_set_category (MASH * mash, MashMetaText * mt,
				  char * category);
char * mash_meta_text_get_description (MASH * mash, MashMetaText * mt);
char * mash_meta_text_set_description (MASH * mash, MashMetaText * mt,
				     char * description);
char * mash_meta_text_get_copyright (MASH * mash, MashMetaText * mt);
char * mash_meta_text_set_copyright (MASH * mash, MashMetaText * mt,
				   char * copyright);
char * mash_meta_text_get_url (MASH * mash, MashMetaText * mt);
char * mash_meta_text_set_url (MASH * mash, MashMetaText * mt, char * url);
CDList * mash_meta_text_get_authors (MASH * mash, MashMetaText * mt);
void mash_meta_text_add_author (MASH * mash, MashMetaText * mt, char * name,
			      char * email);

#if defined(__cplusplus)
}
#endif

#endif /* __MASH_H__ */
