/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashBase: A generic interface for sound generation and processing.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <string.h>

#define __MASH__
#include "mash.h"


MashBase *
mash_new (MASH * mash, MashPlugin * plugin, CDSet * parameters)
{
  MashBase * base;

  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  base = mash_base_new (mash);

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  base->plugin = plugin;

  if (plugin->init != NULL) {
    if (plugin->init (mash, base, parameters) == MashNone) {
      mash_destroy (mash, base);
      return MashNone;
    }
  }

  return base;
}

CDSet *
mash_suggest (MASH * mash, MashPlugin * plugin, CDSet * parameters)
{
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  if (!plugin->suggest) {
    /* XXX: should mash_set_error: is NOOP appropriate ?? */
    return MashNone;
  }

  return plugin->suggest (mash, plugin, parameters, plugin->plugin_data);
}

static int
mash_parameter_scheme_get_key (MASH * mash, CDSet * scheme_set, char * name)
{
  CDSet * s;
  MashParameterScheme * scheme;

  for (s = scheme_set; s; s = s->next) {
    scheme = (MashParameterScheme *)s->data.s_pointer;
    if (!strcmp (scheme->name, name)) return s->key;
  }

  mash_set_error (mash, MASH_ERROR_NOENTITY);

  return -1;
}

int
mash_get_init_parameter_key (MASH * mash, MashPlugin * plugin, char * name)
{
  if (plugin == MashNone) {
    mash_dprintf ("[mash_get_init_parameter_key] plugin == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return mash_parameter_scheme_get_key (mash, plugin->init_scheme, name);
}

int
mash_get_parameter_key (MASH * mash, MashBase * base, char * name)
{
  if (base == MashNone) {
    mash_dprintf ("[mash_get_parameter_key] base == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  if (base->plugin == MashNone) {
    mash_dprintf ("[mash_get_parameter_key] base->plugin == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return mash_parameter_scheme_get_key (mash, base->plugin->process_scheme, name);
}

MashParameter
mash_set_parameter (MASH * mash, MashBase * base, int key,
		    MashParameter parameter)
{
  if (base == MashNone) {
    mash_dprintf ("[mash_set_parameter] base == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return (MashParameter)-1;
  }

  mash_dprintf ("[mash_set_parameter] base %p, [%d] ==> %p\n", base, key,
	      parameter.s_pointer);
  base->parameters = cd_set_replace (mash, base->parameters, key, parameter);
  return parameter;
}

MashParameter
mash_get_parameter (MASH * mash, MashBase * base, int key)
{
  MashParameter p;

  if (base == MashNone) {
    mash_dprintf ("[mash_get_parameter] base == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return (MashParameter)-1;
  }

  p = cd_set_find (mash, base->parameters, key);
  mash_dprintf ("[mash_get_parameter] base %p, [%d] == %p\n", base, key,
	      p.s_pointer);
  return p;
}

MashParameterType
mash_get_parameter_type (MASH * mash, MashBase * base, int key)
{
  MashPlugin * plugin;
  CDScalar k;
  MashParameterScheme * scheme;

  if (base == MashNone) {
    mash_dprintf ("[mash_get_parameter_type] base == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;

  if (plugin == MashNone) {
    mash_dprintf ("[mash_get_parameter_type] base->plugin == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  k = cd_set_find (mash, plugin->process_scheme, key);
  scheme = (MashParameterScheme *)k.s_pointer;

  if (scheme == MashNone) {
    mash_dprintf ("[mash_get_parameter_type] scheme == MashNone\n");
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return scheme->type;
}

MashBase *
mash_base_new_subclass (MASH * mash, size_t size)
{
  MashBase * base = mash_malloc (size);
  _mash_context_copy (mash, &base->context_limit);
  _mash_register_base (mash, base);
  return base;
}

MashBase *
mash_base_new (MASH * mash)
{
  return mash_base_new_subclass (mash, sizeof (struct _MashBase));
}

MashCount
mash_base_get_mixlength (MASH * mash, MashBase * base)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return _mash_base_get_mixlength (mash, base);
}

MashSamplerate
mash_base_get_samplerate (MASH * mash, MashBase * base)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return _mash_base_get_samplerate (mash, base);
}

MashTempo
mash_base_get_tempo (MASH * mash, MashBase * base)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return _mash_base_get_tempo (mash, base);
}

CDSet *
mash_base_get_channels (MASH * mash, MashBase * base)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  return _mash_base_get_channels (mash, base);
}

void *
mash_base_set_instance_data (MASH * mash, MashBase * base, void * data)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  return _mash_set_instance_data (mash, base, data);
}

void *
mash_base_get_instance_data (MASH * mash, MashBase * base)
{
  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  return _mash_get_instance_data (mash, base);
}

int
mash_base_has_samplerate (MASH * mash, MashBase * base)
{
  MashSamplerate asr, bsr;

  asr = mash_get_samplerate (mash);
  bsr = _mash_base_get_samplerate (mash, base);

  return (asr == bsr);
}

int
mash_base_has_tempo (MASH * mash, MashBase * base)
{
  MashTempo at, bt;

  at = mash_get_tempo (mash);
  bt = _mash_base_get_tempo (mash, base);

  return (at == bt);
}

int
mash_base_encompasses_mixlength (MASH * mash, MashBase * base)
{
  MashCount aml, bml;

  aml = mash_get_mixlength (mash);
  bml = _mash_base_get_mixlength (mash, base);

  return (aml < bml);
}

int
mash_base_encompasses_channels (MASH * mash, MashBase * base)
{
  CDSet * s, * as, * bs;
  CDScalar k;

  as = mash_get_channels (mash);
  bs = _mash_base_get_channels (mash, base);

  for (s = as; s; s = s->next) {
    k = cd_set_find (mash, bs, s->key);
    if (k.s_pointer == NULL) return 0;
  }

  return 1;
}

MashMethods *
mash_base_set_methods (MASH * mash, MashBase * base, MashMethods * methods)
{
  MashMethods * old = base->methods;
  _mash_set_methods (mash, base, methods);
  return old;
}

MashMethods *
mash_base_get_methods (MASH * mash, MashBase * base)
{
  return _mash_get_methods (mash, base);
}

MashPlugin *
mash_base_set_plugin (MASH * mash, MashBase * base, MashPlugin * plugin)
{
  MashPlugin * old = base->plugin;
  _mash_set_plugin (mash, base, plugin);
  return old;
}

MashPlugin *
mash_base_get_plugin (MASH * mash, MashBase * base)
{
  return _mash_get_plugin (mash, base);
}

MashBase *
mash_clone_subclass (MASH * mash, MashBase * base)
{
  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return NULL;
  }
  if (!base->methods || !base->methods->clone) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return NULL;
  }
  return _mash_clone (mash, base);
}

int
mash_destroy (MASH * mash, MashBase * base)
{
  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  _mash_unregister_base (mash, base);

  if (!base->methods || !base->methods->destroy) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return -1;
  }
  return _mash_destroy (mash, base);
}

int
mash_destroy_list (MASH * mash, CDList * list)
{
  cd_list_destroy_with (mash, list, (CDDestroyFunc)mash_destroy);
  return 0;
}

/*
 * Prepare the methods for process, seek and length calls.
 *
 * "Prepare" means to make sure the base has enough internal buffers
 * to deal with the current context (sample rate, mixlength).
 *
 * If the methods has a ready() function, that is checked first. If it
 * does not have a ready() function, it is assumed to always be ready.
 */
MashBase *
mash_prepare (MASH * mash, MashBase * base)
{
  int is_ready = 0;

  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return NULL;
  }
  if (base->methods && base->methods->prepare) {
    if (base->methods->ready) {
      is_ready = base->methods->ready (mash, base);
    }

    _mash_context_merge (mash, &base->context_limit);

    if (!is_ready) base =_mash_prepare (mash, base);
  }

  return base;
}

MashCount
mash_process_fast (MASH * mash, MashBase * base, MashCount count,
		   MashStream * input, MashStream * output)
{
  MashCount n;

  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->process) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return -1;
  }
  n = _mash_process (mash, base, count, input, output);
  if (n > 0) base->offset += n;
  return n;
}

MashCount
mash_process (MASH * mash, MashBase * base, MashCount count,
	      MashStream * input, MashStream * output)
{
  MashCount processed;
  MashCount n;
  MashError error;
  char * str;

  mash_debug_down ();

  processed = mash_process_fast (mash, base, count, input, output);

  if (processed == -1) {
    error = mash_last_error (mash);
    str = mash_error_string (mash, error);
    mash_dprintf ("*** ERROR in mash_process: %s\n", str);
    switch (error) {
    case MASH_ERROR_NOOP:
      n = mash_stream_write (mash, output, count, input);
      if (n > 0) {
	base->offset += n;
	processed = n;
      }
      break;
    case MASH_ERROR_SILENCE:
      n = mash_stream_write0 (mash, output, count);
      if (n > 0) {
	base->offset += n;
	processed = n;
      }
      break;
    default: break;
    }
  }

  mash_debug_up ();

  return processed;
}

MashCount
mash_length (MASH * mash, MashBase * base)
{
  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->length) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return -1;
  }
  return _mash_length (mash, base);
}  

MashCount
mash_seek (MASH * mash, MashBase * base, MashCount offset, int whence)
{
  MashCount new_offset, len;

  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  new_offset = base->offset;

  switch (whence) {
  case SEEK_SET: new_offset = offset; break;
  case SEEK_CUR: new_offset += offset; break;
  case SEEK_END:
    len = mash_length (mash, base);
    if (len == -1) return -1;
    new_offset = len + offset;
    break;
  default:
    mash_set_error (mash, MASH_ERROR_INVALID);
    return -1;
    break;
  }

  if (new_offset == base->offset) return new_offset;

  mash_dprintf ("SEEK %p @ %ld\n", base, new_offset);

  if (base->methods && base->methods->seek)
    base->offset = base->methods->seek (mash, base, new_offset);
  else
    base->offset = new_offset;

  return base->offset;
}

MashCount
mash_tell (MASH * mash, MashBase * base)
{
  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }
  return base->offset;
}

int
mash_flush (MASH * mash, MashBase * base)
{
  if (!base) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }
  if (!base->methods || !base->methods->flush) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return -1;
  }
  return _mash_flush (mash, base);
}

MashMetaText *
mash_get_meta_text (MASH * mash, MashBase * base)
{
  MashPlugin * plugin = base->plugin;
  return (plugin ? plugin->metatext : NULL);
}

MashMetaText *
mash_set_meta_text (MASH * mash, MashBase * base, MashMetaText * mt)
{
  MashPlugin * plugin;
  MashMetaText * old;

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  plugin = base->plugin;
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  old = plugin->metatext;
  plugin->metatext = mt;
  return old;
}

int
mash_is_writeable (MASH * mash, MashBase * base)
{
  MashPlugin * plugin;

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & MASH_PLUGIN_WRITEABLE);
}

int
mash_is_seekable (MASH * mash, MashBase * base)
{
  MashPlugin * plugin;

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & MASH_PLUGIN_SEEKABLE);
}

int
mash_is_cacheable (MASH * mash, MashBase * base)
{
  MashPlugin * plugin;

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & MASH_PLUGIN_CACHEABLE);
}

int
mash_is_causal (MASH * mash, MashBase * base)
{
  MashPlugin * plugin;

  if (base == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  plugin = base->plugin;
  if (plugin == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return (plugin->flags & MASH_PLUGIN_CAUSAL);
}
