/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashChannel: An indexed, sparse, monophonic PCM data container.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A channel contains a sequence of chunks which contain raw PCM data.
 *
 * Channels are named, such that when streams are mixed together their
 * correspondingly named channels are mixed together.
 *
 * A channel is indexed by sample count. When reading a channel, all
 * index points for which no chunk is defined are defined as zero. Writing
 * to a channel stops early if no chunk is available at a required point.
 *
 * If two or more chunks overlap, the chunk with the latest start index
 * is always used for both reading and writing data values for the region
 * of overlap.
 *
 * Invariants
 * ----------
 *
 * A channel must be contained within a stream.
 */

#define __MASH__
#include "mash.h"

/* MashChannel */

MashChannel *
mash_channel_new (MASH * mash)
{
  MashChannel * c;

  c = (MashChannel *) mash_malloc (sizeof (struct _MashChannel));

  c->chunks = cd_list_new (mash);
  c->_current_offset = 0;
  c->_current_chunk = MashNone;

  return c;
}

MashChannel *
mash_channel_clone (MASH * mash, MashChannel * channel)
{
  MashChannel * new_channel = mash_channel_new (mash);
  new_channel->chunks = cd_list_clone (mash, channel->chunks,
				       (CDCloneFunc)mash_chunk_clone);
  return new_channel;
}

int
mash_channel_destroy (MASH * mash, MashBase * base)
{
  MashChannel * channel = (MashChannel *)base;
  CDList * l;

  for (l = channel->chunks; l; l = l->next) {
    mash_chunk_free (mash, (MashChunk *)l->data.s_pointer);
  }

  cd_list_free (mash, channel->chunks);
  mash_free (channel);
  return 0;
}

MashChunk *
mash_channel_add_chunk (MASH * mash, MashChannel * channel, MashChunk * chunk)
{
  channel->chunks = cd_list_insert (mash, channel->chunks, CD_TYPE_POINTER,
                                    CD_POINTER(chunk),
				    (CDCmpFunc)mash_chunk_later);
  return chunk;
}

MashChunk *
mash_channel_add_new_chunk (MASH * mash, MashChannel * channel, MashCount offset,
			  MashCount length)
{
  MashChunk * chunk = mash_chunk_new (mash, offset, length);
  return mash_channel_add_chunk (mash, channel, chunk);
}

void
mash_channel_remove_chunk (MASH * mash, MashChannel * channel,
			   MashChunk * chunk)
{
  channel->chunks = cd_list_remove (mash, channel->chunks, CD_TYPE_POINTER,
				    CD_POINTER(chunk));
}

MashChunk *
mash_channel_find_chunk_before (MASH * mash, MashChannel * channel,
				MashCount index)
{
  CDList * l;
  MashChunk * u, * up = MashNone;

  for (l = channel->chunks; l; l = l->next) {
    u = (MashChunk *)l->data.s_pointer;
    if (u->start_index > index) return up;
    up = u;
  }

  return up;
}

/*
 * mash_chunk_item_valid_length (l)
 *
 * Returns the length for which the chunk item 'l' is valid. The valid
 * length of a chunk item in a channel is defined as the minimum of
 * (the chunk's actual length) and (the difference between the offset
 * of the chunk and the offset of the following chunk). ie. if the
 * chunk is followed by another chunk that cuts it off early, the valid
 * length is the difference between the two chunk offsets. Crikey, email
 * me if you're confused.
 */
static MashCount
mash_chunk_item_valid_length (CDList * l)
{
  CDList * ln = l->next;
  MashChunk * u = (MashChunk *)l->data.s_pointer, * un;

  if (ln == MashNone) {
    return u->length;
  } else {
    un = (MashChunk *)ln->data.s_pointer;
    return MIN (u->length, un->start_index - u->start_index);
  }
}

CDList *
mash_channel_get_chunk_item_at (MashChannel * channel, MashCount offset)
{
  CDList * l;
  MashChunk * u;
  MashCount vl; /* valid length */

  for (l = channel->chunks; l; l = l->next) {
    u = (MashChunk *)l->data.s_pointer;
    vl = mash_chunk_item_valid_length (l);
    if (u->start_index <= offset && u->start_index + vl > offset) {
      /* chunk validly spans offset: return its item */
      return l;
    }
  }

  /* No chunks found spanning offset */
  return MashNone;
}

CDList *
mash_channel_get_chunk_item_after (MashChannel * channel, MashCount offset)
{
  CDList * l;
  MashChunk * u;

  for (l = channel->chunks; l; l = l->next) {
    u = (MashChunk *)l->data.s_pointer;
    if (u->start_index >= offset) return l;
  }

  return MashNone;
}

MashCount
mash_channel_write0 (MASH * mash, MashChannel * channel, MashCount length)
{
  CDList * l = channel->_current_chunk;
  MashChunk * u;
  MashCount remaining = length, n, vl;
  MashCount offset = channel->_current_offset;

  while (remaining > 0) {
    if (l == MashNone) break; /* No more chunks */

    u = (MashChunk *)l->data.s_pointer;

    if (u->start_index > offset) { /* skip ahead to start of next chunk */
      n = MIN (remaining, u->start_index - offset);
      offset += n;
      remaining -= n;
    }

    if (remaining > 0) {
      vl = mash_chunk_item_valid_length (l);
      n = _mash_chunk_clear_region (mash, u, offset, MIN(remaining, vl),
				  0, NULL);
      offset += n;
      remaining -= n;
    }

    l = l->next;
  }

  channel->_current_chunk = l;
  channel->_current_offset += length;

  return length;
}

/*
 * mash_channel_chunkfuncify (mash, channel, count, func, data)
 *
 * Apply the MashChunkFunc func to 'count' samples from consecutive chunks
 * of 'channel'.
 * Stops early if the channel runs out of chunks.
 * Returns the number of samples func'ed.
 */
MashCount
mash_channel_chunkfuncify (MASH * mash, MashChannel * channel, MashCount count,
			   MashChunkFunc func, int channelname, void * data)
{
  MashChunk * u;
  MashCount remaining = count, funced = 0, n, vl;
  MashError error;

  mash_dprintf ("[mash_channel_chunkfuncify] (%p, +%ld) @ %ld\n",
	  channel, count, channel->_current_offset);

  while (remaining > 0) {
    channel->_current_chunk =
      mash_channel_get_chunk_item_at (channel, channel->_current_offset);
    if (channel->_current_chunk == MashNone) {
      mash_dprintf ("[mash_channel_chunkfuncify] channel incomplete, funced %ld\n",
		  funced);
      return funced; /* Channel incomplete */
    }

    u = (MashChunk *)channel->_current_chunk->data.s_pointer;
    vl = mash_chunk_item_valid_length (channel->_current_chunk);

    n = func (mash, u, channel->_current_offset, MIN(remaining, vl),
	      channelname, data);

    if (n == -1) {
      error = mash_last_error (mash);
      switch (error) {
      case MASH_ERROR_SILENCE:
	n = _mash_chunk_clear_region (mash, u, channel->_current_offset,
				    MIN(remaining, vl), 0, NULL);
	break;
      default:
	n = 0;
      }
    }

    funced += n;
    remaining -= n;
    channel->_current_offset += n;
  }

  return funced;
}

/*
 * mash_channel_chunkchunkfuncify (mash, src, dest, count, func, data)
 *
 * Apply the MashChunkChunkFunc func to corresponding chunks of 'src' and
 * 'dest' to 'count' samples.
 * Stops early if 'dest' cannot contain part of the region for which
 * 'src' is defined. Copies zeroes to 'dest' wherever 'src' is empty.
 * Returns the number of samples func'ed.
 */
MashCount
mash_channel_chunkchunkfuncify (MASH * mash, MashChannel * src, MashChannel * dest,
			      MashCount count, MashChunkChunkFunc func,
			      int channelname, void * data)
{
  MashChunk * su, * du;
  MashCount remaining = count, funced = 0, n, vl;
  MashError error;

  mash_dprintf ("[mash_channel_ccf...] (%p -> %p, +%ld), src @ %ld, dest @ %ld\n",
	  src, dest, count, src->_current_offset, dest->_current_offset);

  while (remaining > 0) {
    n = 0; /* watch for early changes to n */

    dest->_current_chunk =
	mash_channel_get_chunk_item_at (dest, dest->_current_offset);
    if (dest->_current_chunk == MashNone) {
      mash_dprintf ("[mash_channel_ccf...] channel incomplete after %ld\n", funced);
      return funced; /* Destination channel incomplete */
    }

    src->_current_chunk =
	mash_channel_get_chunk_item_at (src, src->_current_offset);
    if (src->_current_chunk == MashNone) { /* No source data at offset */
      src->_current_chunk =
	mash_channel_get_chunk_item_after (src, src->_current_offset);

      if (src->_current_chunk == MashNone) {
        /* No following source data at all */
	mash_dprintf ("[mash_channel_ccf...] no source data after %ld\n",
		    src->_current_offset);
	n = mash_channel_write0 (mash, dest, remaining);
	funced += n;
	remaining -= n;
	return funced;
      } else {
	mash_dprintf ("[mash_channel_ccf...] no source data at %ld\n",
		    src->_current_offset);
      }
    }

    /* *** Now, sl is the current or following source chunk item *** */

    su = (MashChunk *)src->_current_chunk->data.s_pointer;

    if (su->start_index > src->_current_offset) { /* No source data at offset */
      mash_dprintf ("[mash_channel_ccf...] no source data at %ld (warn 2)\n",
		  src->_current_offset);
      n = mash_channel_write0 (mash, dest, su->start_index - src->_current_offset);
      funced += n;
      remaining -= n;
      src->_current_offset += n;
    }

    if (remaining > 0) {
      if (n > 0) {
	dest->_current_chunk =
	  mash_channel_get_chunk_item_at (dest, dest->_current_offset);
	if (dest->_current_chunk == MashNone) {
	  mash_dprintf ("[mash_channel_ccf...] dest incomplete after %ld\n",
		      dest->_current_chunk);
	  return funced; /* Destination channel incomplete */
	}
      }
 
      du = (MashChunk *)dest->_current_chunk->data.s_pointer;

      vl = mash_chunk_item_valid_length (dest->_current_chunk);
      n = func (mash, su, src->_current_offset, du, dest->_current_offset,
                MIN(remaining, vl), channelname, data);

      if (n == -1) {
	error = mash_last_error (mash);
	switch (error) {
	case MASH_ERROR_SILENCE:
	  n = _mash_chunk_clear_region (mash, du, dest->_current_offset,
				      MIN(remaining, vl), 0, NULL);
	  break;
	default:
	  n = 0;
	}
      }

      funced += n;
      remaining -= n;
      src->_current_offset += n;
      dest->_current_offset += n;
    }
  }

  mash_dprintf ("[mash_channel_ccf...] funced %ld\n", funced);

  return funced;
}

/*
 * mash_channel_chunkchunkchunkfuncify (mash, src1, src2, dest, count, func,
 *                                    data)
 *
 * Apply the MashChunkChunkChunkFunc func to corresponding chunks of 'src1',
 * 'src2' and 'dest' to 'count' samples.
 * Stops early if 'dest' cannot contain part of the region for which
 * both 'src1' and 'src2' is defined. Copies zeroes to 'dest' wherever
 * either 'src1' or 'src2' are empty.
 * Returns the number of samples func'ed.
 */
MashCount
mash_channel_chunkchunkchunkfuncify (MASH * mash,
				   MashChannel * src1, MashChannel * src2,
				   MashChannel * dest, MashCount count,
				   MashChunkChunkChunkFunc func,
				   int channelname, void * data)
{
  MashChunk * s1u, * s2u, * du;
  MashCount remaining = count, funced = 0, n, vl, undef_length;
  MashError error;

  while (remaining > 0) {
    n = 0; /* watch for early changes to n */

    dest->_current_chunk =
      mash_channel_get_chunk_item_at (dest, dest->_current_offset);
    if (dest->_current_chunk == MashNone)
      return funced; /* Destination channel incomplete */

    src1->_current_chunk =
      mash_channel_get_chunk_item_at (src1, src1->_current_offset);
    if (src1->_current_chunk == MashNone) {
      src1->_current_chunk =
	mash_channel_get_chunk_item_after (src1, src1->_current_offset);
    }

    src2->_current_chunk =
      mash_channel_get_chunk_item_at (src2, src2->_current_offset);
    if (src2->_current_chunk == MashNone) {
      src2->_current_chunk =
	mash_channel_get_chunk_item_after (src2, src2->_current_offset);
    }

    if (src1->_current_chunk == MashNone || src2->_current_chunk == MashNone) {
      n = mash_channel_write0 (mash, dest, remaining);
      funced += n;
      remaining -= n;
      return funced;
    }

    s1u = (MashChunk *)src1->_current_chunk->data.s_pointer;
    s2u = (MashChunk *)src2->_current_chunk->data.s_pointer;

    if (s1u->start_index > src1->_current_offset ||
	s2u->start_index > src2->_current_offset) {
      undef_length = MAX (s1u->start_index - src1->_current_offset,
			  s2u->start_index - src2->_current_offset);
      n = mash_channel_write0 (mash, dest, undef_length);
      funced += n;
      remaining -= n;
      src1->_current_offset += n;
      src2->_current_offset += n;
    }
    
    if (remaining > 0) {
      if (n > 0) {
	dest->_current_chunk =
          mash_channel_get_chunk_item_at (dest, dest->_current_offset);
	if (dest->_current_chunk == MashNone)
          return funced; /* Destination channel incomplete */
      }
      
      du = (MashChunk *)dest->_current_chunk->data.s_pointer;
      
      vl = mash_chunk_item_valid_length (dest->_current_chunk);
      n = func (mash, s1u, src1->_current_offset, s2u, src2->_current_offset,
		du, dest->_current_offset, MIN(remaining, vl), channelname,
		data);

      if (n == -1) {
	error = mash_last_error (mash);
	switch (error) {
	case MASH_ERROR_SILENCE:
	  n = _mash_chunk_clear_region (mash, du, dest->_current_offset,
				      MIN(remaining, vl), 0, NULL);
	  break;
	default:
	  n = 0;
	}
      }

      funced += n;
      remaining -= n;
      src1->_current_offset += n;
      src2->_current_offset += n;
      dest->_current_offset += n;
    }
  }

  return funced;
}

/*
 * mash_channel_copy (src, dest, count)
 *
 * Copies 'count' samples from 'src' to 'dest'.
 * Stops early if 'dest' cannot contain part of 'src'. Copies zeroes to
 * 'dest' wherever 'src' is correspondingly empty.
 * Returns the count of samples actually copied.
 */
MashCount
mash_channel_copy (MASH * mash, MashChannel * src, MashChannel * dest, MashCount count)
{
  return mash_channel_chunkchunkfuncify (mash, src, dest, count,
				       (MashChunkChunkFunc)_mash_chunk_copy,
				       0, NULL);
}

MashCount
mash_channel_mix (MASH * mash, MashChannel * src, MashChannel * dest, MashCount count)
{
  return mash_channel_chunkchunkfuncify (mash, src, dest, count,
                                       (MashChunkChunkFunc)_mash_chunk_add_inplace,
				       0, NULL);
}

MashCount
mash_channel_interleave_2 (MASH * mash, MashChannel * src1, MashChannel * src2,
			 MashPCM * dest, MashCount count)
{
  return
    mash_channel_chunkchunkfuncify (mash, src1, src2, count,
				  (MashChunkChunkFunc)_mash_chunk_interleave_2,
				  0, dest);
}

MashCount
mash_channel_deinterleave_2 (MASH * mash, MashChannel * dest1, MashChannel * dest2,
			   MashPCM * src, MashCount count)
{
  return
    mash_channel_chunkchunkfuncify (mash, dest1, dest2, count,
				  (MashChunkChunkFunc)_mash_chunk_deinterleave_2,
				  0, src);
}

MashCount
_mash_channel_write (MASH * mash, MashChannel * channel, MashCount count,
		   MashChannel * data)
{
  MashCount n = mash_channel_copy (mash, data, channel, count);
  data->_current_offset += n;
  data->_current_chunk =
    mash_channel_get_chunk_item_after (data, data->_current_offset);
  channel->_current_offset += n;
  channel->_current_chunk =
    mash_channel_get_chunk_item_after (channel, channel->_current_offset);
  return n;
}

MashCount
_mash_channel_length (MASH * mash, MashChannel * channel)
{
  MashChunk * last = (MashChunk *)
    (cd_list_last (mash, channel->chunks, CD_TYPE_POINTER)).s_pointer;
  if (last == MashNone) return 0;
  return (last->start_index + last->length);
}

MashCount
_mash_channel_seek (MASH * mash, MashChannel * channel, MashCount offset)
{
  if (offset == channel->_current_offset) return offset;
  channel->_current_offset = offset;
  /* Cache the current chunk */
  channel->_current_chunk =
    mash_channel_get_chunk_item_after (channel, offset);
  return offset;
}
