/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashChunk: A contiguous chunk of monophonic PCM data.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A chunk contains raw PCM data.
 *
 * Invariants
 * ----------
 *
 * A chunk must always be contained within a channel. The data within
 * a chunk is only valid where it is not overlapped by a later chunk
 * in the same channel; elsewhere, the chunk's data is not used.
 *
 */

#include <string.h>

#define __MASH__
#include "mash.h"

typedef MashCount (*MashPFunc) (MashPCM * src, MashCount count, void * data);
typedef MashCount (*MashPPFunc) (MashPCM * src, MashPCM * dest, MashCount count,
			     void * data);
typedef MashCount (*MashPPPFunc) (MashPCM * src1, MashPCM * src2, MashPCM * dest,
			      MashCount count, void * data);

MashChunk *
mash_chunk_new (MASH * mash, MashCount start_index, MashCount length)
{
  MashChunk * u;

  u = (MashChunk *) mash_malloc (sizeof (struct _MashChunk));
  u->start_index = start_index;
  u->length = length;

  u->data = (MashPCM *) mash_malloc (length * sizeof (MashPCM));

  return u;
}

MashChunk *
mash_chunk_clone (MASH * mash, MashChunk * chunk)
{
  MashChunk * u = mash_chunk_new (mash, chunk->start_index, chunk->length);
  memcpy (u->data, chunk->data, (chunk->length * sizeof (MashPCM)));
  return u;
}

void
mash_chunk_free (MASH * mash, MashChunk * chunk)
{
  mash_free (chunk->data);
  mash_free (chunk);
}

int
mash_chunk_later (MASH * mash, MashChunk * u1, MashChunk * u2)
{
  return (u1->start_index > u2->start_index);
}

MashCount
mash_chunk_clear (MASH * mash, MashChunk * chunk)
{
  MashCount len = chunk->length;
  memset (chunk->data, (MashPCM)0, len * sizeof (MashPCM));
  return len;
}


/*
 * FUNCTION APPLIERS
 */

static MashCount
_mash_pfunc_apply (MASH * mash, MashPFunc func, MashChunk * chunk, MashCount start,
		 MashCount count, void * data)
{
  MashCount chunk_start = start - chunk->start_index;

  /* chunk_start: 'start' relative to chunk */
  if (chunk_start < 0) {
    count += chunk_start;
    chunk_start = 0;
  }

  if (chunk_start + count > chunk->length) {
    count = chunk->length - chunk_start;
  }

  func (&chunk->data[chunk_start], count, data);

  return (count);
}

/*
 * _mash_ppfunc_apply (src, src_offset, dest, dest_offset, count, func)
 *
 * Apply MashPCMPCMFunc 'func' to 'count' samples of chunks 'src' and 'dest'.
 */
static MashCount
_mash_ppfunc_apply (MASH * mash, MashPPFunc func,
		  MashChunk * src, MashCount src_offset,
		  MashChunk * dest, MashCount dest_offset, MashCount count,
		  void * data)
{
  MashCount dest_start = dest_offset - dest->start_index;
  MashCount src_start;
  MashPCM *s, *d;

  if (dest_start < 0) {
    count += dest_start;
    dest_offset -= dest_start;
    src_offset -= dest_start;
    dest_start = 0;
  }

  src_start = src_offset - src->start_index;

  /* Shorten count if either chunk is shorter than required */
  if (src_start + count > src->length)
    count = src->length - src_start;

  if (dest_start + count > dest->length)
    count = dest->length - dest_start;

  s = &src->data[src_start];
  d = &dest->data[dest_start];

  func (s, d, count, data);

  return count;
}

/*
 * _mash_pppfunc_apply (src, src_offset, dest, dest_offset, count, func)
 *
 * Apply MashPPPFunc 'func' to 'count' samples of chunks 'src1', 'src2'
 * and 'dest'.
 */
static MashCount
_mash_pppfunc_apply (MASH * mash, MashPPPFunc func,
		   MashChunk * src1, MashCount src1_offset,
		   MashChunk * src2, MashCount src2_offset,
		   MashChunk * dest, MashCount dest_offset,
		   MashCount count, void * data)
{
  MashCount dest_start = dest_offset - dest->start_index;
  MashCount src1_start, src2_start;
  MashPCM *s1, *s2, *d;

  if (dest_start < 0) {
    count += dest_start;
    dest_offset -= dest_start;
    src1_offset -= dest_start;
    src2_offset -= dest_start;
    dest_start = 0;
  }

  src1_start = src1_offset - src1->start_index;
  src2_start = src2_offset - src2->start_index;

  /* Shorten count if any chunk is shorter than required */
  if (src1_start + count > src1->length)
    count = src1->length - src1_start;

  if (src2_start + count > src2->length)
    count = src2->length - src2_start;

  if (dest_start + count > dest->length)
    count = dest->length - dest_start;

  s1 = &src1->data[src1_start];
  s2 = &src2->data[src2_start];
  d = &dest->data[dest_start];

  func (s1, s2, d, count, data);

  return count;
}

/* PFunc appliers */

/*
 * _mash_chunk_clear_region (chunk, start, length, unused)
 *
 * Clear 'chunk' from stream index 'start' for 'length' samples.
 * Returns the count of samples actually cleared.
 */
MashCount
_mash_chunk_clear_region (MASH * mash, MashChunk * chunk, MashCount start,
			MashCount length, int channelname, void * unused)
{
  return _mash_pfunc_apply (mash,_mash_pcm_clear_region,
			  chunk, start, length, NULL);
			  
}

MashCount
_mash_chunk_gain (MASH * mash, MashChunk * chunk, MashCount start, MashCount count,
		int channelname, /* (MashPCM *) */ void * gain)
{
  return _mash_pfunc_apply (mash, _mash_pcm_gain, chunk, start, count, gain);
}


/* PPFunc appliers */

/*
 * _mash_chunk_copy (src, dest, offset, count, unused)
 *
 * Copy data from 'src' to 'dest' from stream index 'offset' for
 * 'count'. Returns the count of samples actually copied.
 */
MashCount
_mash_chunk_copy (MASH * mash, MashChunk * src, MashCount src_offset,
		MashChunk * dest, MashCount dest_offset, MashCount count,
		int channelname, void * unused)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_copy, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _mash_chunk_add_inplace (src, dest, start, count, unused)
 *
 * Add data from 'src' to data in 'dest' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually added.
 */
MashCount
_mash_chunk_add_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
		       MashChunk * dest, MashCount dest_offset, MashCount count,
		       int channelname, void * unused)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_add, src, src_offset,
			   dest, dest_offset, count, NULL);
			   
}

/*
 * _mash_chunk_mult_inplace (src, dest, start, count, unused)
 *
 * Multiply data of 'dest' by that of 'src' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually multiplied.
 */
MashCount
_mash_chunk_mult_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
			MashChunk * dest, MashCount dest_offset, MashCount count,
			int channelname, void * unused)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_mult, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _mash_chunk_fade_inplace (src, src_offset, dest, dest_offset, count, unused)
 *
 * Fade data of 'dest' by that of 'src' from stream index 'start' for
 * 'count' samples. Returns the count of samples actually faded.
 */
MashCount
_mash_chunk_fade_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
			MashChunk * dest, MashCount dest_offset, MashCount count,
			int channelname, void * unused)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_fade, src, src_offset,
			   dest, dest_offset, count, NULL);
}

/*
 * _mash_chunk_interleave_2 (mash, src1, src1_offset, src2, src2_offset,
 *                         count, dest)
 *
 * Interleave data of 'src1' from stream index 'src1_offset' with that
 * of 'src2' from stream index 'src2_offset' for 'count' samples, storing
 * the resulting interleaved PCM data in 'dest'.
 *
 * Returns the count of sample frames interleaved.
 */
MashCount
_mash_chunk_interleave_2 (MASH * mash, MashChunk * src1, MashCount src1_offset,
			MashChunk * src2, MashCount src2_offset, MashCount count,
			int unused, void * dest)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_interleave_2, src1, src1_offset,
			   src2, src2_offset, count, dest);
}

/*
 * _mash_chunk_deinterleave_2 (mash, dest1, dest1_offset, dest2, dest2_offset,
 *                           count, src)
 *
 * Deinterleave data of 'src' and store the result in streams 'dest1' (from
 * stream index 'dest1_offset') and 'dest2' (from stream index
 * 'dest2_offset').
 *
 * Returns the number of sample frames deinterleaved.
 */
MashCount
_mash_chunk_deinterleave_2 (MASH * mash, MashChunk * dest1, MashCount dest1_offset,
			  MashChunk * dest2, MashCount dest2_offset, MashCount count,
			  int unused, void * src)
{
  return _mash_ppfunc_apply (mash, _mash_pcm_deinterleave_2, dest1, dest1_offset,
			   dest2, dest2_offset, count, src);
}

/* PPPFunc appliers */

/*
 * _mash_chunk_blend_inplace (src, src_offset, blend, blend_offset,
 *                          dest, dest_offset, count, unused)
 *
 * Blend data of 'src' into that of 'dest' by blend values given in
 * 'blend'.
 */
MashCount
_mash_chunk_blend_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
			 MashChunk * blend, MashCount blend_offset,
			 MashChunk * dest, MashCount dest_offset, MashCount count,
			 int channelname, void * unused)
{
  return _mash_pppfunc_apply (mash, _mash_pcm_blend, src, src_offset,
			    blend, blend_offset, dest, dest_offset,
			    count, NULL);
}
