/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashContext: MASH core context.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <string.h>

#define __MASH__
#include "mash.h"

static void
mash_context_destroy (MASH * mash)
{
  MashContext * ctx = mash->context;
  MashEnvironment * env = mash->environment;

  env->purging = 1;
  /* XXX: mash_destroy_list (mash, env->plugins); */
  /* XXX:  mash_destroy_list (mash, env->bases); */
  mash_free (ctx);
  mash_free (env);
}

static MASH *
mash_add_thread_context (MashContext * ctx, MashEnvironment * env)
{
  MASH * mash = mash_malloc (sizeof (struct _MashThreadContext));
  mash->context = ctx;
  mash->environment = env;
  env->refcount++;
  return mash;
}

MashContext *
_mash_context_copy (MASH * mash, MashContext * dest)
{
  MashContext * ctx = mash->context;

  if (dest == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  dest->samplerate = ctx->samplerate;
  dest->tempo = ctx->tempo;
  dest->mixlength = ctx->mixlength;
  dest->channels = cd_set_clone_keys (mash, ctx->channels);

  return dest;
}

/*
 * _mash_context_merge (mash, dest)
 *
 * Merges the context of mash into context 'dest'. Copies over the samplerate
 * and tempo, and expands the mixlength and channels if they are greater in
 * 'mash's context than in 'dest'.
 */
MashContext *
_mash_context_merge (MASH * mash, MashContext * dest)
{
  MashContext * ctx = mash->context;
  CDSet * s;
  CDScalar k;

  if (dest == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  dest->samplerate = ctx->samplerate;
  dest->tempo = ctx->tempo;

  if (ctx->mixlength > dest->mixlength)
    dest->mixlength = ctx->mixlength;

  for (s = ctx->channels; s; s = s->next) {
    k = cd_set_find (mash, dest->channels, s->key);
    if (k.s_pointer == NULL)
      dest->channels = cd_set_insert (mash, dest->channels, s->key,
				      CD_POINTER(NULL));
  }

  return dest;
}

/*
 * mash_init ()
 */
MASH *
mash_init (void)
{
  MASH * mash;
  MashEnvironment * env =
    (MashEnvironment *) mash_malloc (sizeof (struct _MashEnvironment));
  MashContext * ctx = (MashContext *) mash_malloc (sizeof (struct _MashContext));

  env->refcount = 0;
  env->plugins = cd_list_new (ctx);
  env->bases = cd_list_new (ctx);
  env->purging = FALSE;

  ctx->mixlength = MASH_DEFAULT_MIXLENGTH;
  ctx->samplerate = MASH_DEFAULT_SAMPLERATE;
  ctx->tempo = MASH_DEFAULT_TEMPO;

  mash = mash_add_thread_context (ctx, env);
  mash_channelset_initialise_defaults (mash);
  ctx->channels = MASH_MONO;

  mash_plugin_initialise_defaults (mash);

  return mash;
}

/*
 * mash_init_clone ()
 */
MASH *
mash_init_clone (MASH * mash)
{
  MASH * new_mash = mash_add_thread_context (mash->context, mash->environment);
  return new_mash;
}

void
mash_purge (MASH * mash)
{
  MashEnvironment * env = mash->environment;
  env->refcount--;
  if (env->refcount <= 0) {
    mash_context_destroy (mash);
  }
  mash_free (mash);
}

MashError
mash_set_error (MASH * mash, MashError error)
{
  MashError old = mash->last_error;
  mash->last_error = error;
  return old;
}

MashError
mash_last_error (MASH * mash)
{
  return mash->last_error;
}

MashCount
mash_set_mixlength (MASH * mash, MashCount mixlength)
{
  MashContext * ctx = mash->context;
  MashCount old = ctx->mixlength;
  ctx->mixlength = mixlength;
  return old;
}

MashCount
mash_get_mixlength (MASH * mash)
{
  MashContext * ctx = mash->context;
  return ctx->mixlength;
}

MashSamplerate
mash_set_samplerate (MASH * mash, MashSamplerate samplerate)
{
  MashContext * ctx = mash->context;
  MashSamplerate old = ctx->samplerate;
  ctx->samplerate = samplerate;
  return old;
}

MashSamplerate
mash_get_samplerate (MASH * mash)
{
  MashContext * ctx = mash->context;
  return ctx->samplerate;
}

MashTempo
mash_set_tempo (MASH * mash, MashTempo tempo)
{
  MashContext * ctx = mash->context;
  MashTempo old = ctx->tempo;
  ctx->tempo = tempo;
  return old;
}

MashTempo
mash_get_tempo (MASH * mash)
{
  MashContext * ctx = mash->context;
  return ctx->tempo;
}

CDSet *
mash_set_channels (MASH * mash,  CDSet * channels)
{
  MashContext * ctx = mash->context;
  CDSet * old = ctx->channels;
  ctx->channels = cd_set_clone_keys (mash, channels);
  return old;
}

CDSet *
mash_get_channels (MASH * mash)
{
  MashContext * ctx = mash->context;
  return ctx->channels;
}

MASH *
_mash_register_plugin (MASH * mash, MashPlugin * plugin)
{
  MashEnvironment * env = mash->environment;
  mash_dprintf ("[_mash_register_plugin] REGISTERING %s\n",
	      plugin->metatext ? plugin->metatext->identifier : "(\?\?\?)");
  env->plugins = cd_list_append (mash, env->plugins, CD_POINTER(plugin));
  return mash;
}

MASH *
_mash_unregister_plugin (MASH * mash, MashPlugin * plugin)
{
  MashEnvironment * env = mash->environment;
  if (env->purging) return mash;

  env->plugins = cd_list_remove (mash, env->plugins, CD_TYPE_POINTER,
				 CD_POINTER(plugin));
  return mash;
}

MASH *
_mash_register_base (MASH * mash, MashBase * base)
{
  MashEnvironment * env = mash->environment;
  env->bases = cd_list_append (mash, env->bases, CD_POINTER(base));
  return mash;
}

MASH *
_mash_unregister_base (MASH * mash, MashBase * base)
{
  MashEnvironment * env = mash->environment;
  if (env->purging) return mash;

  env->bases = cd_list_remove (mash, env->bases, CD_TYPE_POINTER,
			       CD_POINTER(base));
  return mash;
}

static int
plugin_id_eq (MASH * mash, MashPlugin * plugin, char * identifier)
{
  if (plugin == MashNone) return FALSE;
  if (plugin->metatext == MashNone) return FALSE;
  return !(strcmp(plugin->metatext->identifier, identifier));
}

MashPlugin *
mash_find_plugin (MASH * mash, char * identifier)
{
  MashEnvironment * env = mash->environment;
  CDList * l = cd_list_find_first (mash, env->plugins, CD_TYPE_POINTER,
				   (CDCmpFunc)plugin_id_eq,
				   CD_POINTER(identifier));

  if (l == MashNone) return MashNone;
  return (MashPlugin *)l->data.s_pointer;
}
