/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashDeck: A high level audio mixing abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A deck contains a number of tracks which are mixed in parallel.
 *
 * Invariants
 * ----------
 *
 * Decks are independent entities. The creation of a new deck does not
 * depend on the existence of any other bases. Any tracks (and, by
 * extension, layers and sounds) added to a deck become part of that
 * deck and are destroyed when the deck is destroyed.
 *
 */

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: optimise on change of nr. tracks */
static MashDeck * mash_deck_optimise (MASH * mash, MashDeck * deck);

static MashDeck *
mash_deck_replace_mixstream (MASH * mash, MashDeck * deck)
{
  MashCount mixlength = _mash_base_get_mixlength (mash, deck);

  if (deck->_mixstream != MashNone)
    mash_destroy (mash, (MashBase *)deck->_mixstream);

  deck->_mixstream = mash_stream_new_contiguous (mash, mixlength);

  return deck;
}

MashBase *
mash_deck_init (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  deck->tracks = cd_list_new (mash);
  deck->_mixstream = MashNone;
  mash_deck_replace_mixstream (mash, deck);
  mash_deck_optimise (mash, deck);
  return (MashBase *)deck;
}

MashDeck *
mash_deck_new (MASH * mash)
{
  MashDeck * deck =
    (MashDeck *) mash_base_new_subclass (mash, sizeof (struct _MashDeck));
  return (MashDeck *)mash_deck_init (mash, (MashBase *)deck);
}

MashBase *
mash_deck_clone (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  MashDeck * new_deck = mash_deck_new (mash);
  new_deck->tracks = cd_list_clone (mash, deck->tracks,
				    (CDCloneFunc)mash_track_clone);
  return (MashBase *)new_deck;
}

static int
mash_deck_destroy (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  mash_destroy_list (mash, deck->tracks);
  mash_destroy (mash, (MashBase *)deck->_mixstream);
  mash_free (deck);
  return 0;
}

static int
mash_deck_ready (MASH * mash, MashBase * base)
{
  return (mash_base_encompasses_mixlength (mash, base) &&
	  mash_base_encompasses_channels (mash, base));
}

static MashBase *
mash_deck_prepare (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  mash_deck_replace_mixstream (mash, deck);
  return base;
}

MashTrack *
_mash_deck_add_track (MASH * mash, MashDeck * deck, MashTrack * track)
{
  deck->tracks = cd_list_prepend (mash, deck->tracks, CD_POINTER(track));
  mash_deck_optimise (mash, deck);
  return track;
}

MashTrack *
_mash_deck_remove_track (MASH * mash, MashDeck * deck, MashTrack * track)
{
  deck->tracks = cd_list_remove (mash, deck->tracks, CD_TYPE_POINTER,
				 CD_POINTER(track));
  mash_deck_optimise (mash, deck);
  return track;
}

int
mash_deck_nr_tracks (MASH * mash, MashDeck * deck)
{
  return cd_list_length (mash, deck->tracks);
}

static MashCount
mash_deck_length (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  MashCount length, maxlength = 0;
  CDList * l;
  MashBase * track;
  
  for (l = deck->tracks; l; l = l->next) {
    track = (MashBase *)l->data.s_pointer;
    length = mash_length (mash, track);
    mash_dprintf ("[mash_deck_length] found track %p length %ld\n", track, length);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

CDList *
mash_deck_get_tracks (MASH * mash, MashDeck * deck)
{
  return cd_list_copy (mash, deck->tracks);
}

#if 0
static MashCount
mash_deck_process (MASH mash, MashBase base, MashCount count, MashStream input,
		 MashStream output)
{
  MashDeck deck = (MashDeck)base;
  CDList tl, ml;
  MashTrack track;
  MashStream mixstream;
  MashCount processed = 0, remaining = count, n, minn;
  MashCount input_offset;

  mash_dprintf ("PROCESS DECK (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, mash_tell (mash, base));

  while (remaining > 0) {
    /* Get mixlength from each track */
    minn = deck->mixlength;
    input_offset = mash_tell (mash, (MashBase)input);
    for (tl = deck->tracks, ml = deck->mixstreams; tl && ml;
         tl = tl->next, ml = ml->next) {
      track = (MashTrack)tl->data;
      mixstream = (MashStream)ml->data;
      mash_seek (mash, (MashBase)input, input_offset);
      n = mash_process (mash, (MashBase)track, MIN (remaining, deck->mixlength),
		      input, mixstream);
      minn = MIN (minn, n);
    }


    if (output != MashNone) {
      /* mix tracks together to the output */
      n = mash_streams_mix (mash, deck->mixstreams, output, minn);
    } else {
      /* don't need to create output */
      n = minn;
    }

    processed += n;
    remaining -= n;
  }

  mash_dprintf ("[mash_deck_process] processed %ld\n", processed);

  return processed;
}
#endif

static MashCount
mash_deck_process (MASH * mash, MashBase * base, MashCount count,
		 MashStream * input, MashStream * output)
{
  MashDeck * deck = (MashDeck *)base;
  CDList * l;
  MashTrack * track;
  MashCount remaining = count, processed = 0, n;
  MashCount current_offset = mash_tell (mash, base);
  MashCount input_offset = mash_tell (mash, (MashBase *)input);
  MashCount output_offset = mash_tell (mash, (MashBase *)output);
  MashCount mixlength = _mash_base_get_mixlength (mash, deck);
  MashStream * mixstream = deck->_mixstream;

  mash_dprintf ("PROCESS DECK (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, current_offset);

  while (remaining > 0) {

    l = deck->tracks;
    track = (MashTrack *)l->data.s_pointer;

    n = MIN (remaining, mixlength);

    n = mash_process (mash, (MashBase *)track, n, input, output);
    mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
    n = mash_stream_gain (mash, output, n, track->gain);

    for (l = l->next; l; l = l->next) {
      track = (MashTrack *)l->data.s_pointer;
      
      mash_seek (mash, (MashBase *)input, input_offset, SEEK_SET);
      mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
      n = mash_process (mash, (MashBase *)track, n, input, mixstream);
      
      mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
      n = mash_stream_gain (mash, mixstream, n, track->gain);
      
      mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
      mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
      n = mash_stream_mix (mash, mixstream, output, n);
    }

    processed += n;
    remaining -= n;
  }

  mash_dprintf ("[mash_deck_process] processed %ld\n", processed);

  return processed;
}

static MashCount
mash_deck_twotrack_process (MASH * mash, MashBase * base, MashCount count,
			  MashStream * input, MashStream * output)
{
  MashDeck * deck = (MashDeck *)base;
  CDList * l;
  MashTrack * track1, * track2;
  MashCount n;
  MashCount current_offset = mash_tell (mash, base);
  MashCount input_offset = mash_tell (mash, (MashBase *)input);
  MashCount output_offset = mash_tell (mash, (MashBase *)output);
  MashStream * mixstream = deck->_mixstream;

  mash_dprintf ("PROCESS DECK [twotrack] (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, current_offset);

  l = deck->tracks;
  track1 = (MashTrack *)l->data.s_pointer;

  l = l->next;
  track2 = (MashTrack *)l->data.s_pointer;

  n = mash_process (mash, (MashBase *)track1, count, input, output);

  mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
  n = mash_stream_gain (mash, output, n, track1->gain);

  mash_seek (mash, (MashBase *)input, input_offset, SEEK_SET);
  mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
  n = mash_process (mash, (MashBase *)track2, n, input, mixstream);

  mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
  n = mash_stream_gain (mash, mixstream, n, track2->gain);

  mash_seek (mash, (MashBase *)mixstream, 0, SEEK_SET);
  mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
  n = mash_stream_mix (mash, mixstream, output, n);

  mash_dprintf ("[mash_deck_twotrack_process] processed %ld\n", n);

  return n;
}

static MashCount
mash_deck_onetrack_process (MASH * mash, MashBase * base, MashCount count,
			  MashStream * input, MashStream * output)
{
  MashDeck * deck = (MashDeck *)base;
  MashTrack * track = (MashTrack *)deck->tracks->data.s_pointer;
  MashCount n;

  mash_dprintf ("PROCESS DECK [onetrack] (%p, +%ld, %p -> %p) @ %ld\n",
	      deck, count, input, output, mash_tell (mash, base));

  n = mash_process (mash, (MashBase *)track, count, input, output);

  mash_dprintf ("*** deck @ %ld\ttrack @ %ld\n", mash_tell (mash, base),
              mash_tell (mash, (MashBase *)track));

  mash_dprintf ("[mash_deck_onetrack_process] processed %ld\n", n);

  return n;
}

static MashCount
mash_deck_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashDeck * deck = (MashDeck *)base;
  CDList * l;
  MashTrack * track;

  for (l = deck->tracks; l; l = l->next) {
    track = (MashTrack *)l->data.s_pointer;
    mash_seek (mash, (MashBase *)track, offset, SEEK_SET);
  }

  return offset;
}

static int
mash_deck_flush (MASH * mash, MashBase * base)
{
  MashDeck * deck = (MashDeck *)base;
  CDList * l;
  MashTrack * track;

  for (l = deck->tracks; l; l = l->next) {
    track = (MashTrack *)l->data.s_pointer;
    mash_flush (mash, (MashBase *)track);
  }

  return 0;
}

static struct _MashMethods _mash_deck_empty_methods = {
  mash_deck_clone,   /* clone */
  mash_deck_destroy, /* destroy */
  mash_deck_ready,   /* ready */
  mash_deck_prepare, /* preapre */
  mash_null_process, /* process */
  mash_null_length,  /* length */
  NULL,            /* seek */
  NULL,            /* flush */
};

static struct _MashMethods _mash_deck_methods = {
  mash_deck_clone,   /* clone */
  mash_deck_destroy, /* destroy */
  mash_deck_ready,   /* ready */
  mash_deck_prepare, /* preapre */
  mash_deck_process, /* process */
  mash_deck_length,  /* length */
  mash_deck_seek,    /* seek */
  mash_deck_flush,   /* flush */
};

static struct _MashMethods _mash_deck_onetrack_methods = {
  mash_deck_clone,            /* clone */
  mash_deck_destroy,          /* destroy */
  mash_deck_ready,            /* ready */
  mash_deck_prepare,          /* preapre */
  mash_deck_onetrack_process, /* process */
  mash_deck_length,           /* length */
  mash_deck_seek,             /* seek */
  mash_deck_flush,   /* flush */
};

static struct _MashMethods _mash_deck_twotrack_methods = {
  mash_deck_clone,            /* clone */
  mash_deck_destroy,          /* destroy */
  mash_deck_ready,            /* ready */
  mash_deck_prepare,          /* preapre */
  mash_deck_twotrack_process, /* process */
  mash_deck_length,           /* length */
  mash_deck_seek,             /* seek */
  mash_deck_flush,   /* flush */
};

static MashDeck *
mash_deck_optimise (MASH * mash, MashDeck * deck)
{
  int nr_tracks = cd_list_length (mash, deck->tracks);

  switch (nr_tracks) {
  case 0: _mash_set_methods (mash, deck, &_mash_deck_empty_methods); break;
  case 1: _mash_set_methods (mash, deck, &_mash_deck_onetrack_methods); break;
  case 2: _mash_set_methods (mash, deck, &_mash_deck_twotrack_methods); break;
  default: _mash_set_methods (mash, deck, &_mash_deck_methods); break;
  }

  return deck;
}
