/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashEnvelope: Generic mathy envelopes for mash.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: optimise when adding and removing points
 * or changing envelope type */
static MashEnvelope * mash_envelope_optimise (MASH * mash, MashEnvelope * envelope);

static void
mash_envelope_debug (MASH * mash, MashEnvelope * envelope)
{
#ifdef DEBUG
  CDList * l;
  MashPoint * point;

  printf ("envelope %p\n", envelope);
  printf ("envelope->points: %p\n", envelope->points);

  /*  printf (" has %d points\n", cd_list_length (mash, envelope->points));*/
  for (l = envelope->points; l; l = l->next) {
    point = (MashPoint *)l->data.s_pointer;
    switch (envelope->timetype) {
    case MASH_TIME_SAMPLES:
      printf ("%ld samples, %f\n", point->time.samples, point->value);
      break;
    case MASH_TIME_SECONDS:
      printf ("%f seconds, %f\n", point->time.seconds, point->value);
      break;
    case MASH_TIME_BEAT24S:
      printf ("%d beat24s, %f\n", point->time.beat24s, point->value);
      break;
    default:
      printf ("*** unknown envelope->timetype ***\n");
      break;
    }
  } 
#endif
}

static MashPoint *
mash_point_new (MashTime time, MashPCM value)
{
  MashPoint * point = mash_malloc (sizeof (struct _MashPoint));
  point->time = time;
  point->value = value;
  return point;
}

static MashPoint *
mash_point_clone (MASH * mash, MashPoint * point)
{
  return mash_point_new (point->time, point->value);
}

static int
mash_point_later_X (MASH * mash, MashPoint * p1, MashPoint * p2)
{
  return _mash_time_gt(MASH_TIME_SAMPLES, p1->time, p2->time);
}

static int
mash_point_later_S (MASH * mash, MashPoint * p1, MashPoint * p2)
{
  return _mash_time_gt(MASH_TIME_SECONDS, p1->time, p2->time);
}

static int
mash_point_later_B (MASH * mash, MashPoint * p1, MashPoint * p2)
{
  return _mash_time_gt(MASH_TIME_BEAT24S, p1->time, p2->time);
}

MashBase *
mash_envelope_init (MASH * mash, MashBase * base)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  envelope->type = MASH_ENVELOPE_LINEAR;
  envelope->points = cd_list_new (mash);
  mash_envelope_optimise (mash, envelope);
  return (MashBase *)envelope;
}

MashEnvelope *
mash_envelope_new (MASH * mash, MashEnvelopeType type)
{
  MashEnvelope * envelope =
    (MashEnvelope *)mash_base_new_subclass (mash, sizeof (struct _MashEnvelope));
  mash_envelope_init (mash, (MashBase *)envelope);
  envelope->type = type;
  mash_envelope_debug (mash, envelope);
  return envelope;
}

MashBase *
mash_envelope_clone (MASH * mash, MashBase * base)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  MashEnvelope * new_envelope = mash_envelope_new (mash, envelope->type);
  new_envelope->points = cd_list_clone (mash, envelope->points,
					(CDCloneFunc)mash_point_clone);
  mash_envelope_optimise (mash, new_envelope);
  return (MashBase *)new_envelope;
}

static int
mash_envelope_destroy (MASH * mash, MashBase * base)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  cd_list_free_all (mash, envelope->points);
  mash_free (envelope);
  return 0;
}

MashEnvelopeType
mash_envelope_set_type (MASH * mash, MashEnvelope * envelope, MashEnvelopeType type)
{
  MashEnvelopeType old = envelope->type;
  envelope->type = type;
  mash_envelope_optimise (mash, envelope);
  return old;
}

MashEnvelopeType
mash_envelope_get_type (MASH * mash, MashEnvelope * envelope)
{
  return envelope->type;
}

MashTimeType
mash_envelope_set_timetype (MASH * mash, MashEnvelope * envelope,
			  MashTimeType timetype)
{
  MashTimeType old = envelope->timetype;
  envelope->timetype = timetype;
  return old;
}

MashTimeType
mash_envelope_get_timetype (MASH * mash, MashEnvelope * envelope)
{
  return envelope->timetype;
}

MashTime
mash_envelope_get_duration (MASH * mash, MashEnvelope * envelope)
{
  CDList * l = cd_list_last_item (mash, envelope->points);
  MashPoint * p;

  if (l == MashNone) return _mash_time_zero (envelope->timetype);

  p = (MashPoint *)l->data.s_pointer;
  if (p == MashNone) return _mash_time_zero (envelope->timetype);

  return p->time;
}

MashPoint *
mash_envelope_add_point (MASH * mash, MashEnvelope * envelope, MashTime time,
		       MashPCM value)
{
  MashPoint * point = mash_point_new (time, value);
  switch (envelope->timetype) {
  case MASH_TIME_SAMPLES:
    envelope->points = cd_list_insert (mash, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)mash_point_later_X);
    break;
  case MASH_TIME_SECONDS:
    envelope->points = cd_list_insert (mash, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)mash_point_later_S);
    break;
  case MASH_TIME_BEAT24S:
    envelope->points = cd_list_insert (mash, envelope->points,
				       CD_TYPE_POINTER, CD_POINTER(point),
				       (CDCmpFunc)mash_point_later_B);
    break;
  default: /* uncommon, we should hope */
    mash_free (point); point = MashNone;
    break;
  }
  mash_envelope_debug (mash, envelope);
  mash_envelope_optimise (mash, envelope);
  return point;
}

MashEnvelope *
mash_envelope_remove_point (MASH * mash, MashEnvelope * envelope, MashPoint * point)
{
  envelope->points = cd_list_remove (mash, envelope->points,
				     CD_TYPE_POINTER, CD_POINTER(point));
  mash_free (point);
  mash_envelope_debug (mash, envelope);
  mash_envelope_optimise (mash, envelope);
  return envelope;
}

MashEnvelope *
mash_envelope_scale (MASH * mash, MashEnvelope * envelope, MashPCM gain)
{
  CDList * l;
  MashPoint * p;

  for (l = envelope->points; l; l = l->next) {
    p = (MashPoint *)l->data.s_pointer;
    p->value *= gain;
  }

  return envelope;
}

MashEnvelope *
mash_envelope_shift (MASH * mash, MashEnvelope * envelope, MashTime delta)
{
  CDList * l;
  MashPoint * p;

  for (l = envelope->points; l; l = l->next) {
    p = (MashPoint *)l->data.s_pointer;
    p->time = _mash_time_add (envelope->timetype, p->time, delta);
  }

  return envelope;
}

static CDList *
mash_envelope_point_item_before (MASH * mash, MashEnvelope * envelope,
			       MashCount offset)
{
  CDList * l, * lp = MashNone;
  MashPoint * point;
  MashTime ptime;

  for (l = envelope->points; l; l = l->next) {
    point = (MashPoint *)l->data.s_pointer;
    ptime = mash_time_convert (mash, point->time, envelope->timetype,
			     MASH_TIME_SAMPLES);
    if (ptime.samples > offset) break;
    lp = l;
  }
  return lp;
}

static MashCount
mash_envelope_constant_write_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
				  MashCount count, int channelname, void * data)
{
  MashEnvelope * envelope = (MashEnvelope *)data;
  MashPoint * point;
  MashPCM value;
  MashPCM * d;
  MashCount n;

  point = (MashPoint *)envelope->points->data.s_pointer;
  value = point->value;
  d = &chunk->data[offset];

  n = _mash_pcm_set (d, value, count);
  envelope->_current_offset += n;
  return n;
}

#if 0
static MashCount
mash_envelope_spline_write_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
				MashCount count, int channelname, void * data)
{
  MashEnvelope * envelope = (MashEnvelope *)data;
  /* XXX: Implement ;) */
  return -1;
}
#endif

/* An MashChunkFunc for creating envelope data */
static MashCount
mash_envelope_linear_write_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
				MashCount count, int channelname, void * data)
{
  MashEnvelope * envelope = (MashEnvelope *)data;
  MashCount remaining = count, written = 0;
  MashCount pos = envelope->_current_offset;
  CDList * l, * nl;
  MashPoint * point, * next_point;
  MashCount px, npx, n;
  MashTime t;
  MashPCM py, npy, gradient;
  MashPCM * d;

  mash_dprintf ("[mash_envelope_linear_write_chunk] (%ld, +%ld) @ %ld\n",
	  offset, count, pos);


  l = envelope->_current_point_item;

  if (l == MashNone) {/* No points before start */
    l = envelope->points;
    if (l == MashNone) {/* No points at all */
      n = _mash_chunk_clear_region (mash, chunk, offset, count, 0, NULL);
      envelope->_current_offset += n;
      return n;
    }
  }

  nl = l->next;
  if (nl == MashNone) {
    /* if the last point was before offset, and there were
     * more points, set l to the second last and nl to the last */
    nl = l; l = l->prev;
    if (l == MashNone) {/* Constant envelope (one point) */
      return mash_envelope_constant_write_chunk (mash, chunk, offset, count,
					       channelname, envelope);
    }
  }

  point = (MashPoint *)l->data.s_pointer;
  t = mash_time_convert (mash, point->time, envelope->timetype, MASH_TIME_SAMPLES);
  px = t.samples;
  py = point->value;
  
  next_point = (MashPoint *)nl->data.s_pointer;
  t = mash_time_convert (mash, next_point->time, envelope->timetype,
		       MASH_TIME_SAMPLES);
  npx = t.samples;
  npy = next_point->value;

  while (remaining > 0) {
    if (nl->next == MashNone) {
      /* These are the last two points, so fill out with this gradient */
      n = remaining;
    } else {
      n = MIN (remaining, npx - pos);
    }
    gradient = (npy - py) / (MashPCM)(npx - px);
    
    d = &chunk->data[offset];
    /*    _mash_pcm_write_linear (d, px - chunk->start_index, py, gradient, n);*/
    n = _mash_pcm_write_linear (d, px, py, npx, npy, pos, n);
    
    remaining -= n;
    written += n;
    pos += n;
    offset += n;
    
    if (remaining > 0) {
      l = nl; point = next_point; px = npx; py = npy;
      
      nl = nl->next;
      next_point = (MashPoint *)nl->data.s_pointer;
      t = mash_time_convert (mash, next_point->time, envelope->timetype,
			   MASH_TIME_SAMPLES);
      npx = t.samples;
      npy = next_point->value;
    }
  }

  envelope->_current_point_item = l;
  envelope->_current_offset = pos;

  return written;
}

static MashCount
mash_envelope_constant_process (MASH * mash, MashBase * base, MashCount count,
			      MashStream * input, MashStream * output)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  return mash_stream_chunkfuncify (mash, output, count,
				 mash_envelope_constant_write_chunk, envelope);
}

static MashCount
mash_envelope_spline_process (MASH * mash, MashBase * base, MashCount count,
			    MashStream * input, MashStream * output)
{
  return -1;
}

static MashCount
mash_envelope_linear_process (MASH * mash, MashBase * base, MashCount count,
			    MashStream * input, MashStream * output)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  return mash_stream_chunkfuncify (mash, output, count,
				 mash_envelope_linear_write_chunk, envelope);
}

static MashCount
mash_envelope_process (MASH * mash, MashBase * base, MashCount count,
		     MashStream * input, MashStream * output)
{
  MashEnvelope * envelope = (MashEnvelope *)base;

  switch (envelope->type) {
  case MASH_ENVELOPE_LINEAR:
    return mash_envelope_linear_process (mash, base, count, input, output);
    break;
  case MASH_ENVELOPE_SPLINE:
    return mash_envelope_spline_process (mash, base, count, input, output);
    break;
  default:
    break;
  }
  mash_set_error (mash, MASH_ERROR_NOOP);
  return 0;
}

static MashCount
mash_envelope_length (MASH * mash, MashBase * base)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  MashTime duration = mash_envelope_get_duration (mash, envelope);
  MashTime t = mash_time_convert (mash, duration, envelope->timetype,
			      MASH_TIME_SAMPLES);
  return t.samples;
}

static MashCount
mash_envelope_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashEnvelope * envelope = (MashEnvelope *)base;
  envelope->_current_point_item =
    mash_envelope_point_item_before (mash, envelope, offset);
  envelope->_current_offset = offset;
  return offset;
}

static struct _MashMethods _mash_envelope_empty_methods = {
  mash_envelope_clone,
  mash_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_null_process,
  mash_null_length,
  mash_envelope_seek,
  NULL, /* flush */
};

static struct _MashMethods _mash_envelope_constant_methods = {
  mash_envelope_clone,
  mash_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_envelope_constant_process,
  mash_envelope_length,
  mash_envelope_seek,
  NULL, /* flush */
};

static struct _MashMethods _mash_envelope_linear_methods = {
  mash_envelope_clone,
  mash_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_envelope_linear_process,
  mash_envelope_length,
  mash_envelope_seek,
  NULL, /* flush */
};

static struct _MashMethods _mash_envelope_spline_methods = {
  mash_envelope_clone,
  mash_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_envelope_spline_process,
  mash_envelope_length,
  mash_envelope_seek,
  NULL, /* flush */
};

static struct _MashMethods _mash_envelope_methods = {
  mash_envelope_clone,
  mash_envelope_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_envelope_process,
  mash_envelope_length,
  mash_envelope_seek,
  NULL, /* flush */
};

static MashEnvelope *
mash_envelope_optimise (MASH * mash, MashEnvelope * envelope)
{
  if (cd_list_is_empty (mash, envelope->points)) {
    _mash_set_methods (mash, envelope, &_mash_envelope_empty_methods);
  } else if (cd_list_is_singleton (mash, envelope->points)) {
    _mash_set_methods (mash, envelope, &_mash_envelope_constant_methods);
  } else {
    switch (envelope->type) {
    case MASH_ENVELOPE_LINEAR:
      _mash_set_methods (mash, envelope, &_mash_envelope_linear_methods); break;
    case MASH_ENVELOPE_SPLINE:
      _mash_set_methods (mash, envelope, &_mash_envelope_spline_methods); break;
    default:
      _mash_set_methods (mash, envelope, &_mash_envelope_methods); break;
    }
  }

  return envelope;
}
