/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashGain: a gain filter
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdlib.h>

#define __MASH_PLUGIN__
#include "mash.h"

#define GAIN_ENVELOPE_KEY 1

typedef struct _MashGain MashGain;

struct _MashGain {
  MashStream * _gain_envstream;
};

/* Optimisation dependencies: none */
static MashBase * mash_gain_optimise (MASH * mash, MashBase * gain);

static void
mash_gain_replace_mixstream (MASH * mash, MashBase * gain)
{
  MashCount mixlength = mash_base_get_mixlength (mash, gain);
  MashGain * gi = (MashGain *)mash_base_get_instance_data (mash, gain);

  if (gi->_gain_envstream != MashNone)
    mash_destroy (mash, (MashBase *)gi->_gain_envstream);

  gi->_gain_envstream =
    mash_stream_new_contiguous (mash, mixlength);
}

static MashBase *
mash_gain_init (MASH * mash, MashBase * base, CDSet * parameters)
{
  mash_base_set_instance_data (mash, base,
			       calloc (1, sizeof (struct _MashGain)));
  mash_gain_replace_mixstream (mash, base);
  mash_gain_optimise (mash, base);
  return base;
}

static MashBase *
mash_gain_clone (MASH * mash, MashBase * base)
{
  MashBase * new_gain = mash_base_new (mash);
  mash_gain_init (mash, new_gain, CD_EMPTY_SET);
  mash_gain_optimise (mash, new_gain);
  return (MashBase *)new_gain;
}

static int
mash_gain_destroy (MASH * mash, MashBase * base)
{
  free (mash_base_get_instance_data (mash, base));
  free (base);
  return 0;
}

static int
mash_gain_ready (MASH * mash, MashBase * base)
{
  return (mash_base_encompasses_mixlength (mash, base) &&
	  mash_base_encompasses_channels (mash, base));
}

static MashBase *
mash_gain_prepare (MASH * mash, MashBase * base)
{
  mash_gain_replace_mixstream (mash, base);
  return base;
}

static MashCount
mash_gain_process (MASH * mash, MashBase * base, MashCount count,
		 MashStream * input, MashStream * output)
{
  MashCount remaining = count, processed = 0, n;
  MashCount output_offset;
  MashCount mixlength = mash_base_get_mixlength (mash, base);
  MashBase * gain_envelope;
  MashGain * gi = mash_base_get_instance_data (mash, base);

  mash_dprintf ("PROCESS GAIN (%p, +%ld) @ %ld\n", base, count,
	      mash_tell (mash, base));

  gain_envelope =
    (MashBase *) (mash_get_parameter (mash, base, GAIN_ENVELOPE_KEY)).s_pointer;

  if (gain_envelope == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOOP);
    return -1;
  }

  while (remaining > 0) {
    n = MIN (remaining, mixlength);

    output_offset = mash_tell (mash, (MashBase *)output);
    n = mash_stream_copy (mash, input, output, n);

    mash_seek (mash, (MashBase *)gi->_gain_envstream, 0, SEEK_SET);
    n = mash_process (mash, gain_envelope, n, MashNone,
		    gi->_gain_envstream);

    mash_seek (mash, (MashBase *)gi->_gain_envstream, 0, SEEK_SET);
    mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
    n = mash_stream_mult (mash, gi->_gain_envstream, output, n);

    remaining -= n;
    processed += n;
  }

  mash_dprintf ("[mash_gain_process] processed %ld\n", processed);

  return processed;
}

static MashCount
mash_gain_length (MASH * mash, MashBase * base)
{
  MashBase * gain_envelope =
    (MashBase *) (mash_get_parameter (mash, base, GAIN_ENVELOPE_KEY)).s_pointer;

  if (gain_envelope == MashNone) {
    return MASH_COUNT_INFINITE;
  }

  return mash_length (mash, gain_envelope);
}

static MashCount
mash_gain_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashBase * gain_envelope =
    (MashBase *) (mash_get_parameter (mash, base, GAIN_ENVELOPE_KEY)).s_pointer;

  return mash_seek (mash, (MashBase *)gain_envelope, offset, SEEK_SET);
}

static struct _MashMethods _mash_gain_methods = {
  mash_gain_clone,
  mash_gain_destroy,
  mash_gain_ready,
  mash_gain_prepare,
  mash_gain_process,
  mash_gain_length,
  mash_gain_seek,
};

static MashBase *
mash_gain_optimise (MASH * mash, MashBase * gain)
{
  mash_base_set_methods (mash, gain, &_mash_gain_methods);
  return gain;
}

static struct _MashParameterScheme gain_envelope_scheme = {
  "Gain envelope",
  "An envelope to control the amplitude",
  MASH_TYPE_BASE,
  MASH_CONSTRAINT_TYPE_NONE,
  MASH_CONSTRAINT_EMPTY,
  MASH_HINT_DEFAULT,
};

static struct _MashMetaText gain_metatext = {
  "builtin::gain",
  "Processors::Gain Adjustment",
  "Adjusts the gain of its input",
  "Copyright (C) 2001 CSIRO Australia",
  "http://www.metadecks.org/mash/plugins/gain.html",
  MASH_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _MashPlugin gain_plugin = {
  &gain_metatext,
  MASH_FLAGS_NONE,
  CD_EMPTY_SET, /* new scheme */
  mash_gain_init,
  CD_EMPTY_SET, /* process_scheme */
  NULL, /* suggests */
  NULL, /* plugin_data */
};

/* module init function */
CDList *
__gain_init (MASH * mash)
{
  CDList * plugins = cd_list_new (mash);

  gain_plugin.process_scheme =
    cd_set_insert (mash, gain_plugin.process_scheme, GAIN_ENVELOPE_KEY,
		   CD_POINTER(&gain_envelope_scheme));

  plugins = cd_list_prepend (mash, plugins, CD_POINTER(&gain_plugin));

  return plugins;
}
