/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashLayer: A sound sequence abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A layer is contained within a track. It consists of a sequence of
 * sounds. A layer has a time type (MashTimeType) with which its sounds
 * are indexed.
 * XXX: Overlapping sounds in a layer should be mixed together.
 *
 * Invariants
 * ----------
 *
 * A layer must be contained in a track.
 *
 */

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: none */
static MashLayer * mash_layer_optimise (MASH * mash, MashLayer * layer);

/* Coherency dependencies: ensure coherency on addition+removal of sounds */
static MashLayer * mash_layer_ensure_coherency (MASH * mash, MashLayer * layer);

void
mash_layer_debug (MASH * mash, MashLayer * layer)
{
#ifdef DEBUG
  CDList * l;
  MashSound * s;

  mash_dprintf ("Layer (0x%p): ", layer);
  if (layer == MashNone) return;

  for (l = layer->sounds; l; l = l->next) {
    s = (MashSound *)l->data.s_pointer;
    /* XXX: assumes samples */
    mash_dprintf ("[0x%p: %ld, +%ld] ", s, s->start_time.samples,
		s->duration.samples);
  }
  mash_dprintf ("\n");
#endif
}

static MashBase *
mash_layer_init (MASH * mash, MashBase * base)
{
  MashLayer * layer = (MashLayer *)base;
  layer->timetype = MASH_TIME_SAMPLES;
  layer->sounds = cd_list_new (mash);
  /*  layer->_current_time = _mash_time_zero (layer->timetype);*/
  layer->_current_sound_item = MashNone;
  layer->_current_tempo = mash_get_tempo (mash);
  layer->_current_offset = 0;
  mash_layer_optimise (mash, layer);
  return (MashBase *)layer;
}

static MashBase *
mash_new_layer (MASH * mash)
{
  return (MashBase *) mash_base_new_subclass (mash, sizeof (struct _MashLayer));
}

MashBase *
mash_layer_clone (MASH * mash, MashBase * base)
{
  MashLayer * layer = (MashLayer *)base;
  MashLayer * new_layer = (MashLayer *)mash_new_layer (mash);
  MashCount offset = mash_tell (mash, base);

  new_layer->timetype = layer->timetype;
  new_layer->sounds =
    cd_list_clone_with (mash, layer->sounds,
			(CDCloneWithFunc)mash_sound_clone_with_layer,
			new_layer);
  mash_seek (mash, (MashBase *)new_layer, offset, SEEK_SET);

  new_layer->track = layer->track;
  _mash_track_add_layer_above (mash, layer->track, new_layer, layer);

  return (MashBase *)new_layer;
}

static int
mash_layer_destroy (MASH * mash, MashBase * base)
{
  MashLayer * layer = (MashLayer *)base;
  if (layer->track)
    _mash_track_remove_layer (mash, layer->track, layer);
  mash_destroy_list (mash, layer->sounds);
  mash_free (layer);
  return 0;
}

MashLayer *
mash_add_layer_ontop (MASH * mash, MashTrack * track, MashTimeType timetype)
{
  MashLayer * layer = (MashLayer *)mash_new_layer (mash);

  layer->track = track;
  _mash_track_add_layer_above (mash, track, layer, MashNone);
  mash_layer_init (mash, (MashBase *)layer);
  layer->timetype = timetype;

  return layer;
}

MashLayer *
mash_add_layer_above (MASH * mash, MashLayer * above, MashTimeType timetype)
{
  MashLayer * layer;
  if (above == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  layer = (MashLayer *)mash_new_layer (mash);
  layer->track = above->track;
  _mash_track_add_layer_above (mash, above->track, layer, above);
  mash_layer_init (mash, (MashBase *)layer);
  layer->timetype = timetype;

  return layer;
}

MashLayer *
mash_move_layer_ontop (MASH * mash, MashLayer * layer, MashTrack * track)
{
  if (layer->track)
    _mash_track_remove_layer (mash, layer->track, layer);

  if (track)
    _mash_track_add_layer_above (mash, track, layer, MashNone);

  return layer;
}

MashLayer *
mash_move_layer_above (MASH * mash, MashLayer * layer, MashLayer * above)
{
  if (layer->track)
    _mash_track_remove_layer (mash, layer->track, layer);

  if (above->track)
    _mash_track_add_layer_above (mash, above->track, layer, above);

  return layer;
}

MashLayer *
mash_raise_layer (MASH * mash, MashLayer * layer)
{
  MashLayer * above;

  if (layer->track) {
    above = _mash_track_get_layer_above (mash, layer->track, layer);
    mash_move_layer_above (mash, layer, above);
  }

  return layer;
}

MashLayer *
mash_lower_layer (MASH * mash, MashLayer * layer)
{
  MashLayer * below;

  if (layer->track) {
    below = _mash_track_get_layer_below (mash, layer->track, layer);
    mash_move_layer_above (mash, below, layer);
  }

  return layer;
}

MashLayer *
_mash_remove_layer (MASH * mash, MashLayer * layer)
{
  if (layer->track)
    _mash_track_remove_layer (mash, layer->track, layer);

  layer->track = MashNone;
  return layer;
}

MashTrack *
mash_layer_get_track (MASH * mash, MashLayer * layer)
{
  return layer->track;
}

MashDeck *
mash_layer_get_deck (MASH * mash, MashLayer * layer)
{
  MashTrack * track = layer->track;
  if (track == MashNone) return MashNone;
  return mash_track_get_deck (mash, track);
}

MashTimeType
mash_layer_set_timetype (MASH * mash, MashLayer * layer, MashTimeType new_type)
{
  MashTimeType old_type = layer->timetype;
  CDList * l;
  MashDeck * deck;
  MashSound * sound;

  if (old_type == new_type) return old_type;

  deck = mash_layer_get_deck (mash, layer);

  for (l = layer->sounds; l; l = l->next) {
    sound = (MashSound *)l->data.s_pointer;
    sound->start_time = mash_time_convert (mash, sound->start_time, old_type,
					   new_type);
    sound->duration = mash_time_convert (mash, sound->duration, old_type,
					 new_type);
  }

  layer->timetype = new_type;

  return old_type;
}

MashTimeType
mash_layer_get_timetype (MASH * mash, MashLayer * layer)
{
  return layer->timetype;
}

MashSound *
_mash_layer_add_sound (MASH * mash, MashLayer * layer, MashSound * sound,
		     MashTime start_time)
{
  sound->start_time = start_time;
  layer->sounds = cd_list_insert (mash, layer->sounds, CD_TYPE_POINTER,
				  CD_POINTER(sound),
				  (CDCmpFunc)mash_sound_later);
  mash_layer_ensure_coherency (mash, layer);
  return sound;
}

MashSound *
_mash_layer_remove_sound (MASH * mash, MashLayer * layer, MashSound * sound)
{
  layer->sounds = cd_list_remove (mash, layer->sounds, CD_TYPE_POINTER,
				  CD_POINTER(sound));
  mash_layer_ensure_coherency (mash, layer);
  return sound;
}

/*
 * mash_layer_get_sound_item_before (layer, time)
 *
 * Finds the last sound with a start_time before 'time'.
 */
static CDList *
mash_layer_get_sound_item_before (MASH * mash, MashLayer * layer, MashTime time)
{
  MashSound * s;
  CDList * l, * lp = MashNone;

  for (l = layer->sounds; l; l = l->next) {
    s = (MashSound *)l->data.s_pointer;
    if (_mash_time_gt (layer->timetype, s->start_time, time)) return lp;
    lp = l;
  }
  
  return lp;
}

/*
 * mash_layer_get_sound_item_at (layer, time)
 *
 * Finds the sound occurring at 'time'. If no sound is playing at 'time',
 * returns MashNone.
 */
CDList *
mash_layer_get_sound_item_at (MASH * mash, MashLayer * layer, MashTime time)
{
  MashTime t;
  CDList * l = mash_layer_get_sound_item_before (mash, layer, time);
  MashSound * s;

  if (l == MashNone) return MashNone;

  s = (MashSound *)l->data.s_pointer;

  t = _mash_time_add (layer->timetype, s->start_time, s->duration);
  if (_mash_time_le (layer->timetype, t, time)) l = MashNone;

  return l;
}

/*
 * mash_layer_get_sound_item_after (layer, time)
 *
 * Finds the first sound at or after 'time' and returns its list item.
 */
static CDList *
mash_layer_get_sound_item_after (MASH * mash, MashLayer * layer, MashTime time)
{
  MashSound * s;
  CDList * l;

  for (l = layer->sounds; l; l = l->next) {
    s = (MashSound *)l->data.s_pointer;
    if (_mash_time_ge (layer->timetype, s->start_time, time)) return l;
  }

  return MashNone;
}

/*
 * mash_layer_get_sound_after (layer, time)
 *
 * Finds the first sound with a start_time at or after 'time'.
 */
MashSound *
mash_layer_get_sound_after (MASH * mash, MashLayer * layer, MashTime time)
{
  CDList * l = mash_layer_get_sound_item_after (mash, layer, time);

  if (l == MashNone) return MashNone;
  else return (MashSound *)l->data.s_pointer;
}

MashSound *
_mash_layer_get_sound_prev (MASH * mash, MashLayer * layer, MashSound * sound)
{
  CDList * sound_item;
  MashSound * sn;

  if (layer->sounds == NULL) return MashNone;

  if (sound == MashNone) {
    sn = (MashSound *) layer->sounds->data.s_pointer;
  } else {
    sound_item = cd_list_find (mash, layer->sounds, CD_TYPE_POINTER,
			       CD_POINTER(sound));
    if (sound_item == NULL) return MashNone;
    sound_item = sound_item->prev;
    if (sound_item == NULL) return MashNone;
    sn = (MashSound *) sound_item->data.s_pointer;
  }

  return sn;
}

MashSound *
_mash_layer_get_sound_next (MASH * mash, MashLayer * layer, MashSound * sound)
{
  CDList * sound_item;
  MashSound * sn;

  if (layer->sounds == NULL) return MashNone;

  if (sound == NULL) {
    sn = (MashSound *)
      (cd_list_last (mash, layer->sounds, CD_TYPE_POINTER)).s_pointer;
  } else {
    sound_item = cd_list_find (mash, layer->sounds, CD_TYPE_POINTER,
			       CD_POINTER(sound));
    if (sound_item == NULL) return MashNone;
    sound_item = sound_item->next;
    if (sound_item == NULL) return MashNone;
    sn = (MashSound *) sound_item->data.s_pointer;
  }

  return sn;
}

MashLayer *
mash_get_layer_below (MASH * mash, MashLayer * layer)
{
  if (layer->track)
    return _mash_track_get_layer_below (mash, layer->track, layer);
  else
    return MashNone;
}

MashLayer *
mash_get_layer_above (MASH * mash, MashLayer * layer)
{
  if (layer->track)
    return _mash_track_get_layer_above (mash, layer->track, layer);
  else
    return MashNone;
}

static MashCount
mash_layer_length (MASH * mash, MashBase * base)
{
  MashLayer * layer = (MashLayer *)base;
  MashSound * sound = (MashSound *)
    (cd_list_last (mash, layer->sounds, CD_TYPE_POINTER)).s_pointer;
  MashTime end, t;

  if (sound == MashNone) {
    mash_dprintf ("[mash_layer_length] layer %p has no sounds\n", layer);
    return 0;
  }

  /* Convert sound's end time to offset and return that */
  end = _mash_time_add (layer->timetype, sound->start_time, sound->duration);
  t = mash_time_convert (mash, end, layer->timetype, MASH_TIME_SAMPLES);

  mash_dprintf ("[mash_layer_length] (%p) last sound ends at %d ticks == %ld samples\n",
	      layer, end.beat24s, t.samples);

  return t.samples;
}

static MashCount
mash_layer_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashLayer * layer = (MashLayer *)base;
  MashTime current_time;

  /* Evaluate this offset as a time value using the current samplerate and
   * tempo */
  current_time = mash_time_convert (mash, (MashTime)offset, MASH_TIME_SAMPLES,
				    layer->timetype);

  /* Cache the current sound item */
  layer->_current_sound_item =
    mash_layer_get_sound_item_at (mash, layer, current_time);

  if (layer->_current_sound_item == MashNone)
    layer->_current_sound_item =
      mash_layer_get_sound_item_after (mash, layer, current_time);

  layer->_current_offset = offset;

  return offset;
}

static MashCount
mash_layer_process_sound (MASH * mash, MashLayer * layer,
			  MashCount current_offset, MashSound * sound,
			  MashCount sound_offset, MashCount sound_length,
			  MashCount count,
			  MashStream * input, MashStream * output)
{
  MashCount remaining = count, processed = 0, n;

  /* If the next sound starts after the current offset, fill up to it */
  if (sound_offset > current_offset) {
    n = MIN (remaining, sound_offset - current_offset);

    mash_dprintf ("[mash_layer_process_sound] %p is after offset, filling %ld\n",
		sound, n);

    if (output != MashNone)
      n = mash_stream_write (mash, output, n, input);
    current_offset += n;
    processed += n;
    remaining -= n;
  }

  /* If that didn't fill the output, process part of the next sound */
  if (processed < count) {
    n = MIN (remaining, sound_offset + sound_length - current_offset);
    /* XXX: fix following line to use SEEK_CUR ??? */
    mash_seek (mash, (MashBase *)sound, current_offset - sound_offset,
	       SEEK_SET);
    n = mash_process (mash, (MashBase *)sound, n, input, output);
    processed += n;
    remaining -= n;
  }

  return processed;
}

static MashCount
mash_layer_process (MASH * mash, MashBase * base, MashCount count,
		    MashStream * input, MashStream * output)
{
  MashLayer * layer = (MashLayer *)base;
  MashTempo tempo = mash_get_tempo (mash);
  MashCount processed = 0, remaining = count, n;
  MashCount sound_offset, sound_length, next_offset;
  MashCount current_offset = mash_tell (mash, (MashBase *)layer);
  MashSound * sound, * sn;
  MashTime t;

  mash_dprintf ("PROCESS LAYER (%p, +%ld, %p -> %p) @ %ld\n",
	      layer, count, input, output, current_offset);


  if (layer->timetype == MASH_TIME_BEAT24S && layer->_current_tempo != tempo) {
#if 0
    MashCount new_offset;

#if 0
    if (layer->_current_sound_item == MashNone) {
      MashSamplerate samplerate = mash_get_samplerate (mash);
      int beat24s;

      beat24s = ((int)((float)mash_tell (mash, (MashBase *)layer) *
		       layer->_current_tempo * 24.0 / (samplerate * 60.0)));
      new_offset = (MashCount)(beat24s * samplerate * 60.0 / (tempo * 24.0));
    } else {
      sound = (MashSound *)layer->_current_sound_item->data.s_pointer;
      t = mash_time_convert (mash, sound->start_time, layer->timetype,
			     MASH_TIME_SAMPLES);
      
      new_offset = t.samples + mash_tell (mash, (MashBase *)sound);
    }
#else
    new_offset = layer->_current_offset * tempo / layer->_current_tempo;
#endif
  
    mash_layer_seek (mash, (MashBase *)layer, new_offset);
    layer->_current_tempo = tempo;
#else
    if (layer->_current_sound_item != MashNone)
      mash_layer_ensure_coherency (mash, layer);
#endif
  }

  while (remaining > 0) {

    if (layer->_current_sound_item == MashNone) {
      /* No more sounds */
      mash_dprintf ("[mash_layer_process] ## no more sounds!\n");

      n = (output == MashNone) ? remaining :
	mash_stream_write (mash, output, remaining, input);
      current_offset += n;
      processed += n;
      remaining -= n;
      break;
    }

    sound = (MashSound *)layer->_current_sound_item->data.s_pointer;
    t = mash_time_convert (mash, sound->start_time, layer->timetype,
			   MASH_TIME_SAMPLES);
    sound_offset = t.samples;

    t = mash_time_convert (mash, sound->duration, layer->timetype,
			   MASH_TIME_SAMPLES);
    sound_length = t.samples;

    if (layer->_current_sound_item->next) {
      sn = (MashSound *)(layer->_current_sound_item->next->data.s_pointer);
      t = mash_time_convert (mash, sn->start_time, layer->timetype,
			     MASH_TIME_SAMPLES);
      next_offset = t.samples;
      if (next_offset < sound_offset + sound_length)
        sound_length = next_offset - sound_offset;
    }

    /* *** We now have the next sound and its valid length *** */
    mash_dprintf ("[mash_layer_process] to process sound %p, [%ld, +%ld]\n",
		  sound, sound_offset, sound_length);

    n = mash_layer_process_sound (mash, layer, current_offset,
				  sound, sound_offset, sound_length,
				  remaining, input, output);
    current_offset += n;
    processed += n;
    remaining -= n;

    if (current_offset >= sound_offset + sound_length)
      layer->_current_sound_item = layer->_current_sound_item->next;
  }

  mash_dprintf ("[mash_layer_process] processed %ld\n", processed);

  if (processed == 0) {
    mash_set_error (mash, MASH_ERROR_NOOP);
    return -1;
  }

  layer->_current_offset = current_offset;

  return processed;
}

static int
mash_layer_flush (MASH * mash, MashBase * base)
{
  MashLayer * layer = (MashLayer *)base;
  MashBase * sound;

  if (layer->_current_sound_item == MashNone) return 0;

  sound = (MashBase *)layer->_current_sound_item->data.s_pointer;

  return mash_flush (mash, sound);
}

static MashLayer *
mash_layer_ensure_coherency (MASH * mash, MashLayer * layer)
{
  MashCount offset = mash_tell (mash, (MashBase *)layer);

  mash_layer_seek (mash, (MashBase *)layer, offset);
  layer->_current_tempo = mash_get_tempo (mash);

  return layer;
}

static struct _MashMethods _mash_layer_methods = {
  mash_layer_clone,   /* clone */
  mash_layer_destroy, /* destroy */
  NULL,             /* ready */
  NULL,             /* prepare */
  mash_layer_process, /* process */
  mash_layer_length,  /* length */
  mash_layer_seek,    /* seek */
  mash_layer_flush,   /* flush */
};

static MashLayer *
mash_layer_optimise (MASH * mash, MashLayer * layer)
{
  _mash_set_methods (mash, layer, &_mash_layer_methods);
  return layer;
}
		  
