/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashPCM: generic functions for anonymous blocks of PCM data.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * This file contains only generic functions to manipulate blocks of
 * MashPCM data. The MashPCM type is defined in <mash_types.h>,  usually as
 * a floating point value (float).
 *
 * The functions in this file are an excellent target for short vector
 * machine optimisations.
 *
 * Invariants
 * ----------
 *
 * N/A
 *
 */

#include <string.h>

#define __MASH__
#include "mash.h"


/* PFunc */

/*
 * _mash_pcm_clear_region (data, count)
 */
MashCount
_mash_pcm_clear_region (MashPCM * data, MashCount count, void * unused)
{
  memset (data, (MashPCM)0, count * sizeof (MashPCM));
  return count;
}


/* PVFunc */

MashCount
_mash_pcm_set (MashPCM * data, MashPCM value, MashCount count)
{
  MashCount i;

  for (i = 0; i < count; i++) {
    *data++ = value;
  }

  return count;
}

MashCount
_mash_pcm_gain (MashPCM * data, MashCount count, void * gain)
{
  MashPCM _gain = *(MashPCM *)gain;
  MashCount i;

  for (i = 0; i < count; i++) {
    *data++ *= _gain;
  }

  return count;
}


/* PPFunc */

/*
 * _mash_pcm_copy (src, dest, count)
 *
 * Copy PCM data from src to dest.
 */
MashCount
_mash_pcm_copy (MashPCM * src, MashPCM * dest, MashCount count, void * unused)
{
  memcpy (dest, src, count * sizeof (MashPCM));
  return count;
}

/*
 * _mash_pcm_add (src, dest, count)
 *
 * Add PCM data from src to dest.
 */
MashCount
_mash_pcm_add (MashPCM * src, MashPCM * dest, MashCount count, void * unused)
{
  MashCount i;

  for (i = 0; i < count; i++) {
    *dest++ += *src++;
  }

  return count;
}

/*
 * _mash_pcm_mult (src, dest, count)
 *
 * Multiply PCM data of dest by that in src.
 */
MashCount
_mash_pcm_mult (MashPCM * src, MashPCM * dest, MashCount count, void * unused)
{
  MashCount i;

  for (i = 0; i < count; i++) {
    *dest++ *= *src++;
  }

  return count;
}

/*
 * _mash_pcm_fade (src, dest, count)
 *
 * Fade PCM data of dest by that in src.
 */
MashCount
_mash_pcm_fade (MashPCM * src, MashPCM * dest, MashCount count, void * unused)
{
  MashCount i;

  for (i = 0; i < count; i++) {
    *dest++ *= (1.0 - *src++);
  }

  return count;
}

/*
 * _mash_pcm_interleave_2 (src1, src2, count, dest)
 *
 * Interleave data of src1 and src2, storing result in dest
 */
MashCount
_mash_pcm_interleave_2 (MashPCM * src1, MashPCM * src2, MashCount count, void * data)
{
  MashPCM * dest = (MashPCM *)data;
  MashCount i;

  for (i = 0; i < count; i++) {
    *dest++ = *src1++;
    *dest++ = *src2++;
  }

  return count;
}

/*
 * _mash_pcm_deinterleave_2 (dest1, dest2, count, src)
 *
 * Deinterleave data of src, storing result in dest1 and dest2
 */
MashCount
_mash_pcm_deinterleave_2 (MashPCM * dest1, MashPCM * dest2, MashCount count, void * data)
{
  MashPCM * src = (MashPCM *)data;
  MashCount i;

  for (i = 0; i < count; i++) {
    *dest1++ = *src++;
    *dest2++ = *src++;
  }

  return count;
}

/* PPPFunc */

/*
 * _mash_pcm_blend (src, blend, dest, count)
 *
 * Blend PCM data of 'src' into 'dest' by blend values in 'blend'.
 */
MashCount
_mash_pcm_blend (MashPCM * src, MashPCM * blend, MashPCM * dest, MashCount count,
	       void * unused)
{
  MashCount i;
  MashPCM b, d;

  for (i = 0; i < count; i++) {
    b = *blend++;
    d = (*dest * b) + (*src++ * (1.0 - b));
    *dest++ = d;
  }

  return count;
}


/* Miscellaneous */

/*
 * mash_pcm_write_linear (data, x1, y1, x2, y2, offset, count)
 *
 * Write 'count' samples at 'data' following line passing through (x1, y1)
 * and (x2, y2), with writing starting at x = 'offset'.
 */
MashCount
_mash_pcm_write_linear (MashPCM * data, MashCount x1, MashPCM y1, MashCount x2, MashPCM y2,
		      MashCount offset, MashCount count)
{
  MashCount i;

  mash_dprintf ("[mash_pcm_write_linear] ((%ld, %f) -> (%ld, %f), %ld +%ld)\n",
	  x1, y1, x2, y2, offset, count);

  for (i = 0; i < count; i++) {
    *data++ = y1 + (MashPCM)(i + offset - x1) * (y2 - y1) / (x2 - x1);
  }

  return count;
}
