/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashPlugin: A container for MashBase types.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <dlfcn.h>

#define __MASH__
#include "mash.h"

static CDList *
mash_plugin_initialise_static (MASH * mash)
{
  CDList * plugins = cd_list_new (mash);

  plugins = cd_list_join (mash, plugins, __gain_init (mash));
  plugins = cd_list_join (mash, plugins, __sndfile_init (mash));

  return plugins;
}

static CDList *
mash_plugin_init (MASH * mash, const char * path)
{
  void * module;
  MashPluginInitFunc init;

  module = dlopen (path, RTLD_NOW);

  if (!module) {
    fprintf (stderr, "Error opening %s: %s\n", path, dlerror ());
    mash_set_error (mash, MASH_ERROR_SYSTEM);
    return CD_EMPTY_LIST;
  }

  if ((init = dlsym (module, "mash_load")) != NULL) {
    return init (mash);
  }

  return CD_EMPTY_LIST;
}


/*
 * Implementation of stripname for matching plugins of form
 * lib(.*).so
 *
 * If you want to implement support for other dynamic library
 * naming schemes, the correct thing to do is implement another
 * one of these functions. The g_module stuff deals with the
 * actual filenames portably.
 */
static char *
stripname (char * d_name)
{
  int len, blen;
#define BUF_LEN 8
  char buf[BUF_LEN];

  if (strncmp (d_name, "lib", 3)) return 0;

  len = strlen (d_name);
  snprintf (buf, BUF_LEN, ".so.%d", MASH_PLUGIN_API_MAJOR);
  blen = strlen (buf);
  if (strncmp (&d_name[len-blen], buf, blen)) return 0;

  d_name[len-blen] = '\0';

  return &d_name[3];
}

#define BUFLEN 256

static CDList *
init_dynamic_plugins_dir (MASH * mash, char * dirname)
{
  CDList * plugins = cd_list_new (mash);
  DIR * dir;
  struct dirent * dirent;
  char * name;
  static char buf[BUFLEN];

  dir = opendir (dirname);
  if (!dir) {
    /* fail silently */
    return CD_EMPTY_LIST;
  }

  while ((dirent = readdir (dir)) != NULL) {
    if ((name = stripname (dirent->d_name)) != NULL) {
      snprintf (buf, BUFLEN, "%s/%s.so", dirname, dirent->d_name);
      plugins = cd_list_join (mash, plugins, mash_plugin_init (mash, buf));
    }
  }

  return plugins;
}


static CDList *
mash_plugin_initialise_dynamic (MASH * mash)
{
  return init_dynamic_plugins_dir (mash, PACKAGE_PLUGIN_DIR);
}

void
mash_plugin_initialise_defaults (MASH * mash)
{
  CDList * plugins = cd_list_new (mash);

  plugins = cd_list_join (mash, plugins, mash_plugin_initialise_static (mash));
  plugins = cd_list_join (mash, plugins, mash_plugin_initialise_dynamic (mash));

  cd_list_apply (mash, plugins, (CDFunc)_mash_register_plugin);
}

#if 0

MashPlugin
mash_plugin_new (void)
{
  MashPlugin plugin = mash_malloc (sizeof (struct _MashPlugin));
#if 0
  plugin->text = mash_meta_text_new ();
#endif
  return plugin;
}

MashMetaText *
mash_plugin_get_meta_text (MashPlugin * plugin)
{
  return plugin->text;
}

MashMetaText *
mash_plugin_set_meta_text (MashPlugin * plugin, MashMetaText * mt)
{
  MashMetaText * old = plugin->text;
  plugin->text = mt;
  return old;
}

int
mash_plugin_writeable (MashPlugin * plugin)
{
  return (plugin->flags & MASH_PLUGIN_WRITEABLE);
}

int
mash_plugin_seekable (MashPlugin * plugin)
{
  return (plugin->flags & MASH_PLUGIN_SEEKABLE);
}

int
mash_plugin_cacheable (MashPlugin * plugin)
{
  return (plugin->flags & MASH_PLUGIN_CACHEABLE);
}

int
mash_plugin_causal (MashPlugin * plugin)
{
  return (plugin->flags & MASH_PLUGIN_CAUSAL);
}

#endif
