/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * mash_plugin.h -- MASH internal data types and functions.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifndef __MASH_PLUGIN_H__
#define __MASH_PLUGIN_H__

/*#define DEBUG*/

#if defined(__MASH_PLUGIN__) || defined(__MASH__)

#include <stdio.h>

#include "ctxdata.h"

#define MASH_PLUGIN_API_MAJOR 1
#define MASH_PLUGIN_API_MINOR 0
#define MASH_PLUGIN_API_REVISION 0

typedef struct _MashMetaAuthor MashMetaAuthor;
typedef struct _MashMetaText MashMetaText;

typedef struct _MashParameterScheme MashParameterScheme;
typedef struct _MashParameterRange MashParameterRange;

typedef struct _MashNamedParameter MashNamedParameter;

typedef union _MashConstraint MashConstraint;

typedef struct _MashMethods MashMethods;;

#define MASH_PLUGIN_WRITEABLE 1<<0
#define MASH_PLUGIN_SEEKABLE  1<<1
#define MASH_PLUGIN_CACHEABLE 1<<2
#define MASH_PLUGIN_CAUSAL    1<<3

/* A base of a plugin */
typedef struct _MashPlugin MashPlugin;
/*typedef struct _MashBase * MashBase;*/

typedef struct _MashChunk MashChunk;

#if defined (__MASH__)
#include "mash_private.h"
#else
typedef MashOpaque MASH;
typedef MashOpaque MashPoint;
typedef MashOpaque MashEnvelope;
typedef MashOpaque MashStream;
typedef MashOpaque MashChannel;
typedef MashOpaque MashDeck;
typedef MashOpaque MashTrack;
typedef MashOpaque MashLayer;
typedef MashOpaque MashSound;
typedef MashOpaque MashSquareTone;
typedef MashOpaque MashMonitor;
#endif


typedef CDList * (*MashPluginInitFunc) (MASH * mash);

typedef MashBase * (*MashInitFunc) (MASH * mash, MashBase * base,
				    CDSet * parameters);

typedef CDSet * (*MashSuggestFunc) (MASH * mash, MashPlugin * plugin,
				    CDSet * parameters,
				    void * plugin_data);

typedef MashBase * (*MashCloneFunc) (MASH * mash, MashBase * base);
typedef int (*MashDestroyFunc) (MASH * mash, MashBase * base);
typedef int (*MashReadyFunc) (MASH * mash, MashBase * base);
typedef MashBase * (*MashPrepareFunc) (MASH * mash, MashBase * base);
typedef MashCount (*MashSeekFunc) (MASH * mash, MashBase * base, MashCount count);
typedef MashCount (*MashLengthFunc) (MASH * mash, MashBase * base);
typedef MashCount (*MashProcessFunc) (MASH * mash, MashBase * base, MashCount count,
                                  MashStream * input, MashStream * output);
typedef int (*MashFlushFunc) (MASH * mash, MashBase * base);

#define MASH_FLAGS_NONE (0)

#define MASH_AUTHOR(n,e) ((struct _MashMetaAuthor){(n),(e)})
#define MASH_ONE_AUTHOR(n,e) CD_SINGLETON_LIST(CD_POINTER(&(MASH_AUTHOR((n),(e)))))

/* ChunkChunkFuncs and all the rest */

/* MashChunkFunc: a function to apply to one chunk */
typedef MashCount (*MashChunkFunc) (MASH * mash, MashChunk * chunk, MashCount offset,
				MashCount count, int channelname, void * data);

/* MashChunkChunkFunc: a function to apply between two chunks :) */
typedef MashCount (*MashChunkChunkFunc) (MASH * mash,
				     MashChunk * src, MashCount src_offset,
				     MashChunk * dest, MashCount dest_offset,
				     MashCount count, int channelname,
				     void * data);

/* MashChunkChunkChunkFunc: a function to apply between THREE chunks !! */
typedef MashCount (*MashChunkChunkChunkFunc) (MASH * mash,
					  MashChunk * src1, MashCount src1_offset,
					  MashChunk * src2, MashCount src2_offset,
					  MashChunk * dest, MashCount dest_offset,
					  MashCount count, int channelname,
					  void * data);

struct _MashMetaAuthor {
  char * name;
  char * email;
};

struct _MashMetaText {
  char * identifier;
  char * category;
  char * description;
  char * copyright;
  char * url;
  CDList * authors;
};

struct _MashParameterRange {
  MashFlags valid_mask;
  MashParameter lower;
  MashParameter upper;
  MashParameter step;
};

struct _MashNamedParameter {
  char * name;
  MashParameter parameter;
};

#define MASHNAMED_PARAMETER(n,p) (&((struct _MashNamedParameter){(n),(p)}))

union _MashConstraint {
  CDList * list; /* list of MashNamedParameter */
  MashParameterRange * range;
};

struct _MashParameterScheme {
  char * name;
  char * description;
  MashParameterType type;
  MashConstraintType constraint_type;
  MashConstraint constraint;
  MashFlags hints;
};

struct _MashPlugin {
  MashMetaText * metatext;
  MashFlags flags;
  CDSet * init_scheme;
  MashInitFunc init;
  CDSet * process_scheme;
  MashSuggestFunc suggest;
  void * plugin_data;
};

struct _MashMethods {
  MashCloneFunc clone;
  MashDestroyFunc destroy;
  MashReadyFunc ready;
  MashPrepareFunc prepare;
  MashProcessFunc process;
  MashLengthFunc length;
  MashSeekFunc seek;
  MashFlushFunc flush;
};


struct _MashChunk {
  MashCount start_index;
  MashCount length;
  MashPCM * data;
};


/* debug */
void mash_dprintf (const char * fmt, ...);

/* MASH, mash_context */

MashError mash_set_error (MASH * mash, MashError error);


/* mash_base */
MashBase * mash_base_new (MASH * mash);

MashPlugin * mash_base_set_plugin (MASH * mash, MashBase * base, MashPlugin * plugin);
MashPlugin * mash_base_get_plugin (MASH * mash, MashBase * base);
MashMethods * mash_base_set_methods (MASH * mash, MashBase * base,
				 MashMethods * methods);
void * mash_base_set_instance_data (MASH * mash, MashBase * base, void * data);
void * mash_base_get_instance_data (MASH * mash, MashBase * base);

MashCount mash_base_get_mixlength (MASH * mash, MashBase * base);
MashSamplerate mash_base_get_samplerate (MASH * mash, MashBase * base);
MashTempo mash_base_get_tempo (MASH * mash, MashBase * base);
CDSet * mash_base_get_channels (MASH * mash, MashBase * base);

int mash_base_has_samplerate (MASH * mash, MashBase * base);
int mash_base_has_tempo (MASH * mash, MashBase * base);
int mash_base_encompasses_mixlength (MASH * mash, MashBase * base);
int mash_base_encompasses_channels (MASH * mash, MashBase * base);


/* mash_meta */
MashMetaText * mash_meta_text_new (MASH * mash);
void mash_meta_text_free (MASH * mash, MashMetaText * mt);
MashMetaText * mash_get_meta_text (MASH * mash, MashBase * base);
MashMetaText * mash_set_meta_text (MASH * mash, MashBase * base, MashMetaText * mt);


/* mash_null */
MashCount mash_null_length (MASH * mash, MashBase * base);
MashCount mash_null_process (MASH * mash, MashBase * base, MashCount count,
			 MashStream * input, MashStream * output);
MashCount mash_null_seek (MASH * mash, MashBase * base, MashCount offset);


/* mash_stream */
MashCount mash_stream_chunkfuncify (MASH * mash, MashStream * stream, MashCount count,
				MashChunkFunc func, void * data);
MashCount mash_stream_chunkchunkfuncify (MASH * mash,
				     MashStream * src, MashStream * dest,
				     MashCount count, MashChunkChunkFunc func,
				     void * data);
MashCount mash_stream_chunkchunkchunkfuncify (MASH * mash,
					  MashStream * src1, MashStream * src2,
					  MashStream * dest, MashCount count,
					  MashChunkChunkChunkFunc func,
					  void * data);

/* MashChannel */
MashCount mash_channel_chunkfuncify (MASH * mash, MashChannel * channel,
				 MashCount count,
				 MashChunkFunc func, int channelname,
				 void * data);
MashCount mash_channel_chunkchunkfuncify (MASH * mash,
				      MashChannel * src, MashChannel * dest,
				      MashCount count, MashChunkChunkFunc func,
				      int channelname, void * data);
MashCount mash_channel_chunkchunkchunkfuncify (MASH * mash,
					   MashChannel * src1, MashChannel * src2,
					   MashChannel * dest, MashCount count,
					   MashChunkChunkChunkFunc func,
					   int channelname, void * data);
/* MashPCM */

MashCount _mash_pcm_clear_region (MashPCM * data, MashCount count, void * unused);
MashCount _mash_pcm_set (MashPCM * data, MashPCM value, MashCount count);
MashCount _mash_pcm_gain (MashPCM * data, MashCount count,
		      /* (MashPCM *) */ void * gain);
MashCount _mash_pcm_copy (MashPCM * src, MashPCM * dest, MashCount count, void * unused);
MashCount _mash_pcm_add (MashPCM * src, MashPCM * dest, MashCount count, void * unused);
MashCount _mash_pcm_mult (MashPCM * src, MashPCM * dest, MashCount count, void * unused);
MashCount _mash_pcm_fade (MashPCM * src, MashPCM * dest, MashCount count, void * unused);
MashCount _mash_pcm_interleave_2 (MashPCM * src1, MashPCM * src2, MashCount count,
			      void * data);
MashCount _mash_pcm_deinterleave_2 (MashPCM * dest1, MashPCM * dest2, MashCount count,
				void * data);
MashCount _mash_pcm_blend (MashPCM * src, MashPCM * blend, MashPCM * dest,
		       MashCount count, void * unused);
MashCount _mash_pcm_write_linear (MashPCM * data, MashCount x1, MashPCM y1,
			      MashCount x2, MashPCM y2,
			      MashCount offset, MashCount count);

#endif /* defined(__MASH__) */

#endif /* __MASH_PLUGIN_H__ */
