/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * mash_private.h -- MASH internal data types and functions.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#ifndef __MASH_PRIVATE_H__
#define __MASH_PRIVATE_H__

/*#define DEBUG*/

#if defined(__MASH__)

#include <stdlib.h>
#include <stdio.h>

#if (HAVE_ALSA == 1)
#include <alsa/asoundlib.h>
#endif

#include "ctxdata.h"

#include "mash_plugin.h"

/* Max line length for error messages etc. */
#define MASH_MAXLINE 4096

#define MASH_DEFAULT_MIXLENGTH 1024
#define MASH_DEFAULT_SAMPLERATE 44100
#define MASH_DEFAULT_TEMPO 120

typedef struct _MashThreadContext MashThreadContext;
typedef struct _MashEnvironment MashEnvironment;
typedef struct _MashContext MashContext;

typedef MashThreadContext MASH;

typedef struct _MashPoint MashPoint;
typedef struct _MashEnvelope MashEnvelope;
/*typedef struct _MashChunk MashChunk;*/
typedef struct _MashChannel MashChannel;
typedef struct _MashStream MashStream;
typedef struct _MashDeck MashDeck;
typedef struct _MashTrack MashTrack;
typedef struct _MashLayer MashLayer;
typedef struct _MashSound MashSound;


struct _MashThreadContext {
  MashError last_error;
  MashContext * context;
  MashEnvironment * environment;
};

struct _MashEnvironment {
  MashCount refcount;
  CDList * plugins;
  CDList * bases;
  int purging;
};

struct _MashContext {
  MashSamplerate samplerate;
  MashTempo tempo;
  CDSet * channels;
  MashCount mixlength;
};

struct _MashBase {
  MashPlugin * plugin;
  MashMethods * methods;
  CDSet * parameters;
  MashCount offset; /* current position */
  MashContext context_limit;
  void * instance_data;
};

struct _MashPoint {
  MashTime time;
  MashPCM value;
};

struct _MashEnvelope {
  MashBase base;
  MashEnvelopeType type;
  MashTimeType timetype;
  CDList * points;
  CDList * _current_point_item;
  MashCount _current_offset;
};

/* XXX: multichannel envelopes ? */


struct _MashStream {
  MashBase base;
  CDSet * channels;
};

struct _MashChannel {
  CDList * chunks;
  MashCount _current_offset;
  CDList * _current_chunk;
};

struct _MashDeck {
  MashBase base;
  CDList * tracks;
  MashStream * _mixstream;
};

struct _MashTrack {
  MashBase base;
  MashDeck * deck;
  MashPCM gain;
  CDList * layers;
  MashStream * _mixstream_a;
  MashStream * _mixstream_b;
};

struct _MashLayer {
  MashBase base;
  MashTrack * track;
  MashTimeType timetype;
  CDList * sounds;
  /*MashTime _current_time;*/
  CDList * _current_sound_item;
  MashTempo _current_tempo;
  MashCount _current_offset;
};

struct _MashSound {
  MashBase base;
  MashBase * source;
  MashBase * rate_envelope;
  MashBase * gain_envelope;
  MashBase * blend_envelope;
  MashLayer * layer;
  MashTime start_time; /* position in layer */
  MashTime duration; /* maximum time length */
  MashCount cutin;  /* start offset into sound source */
  MashCount cutlength;
  MashCount _current_source_offset;
  MashStream * _rate_envstream;
  MashStream * _gain_envstream;
  MashStream * _blend_envstream;
};

typedef struct _MashSquareToneChannel MashSquareToneChannel;
typedef struct _MashSquareTone MashSquareTone;

struct _MashSquareToneChannel {
  MashCount _cycle_offset;
};

struct _MashSquareTone {
  MashBase base;
  float frequency;
  CDSet * channels;
};

typedef struct _MashMonitor MashMonitor;

#define MASHMONITOR_BUFFERLEN 2048

struct _MashMonitor {
  MashBase base;
  MashPCM databuffer[MASHMONITOR_BUFFERLEN];
  short playbuffer[MASHMONITOR_BUFFERLEN];
  int dev_dsp_fd;
  int mode;
  int mask;
  int format;
  int stereo;
  int frequency;
  int numfrags;
  int fragsize;
#if (HAVE_ALSA == 1)
  snd_pcm_t *alsa_dev;
#else
  void *alsa_dev;
#endif
};

#define _mash_time_zero(t) (MashTime)\
  (((t)==MASH_TIME_SAMPLES) ? ((MashCount)0) : \
  (((t)==MASH_TIME_SECONDS) ? ((float)0.0) : \
                            ((int)0)))

#define _mash_time_invalid(t) (MashTime)\
  (((t)==MASH_TIME_SAMPLES) ? ((MashCount)-1) : \
  (((t)==MASH_TIME_SECONDS) ? ((float)-1.0) : \
                            ((int)-1)))

#define _mash_time_is_invalid(t,ti) \
  (((t)==MASH_TIME_SAMPLES) ? ((ti).samples < (MashCount)0) : \
  (((t)==MASH_TIME_SECONDS) ? ((ti).seconds < (float)0.0) : \
                            ((ti).beat24s < (int)0)))

#define _mash_time_add(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? MASH_SAMPLES((t1).samples + (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? MASH_SECONDS((t1).seconds + (t2).seconds) : \
                            MASH_BEAT24S((t1).beat24s + (t2).beat24s)))

#define _mash_time_sub(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? MASH_SAMPLES((t1).samples - (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? MASH_SECONDS((t1).seconds - (t2).seconds) : \
                            MASH_BEAT24S((t1).beat24s - (t2).beat24s)))

#define _mash_time_eq(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples == (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds == (t2).seconds) : \
                            ((t1).beat24s == (t2).beat24s)))

#define _mash_time_gt(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples > (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds > (t2).seconds) : \
                            ((t1).beat24s > (t2).beat24s)))

#define _mash_time_lt(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples < (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds < (t2).seconds) : \
                            ((t1).beat24s < (t2).beat24s)))

#define _mash_time_ge(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples >= (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds >= (t2).seconds) : \
                            ((t1).beat24s >= (t2).beat24s)))

#define _mash_time_le(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples <= (t2).samples) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds <= (t2).seconds) : \
                            ((t1).beat24s <= (t2).beat24s)))

#define _mash_time_min(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples < (t2).samples ? (t1) : (t2)) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds < (t2).seconds ? (t1) : (t2)) : \
                            ((t1).beat24s < (t2).beat24s ? (t1) : (t2))))

#define _mash_time_max(t,t1,t2) \
  (((t)==MASH_TIME_SAMPLES) ? ((t1).samples > (t2).samples ? (t1) : (t2)) : \
  (((t)==MASH_TIME_SECONDS) ? ((t1).seconds > (t2).seconds ? (t1) : (t2)) : \
                            ((t1).beat24s > (t2).beat24s ? (t1) : (t2))))

#define _mash_base_get_samplerate(a,b) (((MashBase*)b)->context_limit.samplerate)
#define _mash_base_get_tempo(a,b) (((MashBase*)b)->context_limit.tempo)
#define _mash_base_get_mixlength(a,b) (((MashBase*)b)->context_limit.mixlength)
#define _mash_base_get_channels(a,b) (((MashBase*)b)->context_limit.channels)

#define _mash_set_plugin(a,b,p) (((MashBase*)b)->plugin = (p))
#define _mash_get_plugin(a,b) (((MashBase*)b)->plugin)
#define _mash_set_methods(a,b,m) (((MashBase*)b)->methods = (m))
#define _mash_get_methods(a,b) (((MashBase*)b)->methods)
#define _mash_set_instance_data(a,b,d) (((MashBase*)b)->instance_data = (d))
#define _mash_get_instance_data(a,b) (((MashBase*)b)->instance_data)
#define _mash_set_name(a,b,n) (((MashBase*)b)->name = (n))
#define _mash_get_name(a,b) (((MashBase*)b)->name)
#define _mash_clone(a,b) (((MashBase*)b)->methods->clone ((a), ((MashBase*)b)))
#define _mash_destroy(a,b) (((MashBase*)b)->methods->destroy ((a), ((MashBase*)b)))
#define _mash_prepare(a,b) (((MashBase*)b)->methods->prepare ((a), ((MashBase*)b)))
#define _mash_process(a,b,c,i,o) \
        (((MashBase*)b)->methods->process ((a),((MashBase*)b),(c),(i),(o)))
#define _mash_length(a,b) (((MashBase*)b)->methods->length ((a), ((MashBase*)b)))
#define _mash_flush(a,b) (((MashBase*)b)->methods->flush ((a), ((MashBase*)b)))


/* util */
#define mash_malloc(x) calloc(1, x)
#define mash_free free

/* debug */
void mash_debug_down (void);
void mash_debug_up (void);

/* MASH, mash_context */

MashBase * mash_base_new_subclass (MASH * mash, size_t size);

MashContext * _mash_context_copy (MASH * mash, MashContext * dest);
MashContext * _mash_context_merge (MASH * mash, MashContext * dest);
MASH * _mash_register_plugin (MASH * mash, MashPlugin * plugin);
MASH * _mash_unregister_plugin (MASH * mash, MashPlugin * plugin);
MASH * _mash_register_base (MASH * mash, MashBase * base);
MASH * _mash_unregister_base (MASH * mash, MashBase * base);

/* mash_plugin */
void mash_plugin_initialise_defaults (MASH * mash);

/* mash_deck */
MashTrack * _mash_deck_add_track (MASH * mash, MashDeck * deck, MashTrack * track);
MashTrack * _mash_deck_remove_track (MASH * mash, MashDeck * deck, MashTrack * track);

/* mash_track */
MashBase * mash_track_clone (MASH * mash, MashBase * base);
MashLayer * _mash_track_add_layer_above (MASH * mash, MashTrack * track,
				     MashLayer * layer, MashLayer * above);
MashLayer * _mash_track_remove_layer (MASH * mash, MashTrack * track,
				  MashLayer * layer);
MashLayer * _mash_track_get_layer_above (MASH * mash, MashTrack * track,
				     MashLayer * above);
MashLayer * _mash_track_get_layer_below (MASH * mash, MashTrack * track,
				     MashLayer * below);

/* mash_layer */
MashLayer * _mash_remove_layer (MASH * mash, MashLayer * layer);
MashBase * mash_layer_clone (MASH * mash, MashBase * base);
MashSound * _mash_layer_add_sound (MASH * mash, MashLayer * layer, MashSound * sound,
			     MashTime position);
MashSound * _mash_layer_remove_sound (MASH * mash, MashLayer * layer,
				  MashSound * sound);
MashSound * _mash_layer_get_sound_prev (MASH * mash, MashLayer * layer,
				    MashSound * sound);
MashSound * _mash_layer_get_sound_next (MASH * mash, MashLayer * layer,
				    MashSound * sound);

/* mash_sound */
MashBase *  mash_sound_clone_with_layer (MASH * mash, MashBase * base,
				    MashLayer * new_layer);
int mash_sound_later (MASH * mash, MashSound * s1, MashSound * s2);

/* mash_envelope */
MashBase * mash_envelope_clone (MASH * mash, MashBase * base);


/* mash_channel */
MashChannel * mash_channel_new (MASH * mash);
MashChannel * mash_channel_clone (MASH * mash, MashChannel * channel);
int mash_channel_destroy (MASH * mash, MashBase * base);

MashChunk * mash_channel_add_new_chunk (MASH * mash, MashChannel * channel,
				    MashCount offset, MashCount length);

MashCount mash_channel_write0 (MASH * mash, MashChannel * channel, MashCount length);

MashCount _mash_channel_write (MASH * mash, MashChannel * channel, MashCount count,
			   MashChannel * data);
MashCount _mash_channel_length (MASH * mash, MashChannel * channel);
MashCount _mash_channel_seek (MASH * mash, MashChannel * channel, MashCount offset);

MashCount mash_channel_interleave_2 (MASH * mash,
				 MashChannel * src1, MashChannel * src2,
				 MashPCM * dest, MashCount count);
MashCount mash_channel_deinterleave_2 (MASH * mash,
				   MashChannel * dest1, MashChannel * dest2,
				   MashPCM * src, MashCount count);
MashCount mash_channel_mix (MASH * mash, MashChannel * src, MashChannel * dest,
			MashCount count);


/* mash_channelset */
void mash_channelset_initialise_defaults (MASH * mash);

/* mash_chunk */
MashChunk * mash_chunk_new (MASH * mash, MashCount start_index, MashCount length);
MashChunk * mash_chunk_clone (MASH * mash, MashChunk * chunk);
void mash_chunk_free (MASH * mash, MashChunk * chunk);
MashCount _mash_chunk_clear_region (MASH * mash, MashChunk * chunk, MashCount start,
				MashCount length, int channelname,
				void * unused);
MashCount _mash_chunk_gain (MASH * mash, MashChunk * chunk, MashCount start,
			MashCount count,
			int channelname, /* (MashPCM *) */ void * gain);
MashCount _mash_chunk_copy (MASH * mash, MashChunk * src, MashCount src_offset,
			MashChunk * dest, MashCount dest_offset, MashCount count,
			int channelname, void * unused);
MashCount _mash_chunk_add_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
			       MashChunk * dest, MashCount dest_offset,
			       MashCount count, int channelname, void * unused);
MashCount _mash_chunk_mult_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
				MashChunk * dest, MashCount dest_offset,
				MashCount count, int channelname, void * unused);
MashCount _mash_chunk_fade_inplace (MASH * mash, MashChunk * src, MashCount src_offset,
				MashChunk * dest, MashCount dest_offset,
				MashCount count, int channelname, void * unused);
MashCount _mash_chunk_interleave_2 (MASH * mash,
				MashChunk * src1, MashCount src1_offset,
				MashChunk * src2, MashCount src2_offset,
				MashCount count, int unused, void * dest);
MashCount _mash_chunk_deinterleave_2 (MASH * mash,
				  MashChunk * dest1, MashCount dest1_offset,
				  MashChunk * dest2, MashCount dest2_offset,
				  MashCount count, int unused, void * src);
MashCount _mash_chunk_blend_inplace (MASH * mash,
				 MashChunk * src, MashCount src_offset,
				 MashChunk * blend, MashCount blend_offset,
				 MashChunk * dest, MashCount dest_offset,
				 MashCount count, int channelname,
				 void * unused);

/* XXX: remove these when dynamic! */
CDList * __gain_init (MASH * mash);
CDList * __sndfile_init (MASH * mash);

#endif /* defined(__MASH__) */

#endif /* __MASH_PRIVATE_H__ */
