/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashSndfile: a libsndfile handler
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#include <stdio.h>
#include <sndfile.h>
#include <string.h>

#define __MASH__
#include "mash.h"

#define PATH_KEY 1

typedef struct _MashSndfileInstance MashSndfileInstance;

struct _MashSndfileInstance {
  char * path;
  int writing;
  SNDFILE * file;
  SF_INFO info;
};


/* Optimisation dependencies: none */
static MashBase * mash_sndfile_optimise (MASH * mash, MashBase * sndfile);


static MashBase *
mash_sndfile_create (MASH * mash, MashBase * sndfile,
		     const char * path, int writing)
{
  MashSndfileInstance * si =
    mash_malloc (sizeof (struct _MashSndfileInstance));

  si->path = strdup (path);
  si->writing = writing;

  if (writing) {
    si->info.samplerate = mash_get_samplerate (mash);
    si->info.channels = 1; /* XXX: how many channels, or specify? */
    si->info.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16; /* XXX: assumes WAV */

    si->file = sf_open  (path, SFM_WRITE, &si->info);
  } else {
    si->file = sf_open (path, SFM_READ, &si->info);
  }

  if (si->file == NULL) {
    mash_set_error (mash, MASH_ERROR_SYSTEM);
    mash_destroy (mash, (MashBase *)sndfile);
    return MashNone;
  }

  sf_command (si->file, SFC_SET_NORM_FLOAT, NULL, SF_TRUE);

  if (writing)
    sf_command (si->file, SFC_SET_ADD_DITHER_ON_WRITE, NULL, SF_TRUE);

  sndfile->instance_data = si;

  return sndfile;
}

static MashBase *
mash_sndfile_reader_init (MASH * mash, MashBase * base, CDSet * parameters)
{
  char * path;

  path = (cd_set_find (mash, parameters, PATH_KEY)).s_string;

  if (mash_sndfile_create (mash, base, path, 0) == MashNone)
  	return MashNone;

  mash_sndfile_optimise (mash, base);
  return base;
}

static MashBase *
mash_sndfile_writer_init (MASH * mash, MashBase * base, CDSet * parameters)
{
  char * path;

  path = (cd_set_find (mash, parameters, PATH_KEY)).s_string;

  mash_sndfile_create (mash, base, path, 1);
  mash_sndfile_optimise (mash, base);
  return base;
}

static MashBase *
mash_sndfile_clone (MASH * mash, MashBase * base)
{
  MashBase * new_sndfile = mash_base_new (mash);
  MashSndfileInstance * si = (MashSndfileInstance *)base->instance_data;
  mash_sndfile_create (mash, new_sndfile, si->path, si->writing);
  mash_sndfile_optimise (mash, new_sndfile);
  return new_sndfile;
}

static int
mash_sndfile_destroy (MASH * mash, MashBase * base)
{
  MashSndfileInstance * si = (MashSndfileInstance *)base->instance_data;
  if (si->file != NULL) sf_close (si->file);
  mash_free (si);
  mash_free (base);
  return 0;
}

/* An MashChunkFunc for creating sndfile */
static MashCount
mash_sndfile_read_into_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
			      MashCount count, int channelname, void * data)
{
  MashBase * sndfile = (MashBase *)data;
  MashPCM * d;
  MashCount remaining = count, written = 0, n;
  MashSndfileInstance * si = (MashSndfileInstance *)sndfile->instance_data;

  mash_dprintf ("[mash_sndfile_read_into_chunk] (%p, +%ld) @ %ld\n",
	      sndfile, count, mash_tell (mash, sndfile));

  d = &chunk->data[offset];

  while (remaining > 0) {
    n = MIN (remaining, 4096);
    n = sf_readf_float (si->file, d, n);

    if (n == 0) { /* EOF */
      n = _mash_pcm_set (d, 0.0, remaining);
    }

    remaining -= n;
    written += n;

    d += n;
  }

  return written;
}

static MashCount
mash_sndfile_write_from_chunk (MASH * mash, MashChunk * chunk,
			       MashCount offset, MashCount count,
			       int channelname, void * data)
{
  MashBase * sndfile = (MashBase *)data;
  MashPCM * d;
  MashCount remaining = count, read = 0, n;
  MashSndfileInstance * si = (MashSndfileInstance *) sndfile->instance_data;

  mash_dprintf ("[mash_sndfile_write_from_chunk] (%p, +%ld) @ %ld\n",
		sndfile, count, mash_tell (mash, sndfile));

  d = &chunk->data[offset];

  while (remaining > 0) {
    n = MIN (remaining, 4096);
    n = sf_write_float (si->file, d, n);

    if (n == 0) { /* EOF */
      n = remaining;
    }

    remaining -= n;
    read += n;

    d += n;
  }

  return read;
}

static MashCount
mash_sndfile_reader_process (MASH * mash, MashBase * base, MashCount count,
			   MashStream * input, MashStream * output)
{
  return mash_stream_chunkfuncify (mash, output, count,
				   mash_sndfile_read_into_chunk, base);
}

static MashCount
mash_sndfile_writer_process (MASH * mash, MashBase * base, MashCount count,
			   MashStream * input, MashStream * output)
{
  return mash_stream_chunkfuncify (mash, output, count,
				   mash_sndfile_write_from_chunk, base);
}

static MashCount
mash_sndfile_length (MASH * mash, MashBase * base)
{
  MashSndfileInstance * si = (MashSndfileInstance *)base->instance_data;
  return si->info.frames;
}

static MashCount
mash_sndfile_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashSndfileInstance * si = (MashSndfileInstance *)base->instance_data;
  return sf_seek (si->file, offset, SEEK_SET);
}

static struct _MashMethods _mash_sndfile_reader_methods = {
  mash_sndfile_clone,
  mash_sndfile_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_sndfile_reader_process,
  mash_sndfile_length,
  mash_sndfile_seek,
  NULL, /* flush */
};

static struct _MashMethods _mash_sndfile_writer_methods = {
  mash_sndfile_clone,
  mash_sndfile_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_sndfile_writer_process,
  mash_sndfile_length,
  mash_sndfile_seek,
  NULL, /* flush */
};  

static MashBase *
mash_sndfile_optimise (MASH * mash, MashBase * sndfile)
{
  MashSndfileInstance * si = (MashSndfileInstance *)sndfile->instance_data;

  if (si->writing)
    mash_base_set_methods (mash, sndfile, &_mash_sndfile_writer_methods);
  else
    mash_base_set_methods (mash, sndfile, &_mash_sndfile_reader_methods);

  return sndfile;
}


static struct _MashParameterScheme path_scheme = {
  "path",
  "Path to sound file",
  MASH_TYPE_STRING,
  MASH_CONSTRAINT_TYPE_NONE,
  MASH_CONSTRAINT_EMPTY,
  MASH_HINT_FILENAME,
};

#if 0
static MashNamedParameter types[] = {
  MASHNAMED_PARAMETER ("WAV", CD_INT(SF_FORMAT_WAV)),
  MASHNAMED_PARAMETER ("AIFF", CD_INT(SF_FORMAT_AIFF)),
  MASHNAMED_PARAMETER ("Sun/NeXT AU format (big endian)", CD_INT(SF_FORMAT_AU)),
  MASHNAMED_PARAMETER ("DEC AU format (little endian)", CD_INT(SF_FORMAT_AULE)),
  MASHNAMED_PARAMETER ("RAW PCM data", CD_INT(SF_FORMAT_RAW)),
  MASHNAMED_PARAMETER ("Ensoniq PARIS", CD_INT(SF_FORMAT_PAF)),
  MASHNAMED_PARAMETER ("Amiga IFF / SVX8 / SV16", CD_INT(SF_FORMAT_SVX)),
  MASHNAMED_PARAMETER ("Sphere NIST", CD_INT(SF_FORMAT_NIST)),
  MASHNAMED_PARAMETER ("Windows Media Audio", CD_INT(SF_FORMAT_WMA)),
  MASHNAMED_PARAMETER ("Sekd Samplitude", CD_INT(SF_FORMAT_SMPLTD)),
  MASHNAMED_PARAMETER ("VOC", CD_INT(SF_FORMAT_VOC)),
  MASHNAMED_PARAMETER ("Sound Designer 2", CD_INT(SF_FORMAT_SD2)),
  MASHNAMED_PARAMETER ("Rex2", CD_INT(SF_FORMAT_REX2)),
  MASHNAMED_PARAMETER ("PCM", CD_INT(SF_FORMAT_PCM)),
  MASHNAMED_PARAMETER ("32 bit floats", CD_INT(SF_FORMAT_FLOAT)),
  MASHNAMED_PARAMETER ("U-Law encoded", CD_INT(SF_FORMAT_ULAW)),
  MASHNAMED_PARAMETER ("A-Law encoded", CD_INT(SF_FORMAT_ALAW)),
  MASHNAMED_PARAMETER ("IMA ADPCM", CD_INT(SF_FORMAT_IMASHADPCM)),
  MASHNAMED_PARAMETER ("Microsoft ADPCM", CD_INT(SF_FORMAT_MS_ADPCM)),
  MASHNAMED_PARAMETER ("Big endian PCM", CD_INT(SF_FORMAT_PCM_BE)),
  MASHNAMED_PARAMETER ("Little endian PCM", CD_INT(SF_FORMAT_PCM_LE)),
  MASHNAMED_PARAMETER ("Signed 8 bit PCM", CD_INT(SF_FORMAT_PCM_S8)),
  MASHNAMED_PARAMETER ("Unsigned 8 bit PCM", CD_INT(SF_FORMAT_PCM_U8)),
  MASHNAMED_PARAMETER ("SVX Fibonacci Delta", CD_INT(SF_FORMAT_SVX_FIB)),
  MASHNAMED_PARAMETER ("SVX Exponential Delta", CD_INT(SF_FORMAT_SVX_EXP)),
  MASHNAMED_PARAMETER ("GSM 6.10 Encoding", CD_INT(SF_FORMAT_GSM610)),
  MASHNAMED_PARAMETER ("32kbs G721 ADPCM", CD_INT(SF_FORMAT_G721_32)),
  MASHNAMED_PARAMETER ("24kbs G723 ADPCM", CD_INT(SF_FORMAT_G723_24)),
};
#endif

static struct _MashParameterScheme format_scheme = {
  "format",
  "Format of sound file",
  MASH_TYPE_INT,
  MASH_CONSTRAINT_TYPE_LIST,
  MASH_CONSTRAINT_EMPTY,
  MASH_HINT_DEFAULT,
};

static struct _MashMetaText sndfile_reader_metatext = {
  "builtin::sndfile_reader",
  "File::Sndfile Reader",
  "Reads PCM audio files using libsndfile",
  "Copyright (C) 2001, 2002 CSIRO Australia",
  "http://www.metadecks.org/software/mash/plugins/sndfile.html",
  MASH_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _MashMetaText sndfile_writer_metatext = {
  "builtin::sndfile_writer",
  "File::Sndfile Writer",
  "Writes PCM audio files using libsndfile",
  "Copyright (C) 2001, 2002 CSIRO Australia",
  "http://www.metadecks.org/software/mash/plugins/sndfile.html",
  MASH_ONE_AUTHOR ("Conrad Parker", "Conrad.Parker@CSIRO.AU"),
};

static struct _MashPlugin sndfile_reader_plugin = {
  &sndfile_reader_metatext,
  MASH_FLAGS_NONE,
  CD_EMPTY_SET, /* init scheme */
  mash_sndfile_reader_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

static struct _MashPlugin sndfile_writer_plugin = {
  &sndfile_writer_metatext,
  MASH_FLAGS_NONE,
  CD_EMPTY_SET, /* init scheme */
  mash_sndfile_writer_init,
  CD_EMPTY_SET, /* process scheme */
  NULL, /* suggests */
  NULL, /* plugin data */
};

/* module init function */
CDList *
__sndfile_init (MASH * mash)
{
  CDList * plugins = cd_list_new (mash);
  int i, count;
  SF_FORMAT_INFO info;
  MashNamedParameter * param;

  sndfile_reader_plugin.init_scheme =
    cd_set_insert (mash, sndfile_reader_plugin.init_scheme, PATH_KEY,
		   CD_POINTER(&path_scheme));

  plugins = cd_list_prepend (mash, plugins,
			     CD_POINTER(&sndfile_reader_plugin));

  format_scheme.constraint.list = cd_list_new (mash);

  sf_command (NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof (int)) ;

  for (i = 0; i < count ; i++) {
    info.format = i ;
    sf_command (NULL, SFC_GET_FORMAT_MAJOR, &info, sizeof (info)) ;

    param = mash_malloc (sizeof(MashNamedParameter));
    param->name = strdup (info.name);
    param->parameter = CD_INT(info.format);

    format_scheme.constraint.list =
      cd_list_append (mash, format_scheme.constraint.list, CD_POINTER(param));
  }

  sndfile_writer_plugin.init_scheme =
    cd_set_insert (mash, sndfile_writer_plugin.init_scheme, PATH_KEY,
		   CD_POINTER(&path_scheme));

  plugins = cd_list_prepend (mash, plugins,
			     CD_POINTER(&sndfile_writer_plugin));

  return plugins;
}

