/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashSound: An instance of an base within an MashLayer sequence.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A sound is contained within a layer. Each sound is a unique entity, but
 * many sounds may have the same source base.
 *
 * Invariants
 * ----------
 *
 * A sound must be contained in a layer.
 */

#include <string.h>

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: none */
static MashSound * mash_sound_optimise (MASH * mash, MashSound * sound);

/*
 * mash_sound_replace_mixstreams (mash, sound)
 *
 * Replaces mix- (and envelope-) streams of 'sound' with new ones of
 * the mash's mixlength.
 */
static void
mash_sound_replace_mixstreams (MASH * mash, MashSound * sound)
{
  MashCount mixlength = _mash_base_get_mixlength (mash, sound);

  if (sound->_rate_envstream != MashNone)
    mash_destroy (mash, (MashBase *)sound->_rate_envstream);
  if (sound->_gain_envstream != MashNone)
    mash_destroy (mash, (MashBase *)sound->_gain_envstream);
  if (sound->_blend_envstream != MashNone)
    mash_destroy (mash, (MashBase *)sound->_blend_envstream);

  sound->_rate_envstream =
    mash_stream_new_contiguous (mash, mixlength);
  sound->_gain_envstream =
    mash_stream_new_contiguous (mash, mixlength);
  sound->_blend_envstream =
    mash_stream_new_contiguous (mash, mixlength);
}

static MashBase *
mash_sound_init (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;
  sound->rate_envelope = sound->gain_envelope = sound->blend_envelope =
    MashNone;
  sound->cutin = sound->cutlength = 0;
  sound->_rate_envstream = sound->_gain_envstream = sound->_blend_envstream =
    MashNone;
  mash_sound_replace_mixstreams (mash, sound);
  mash_sound_optimise (mash, sound);
  return (MashBase *)sound;
}

static MashSound *
mash_sound_new (MASH * mash)
{
  return (MashSound *)
    mash_base_new_subclass (mash, sizeof (struct _MashSound));
}

MashBase *
mash_sound_clone_invalid (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;
  MashSound * new_sound = mash_sound_new (mash);

  memcpy (new_sound, sound, sizeof (struct _MashSound));
  new_sound->layer = MashNone;

  return (MashBase *)new_sound;
}

MashBase *
mash_sound_clone_with_layer (MASH * mash, MashBase * base, MashLayer * new_layer)
{
  MashSound * sound = (MashSound *)base;
  MashSound * new_sound = mash_sound_new (mash);

  memcpy (new_sound, sound, sizeof (struct _MashSound));
  new_sound->layer = new_layer;
  _mash_layer_add_sound (mash, new_layer, new_sound, new_sound->start_time);

  return (MashBase *)new_sound;
}

static int
mash_sound_destroy (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;

  mash_remove_sound (mash, sound);

  if (sound->rate_envelope)
    mash_destroy (mash, sound->rate_envelope);
  if (sound->gain_envelope)
    mash_destroy (mash, sound->gain_envelope);
  if (sound->blend_envelope)
    mash_destroy (mash, sound->blend_envelope);
  if (sound->_rate_envstream)
    mash_destroy (mash, (MashBase *)sound->_rate_envstream);
  if (sound->_gain_envstream)
    mash_destroy (mash, (MashBase *)sound->_gain_envstream);
  if (sound->_blend_envstream)
    mash_destroy (mash, (MashBase *)sound->_blend_envstream);
  mash_free (sound);

  return 0;
}

static int
mash_sound_ready (MASH * mash, MashBase * base)
{
  return 0;
}

static MashBase *
mash_sound_prepare (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;
  mash_sound_replace_mixstreams (mash, sound);
  return base;
}

MashBase *
mash_sound_set_source (MASH * mash, MashSound * sound, MashBase * source)
{
  MashBase * old = sound->source;
  sound->source = source;
  return old;
}

MashBase *
mash_sound_get_source (MASH * mash, MashSound * sound)
{
  return sound->source;
}

MashSound *
mash_add_sound (MASH * mash, MashBase * source, MashLayer * layer, MashTime start_time,
	      MashTime duration)
{
  MashSound * sound = mash_sound_new (mash);

  sound->layer = layer;
  sound->start_time = start_time;
  sound->duration = duration;
  _mash_layer_add_sound (mash, layer, sound, start_time);
  sound->source = source;
  mash_sound_init (mash, (MashBase *)sound);
  return sound;
}

/* XXX: This breaks sound invariant */
MashSound *
mash_remove_sound (MASH * mash, MashSound * sound)
{
  if (sound->layer == MashNone) return MashNone;
  return _mash_layer_remove_sound (mash, sound->layer, sound);
}

MashLayer *
mash_sound_get_layer (MASH * mash, MashSound * sound)
{
  return sound->layer;
}

MashTrack *
mash_sound_get_track (MASH * mash, MashSound * sound)
{
  if (sound->layer == MashNone) return MashNone;
  return mash_layer_get_track (mash, sound->layer);
}

MashDeck *
mash_sound_get_deck (MASH * mash, MashSound * sound)
{
  if (sound->layer == MashNone) return MashNone;
  return mash_layer_get_deck (mash, sound->layer);
}

MashTime
mash_sound_move (MASH * mash, MashSound * sound, MashTime start_time)
{
  MashTime old = sound->start_time;
  MashLayer * layer = sound->layer;
  if (mash_remove_sound (mash, sound) != MashNone) {
    _mash_layer_add_sound (mash, layer, sound, start_time);
    return old;
  } else {
    return _mash_time_invalid (sound->layer->timetype);
  } 
}

MashSound *
mash_sound_get_prev (MASH * mash, MashSound * sound)
{
  if (sound->layer)
    return _mash_layer_get_sound_prev (mash, sound->layer, sound);
  else
    return MashNone;
}

MashSound *
mash_sound_get_next (MASH * mash, MashSound * sound)
{
  if (sound->layer)
    return _mash_layer_get_sound_next (mash, sound->layer, sound);
  else
    return MashNone;
}

int
mash_sound_later (MASH * mash, MashSound * s1, MashSound * s2)
{
  MashLayer * layer = mash_sound_get_layer (mash, s1);
  return _mash_time_gt (layer->timetype, s1->start_time, s2->start_time);
}

MashTime
mash_sound_set_start_time (MASH * mash, MashSound * sound, MashTime start_time)
{
  MashTime old = sound->start_time;
  MashLayer * layer = sound->layer;
  _mash_layer_remove_sound (mash, layer, sound);
  _mash_layer_add_sound (mash, layer, sound, start_time);
  return old;
}

MashTime
mash_sound_get_start_time (MASH * mash, MashSound * sound)
{
  return sound->start_time;
}

MashTime
mash_sound_set_duration (MASH * mash, MashSound * sound, MashTime duration)
{
  MashTime old = sound->duration;
  sound->duration = duration;
  return old;
}

MashTime
mash_sound_get_duration (MASH * mash, MashSound * sound)
{
  return sound->duration;
}

MashBase *
mash_sound_set_rate_envelope (MASH * mash, MashSound * sound,
			    MashBase * rate_envelope)
{
  MashBase * old = sound->rate_envelope;
  sound->rate_envelope = rate_envelope;
  return old;
}

MashBase *
mash_sound_get_rate_envelope (MASH * mash, MashSound * sound)
{
  if (sound == MashNone) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return MashNone;
  }
  return sound->rate_envelope;
}

MashBase *
mash_sound_set_gain_envelope (MASH * mash, MashSound * sound,
			    MashBase * gain_envelope)
{
  MashBase * old;

  if (sound == MashNone) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return MashNone;
  }
  old = sound->gain_envelope;
  sound->gain_envelope = gain_envelope;

  return old;
}

MashBase *
mash_sound_get_gain_envelope (MASH * mash, MashSound * sound)
{
  if (sound == MashNone) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return MashNone;
  }
  return sound->gain_envelope;
}

MashBase *
mash_sound_set_blend_envelope (MASH * mash, MashSound * sound,
			     MashBase * blend_envelope)
{
  MashBase * old = sound->blend_envelope;
  sound->blend_envelope = blend_envelope;
  return old;
}

MashBase *
mash_sound_get_blend_envelope (MASH * mash, MashSound * sound)
{
  if (sound == MashNone) {
    mash_set_error (mash, MASH_ERROR_INVALID);
    return MashNone;
  }
  return sound->blend_envelope;
}

static MashCount
_mash_sound_fade (MASH * mash, MashSound * sound, MashCount count,
		  MashStream * input, MashStream * output)
{
  MashCount output_offset, n;

  output_offset = mash_tell (mash, (MashBase *)output);

  /* XXX: this wanted to use 'block' not 'count': have we bounds
   * checked the sound ?? */

  mash_seek (mash, (MashBase *)sound->_blend_envstream, 0, SEEK_SET);
  n = mash_process (mash, sound->blend_envelope, count, MashNone,
		  sound->_blend_envstream);
  mash_stream_write (mash, output, count, input);
  mash_seek (mash, (MashBase *)sound->_blend_envstream, 0, SEEK_SET);
  mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
  n = mash_stream_fade (mash, sound->_blend_envstream, output, count);

  return n;
}

/* Do rate conversion, handle offset etc.: get raw sound data */
static MashCount
_mash_sound_get_raw (MASH * mash, MashSound * sound, MashCount offset,
		     MashCount count, MashStream * input, MashStream * output)
{
  MashCount block, n;

  mash_dprintf ("[_mash_sound_get_raw] (%p, +%ld, %p -> %p) @ %ld\n",
	      sound, count, input, output, offset);

  /* XXX: deal with rate conversion!! */

  if (sound->cutlength > 0) {
    if (offset > sound->cutlength) {
      mash_set_error (mash, MASH_ERROR_SILENCE);
      return -1;
    }
    block = MIN (count, sound->cutlength - offset);
  } else {
    block = count;
  }

  mash_dprintf ("[_mash_sound_get_raw] block +%ld (cutin: %ld, cutlength: %ld)\n",
	      block, sound->cutin, sound->cutlength);

  mash_seek (mash, sound->source, sound->cutin + offset, SEEK_SET);
  n = mash_process (mash, sound->source, block, input, output);

  if (n == -1) {
    mash_dprintf ("error getting source data: %s\n",
		mash_error_string(mash, mash_last_error(mash)));
  } else {

    if (block < count)
      n += mash_stream_write0 (mash, output, count - block);
  }

  mash_dprintf ("[_mash_sound_get_raw] got %ld raw samples\n", n);

  return n;
}

static MashCount
_mash_sound_apply_gain (MASH * mash, MashSound * sound,
		      MashCount offset, MashCount count,
		      MashStream * data, MashCount data_offset)
{
  MashCount n;

  mash_dprintf ("in _mash_sound_apply_gain (%p, +%ld)\n", sound, count);

  mash_seek (mash, sound->gain_envelope, offset, SEEK_SET);
  mash_seek (mash, (MashBase *)sound->_gain_envstream, 0, SEEK_SET);
  n = mash_process (mash, sound->gain_envelope, count,
		  MashNone, sound->_gain_envstream);
  mash_dprintf ("Got %ld values from gain_envelope %p onto stream %p\n",
	  n, sound->gain_envelope, sound->_gain_envstream);

  mash_seek (mash, (MashBase *)data, data_offset, SEEK_SET);
  mash_seek (mash, (MashBase *)sound->_gain_envstream, 0, SEEK_SET);
  n = mash_stream_mult (mash, sound->_gain_envstream, data, n);
  mash_dprintf ("Multiplied %ld values of gain\n", n);

  return n;
}

static MashCount
_mash_sound_blend (MASH * mash, MashSound * sound, MashCount count,
		   MashStream * input, MashStream * output)
{
  MashCount n;

  /* XXX: this wanted to use 'block' not 'count': have we bounds
   * checked the sound ?? */

  mash_seek (mash, (MashBase * )sound->_blend_envstream, 0, SEEK_SET);
  n = mash_process (mash, sound->blend_envelope, count, MashNone,
		    sound->_blend_envstream);
  mash_seek (mash, (MashBase *)sound->_blend_envstream, 0, SEEK_SET);
  n = mash_stream_blend (mash, input, sound->_blend_envstream, output,
			 count);

  return n;
}

static MashCount
mash_sound_process (MASH * mash, MashBase * base, MashCount count, MashStream * input,
                  MashStream * output)
{
  MashSound * sound = (MashSound *)base;
  MashCount remaining = count, processed = 0, block, m, n;
  MashCount offset, input_offset, output_offset;
  MashCount mixlength = _mash_base_get_mixlength (mash, sound);

  offset = mash_tell (mash, (MashBase *)sound);

  mash_dprintf ("PROCESS SOUND (%p [%p], +%ld, %p -> %p) @ %ld\n",
	  sound, sound ? sound->source : NULL, count, input, output, offset);

  /* if we're beyond the source's actual length, then
   * just fade the input to the output using the sound's blend
   * envelope */
  if (offset > mash_length (mash, sound->source)) {
    mash_dprintf ("## offset %ld > length %ld\n", offset,
	    mash_length (mash, sound->source));

    /* Fade input and copy to output */
    if (sound->blend_envelope == MashNone) {
      mash_set_error (mash, MASH_ERROR_NOOP);
      return -1;
    } else {
      return _mash_sound_fade (mash, sound, count, input, output);
    }
  }

  while (remaining > 0) {
    block = MIN (remaining, mixlength);

    input_offset = mash_tell (mash, (MashBase *)input);
    output_offset = mash_tell (mash, (MashBase *)output);

    /* Get raw sound data; handle rate conversion etc. */
    m = _mash_sound_get_raw (mash, sound, offset, block, input, output);
    if (m == -1) {
      mash_dprintf ("error getting raw sound data\n");
      break;
    } else {
      n = m;
    }

    /* Apply gain envelope */
    if (sound->gain_envelope != MashNone) {
      m = _mash_sound_apply_gain (mash, sound, offset, n, output, output_offset);
      if (m == -1) {
	mash_dprintf ("error applying gain!\n");
      } else {
	n = m;
      }
    }
    
    /* Blend input back in */
    if (sound->blend_envelope != MashNone) {
      mash_seek (mash, (MashBase *)input, input_offset, SEEK_SET);
      mash_seek (mash, (MashBase *)output, output_offset, SEEK_SET);
      n = _mash_sound_blend (mash, sound, n, input, output);
    }
    
    offset += n;
    processed += n;
    remaining -= n;
  }

  mash_dprintf ("[mash_sound_process] processed %ld from sound %p\n",
	      processed, sound);

  return processed;
}

static MashCount
mash_sound_length (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;
  MashTime duration = mash_sound_get_duration (mash, sound);
  MashTime t = mash_time_convert (mash, duration, sound->layer->timetype,
			      MASH_TIME_SAMPLES);
  return t.samples;
}

static MashCount
mash_sound_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashSound * sound = (MashSound *)base;
  if (sound->cutlength > 0 && offset > sound->cutlength) {
    offset = sound->cutlength;
  }
  mash_seek (mash, sound->source, sound->cutin + offset, SEEK_SET);
  return offset;
}

static int
mash_sound_flush (MASH * mash, MashBase * base)
{
  MashSound * sound = (MashSound *)base;
  return mash_flush (mash, sound->source);
}

static struct _MashMethods _mash_sound_methods = {
  mash_sound_clone_invalid, /* clone */
  mash_sound_destroy,       /* destroy */
  mash_sound_ready,         /* ready */
  mash_sound_prepare,       /* prepare */
  mash_sound_process,       /* process */
  mash_sound_length,        /* length */
  mash_sound_seek,          /* seek */
  mash_sound_flush,         /* flush */
};

static MashSound *
mash_sound_optimise (MASH * mash, MashSound * sound)
{
  _mash_set_methods (mash, sound, &_mash_sound_methods);
  return sound;
}

