/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashSquareTone: a square wave tone generator
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: none */
static MashSquareTone * mash_squaretone_optimise (MASH * mash,
					      MashSquareTone * squaretone);

static MashSquareTone *
mash_squaretone_replace_channels (MASH * mash, MashSquareTone * squaretone)
{
  MashSquareToneChannel * sqch;
  CDSet * s, * channels = mash_get_channels (mash);
  MashCount offset = mash_tell (mash, (MashBase *)squaretone);

  cd_set_free_all (mash, squaretone->channels);

  for (s = channels; s; s = s->next) {
    mash_dprintf ("[mash_squaretone_replace_channels] %p replacing channel %d\n",
		squaretone, s->key);
    sqch = (MashSquareToneChannel *)
      mash_malloc (sizeof (struct _MashSquareToneChannel));
    sqch->_cycle_offset = 0;
    squaretone->channels =
      cd_set_insert (mash, squaretone->channels, s->key, CD_POINTER(sqch));
  }

  if (offset > 0) mash_seek (mash, (MashBase *)squaretone, offset, SEEK_SET);

  return squaretone;
}

static MashBase *
mash_squaretone_init (MASH * mash, MashBase * base)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  squaretone->channels = NULL;
  mash_squaretone_replace_channels (mash, squaretone);
  mash_squaretone_optimise (mash, squaretone);
  return (MashBase *)squaretone;
}

MashSquareTone *
mash_squaretone_new (MASH * mash, float frequency)
{
  MashSquareTone * squaretone =
    (MashSquareTone *)mash_base_new_subclass (mash, sizeof (struct _MashSquareTone));
  mash_squaretone_init (mash, (MashBase *)squaretone);
  squaretone->frequency = frequency;
  return squaretone;
}

static MashBase *
mash_squaretone_clone (MASH * mash, MashBase * base)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  MashSquareTone * new_squaretone =
    mash_squaretone_new (mash, squaretone->frequency);
  mash_squaretone_optimise (mash, new_squaretone);
  return (MashBase *)new_squaretone;
}

static int
mash_squaretone_destroy (MASH * mash, MashBase * base)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  mash_free (squaretone);
  return 0;
}

static int
mash_squaretone_ready (MASH * mash, MashBase * base)
{
  return (mash_base_has_samplerate (mash, base) &&
	  mash_base_encompasses_channels (mash, base));
}

static MashBase *
mash_squaretone_prepare (MASH * mash, MashBase * base)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  mash_squaretone_replace_channels (mash, squaretone);
  return base;
}

float
mash_squaretone_set_frequency (MASH * mash, MashSquareTone * squaretone,
			     float frequency)
{
  float old = squaretone->frequency;
  squaretone->frequency = frequency;
  return old;
}

float
mash_squaretone_get_frequency (MASH * mash, MashSquareTone * squaretone)
{
  return squaretone->frequency;
}

/* An MashChunkFunc for creating squaretone data */
static MashCount
mash_squaretone_write_chunk (MASH * mash, MashChunk * chunk, MashCount offset,
			   MashCount count, int channelname, void * data)
{
  MashSquareTone * squaretone = (MashSquareTone *)data;
  MashSquareToneChannel * sqch;
  MashCount remaining = count, written = 0, n;
  MashCount wavelength;
  MashPCM * d, value;
  CDScalar k;

  mash_dprintf ("[mash_squaretone_write_chunk] (+%ld) @ %ld\n", count, offset);

  k = cd_set_find (mash, squaretone->channels, channelname);
  sqch = k.s_pointer;

  if (sqch == MashNone) {
    mash_dprintf ("[mash_squaretone_write_chunk] channel %d not found\n",
		channelname);
    mash_set_error (mash, MASH_ERROR_SILENCE);
    return -1;
  }

  wavelength = (MashCount)(mash_get_samplerate (mash) / squaretone->frequency);

  mash_dprintf ("[mash_squaretone_write_chunk] wavelength %ld, cycle_offset %ld\n",
	      wavelength, sqch->_cycle_offset);
  
  if (sqch->_cycle_offset < wavelength/2) {
    n = MIN (remaining, wavelength/2 - sqch->_cycle_offset);
    value = 1.0;
  } else {
    n = MIN (remaining, wavelength - sqch->_cycle_offset);
    value = -1.0;
  }

  d = &chunk->data[offset];
  _mash_pcm_set (d, value, n);
  remaining -= n; written += n; offset += n;

  while (remaining > 0) {
    n = MIN (remaining, wavelength/2);
    value = -(value);
    d = &chunk->data[offset];
    _mash_pcm_set (d, value, n);
    remaining -= n; written += n; offset += n;
  }

  sqch->_cycle_offset += written;
  sqch->_cycle_offset %= wavelength;

  mash_dprintf ("[mash_squaretone_write_chunk] written %ld\n", written);

  return written;
}

static MashCount
mash_squaretone_process (MASH * mash, MashBase * base, MashCount count,
		       MashStream * input, MashStream * output)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  return mash_stream_chunkfuncify (mash, output, count,
				 mash_squaretone_write_chunk, squaretone);
}

static MashCount
mash_squaretone_length (MASH * mash, MashBase * base)
{
  return MASH_COUNT_INFINITE;
}

static MashCount
mash_squaretone_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashSquareTone * squaretone = (MashSquareTone *)base;
  MashCount wavelength;
  CDSet * s;
  MashCount cycle_offset;
  MashSquareToneChannel * sqch;

  wavelength = (MashCount)(mash_get_samplerate (mash) / squaretone->frequency);
  cycle_offset = offset % wavelength;

  for (s = squaretone->channels; s; s = s->next) {
    sqch = (MashSquareToneChannel *)s->data.s_pointer;
    sqch->_cycle_offset = cycle_offset;
  }

  return offset;
}

static struct _MashMethods _mash_squaretone_methods = {
  mash_squaretone_clone,
  mash_squaretone_destroy,
  mash_squaretone_ready,
  mash_squaretone_prepare,
  mash_squaretone_process,
  mash_squaretone_length,
  mash_squaretone_seek,
  NULL, /* flush */
};

static MashSquareTone *
mash_squaretone_optimise (MASH * mash, MashSquareTone * squaretone)
{
  _mash_set_methods (mash, squaretone, &_mash_squaretone_methods);
  return squaretone;
}
