/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashStream: An indexed, sparse, polyphonic PCM data container.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A stream consists of multiple channels of PCM data.
 *
 * Invariants
 * ----------
 *
 * A stream is an independent entity.
 */

#define __MASH__
#include "mash.h"

/* MashStream */

/* Optimisation dependencies: none */
static MashStream * mash_stream_optimise (MASH * mash, MashStream * stream);

MashBase *
mash_stream_init (MASH * mash, MashBase * base)
{
  MashStream * stream = (MashStream *)base;
  CDSet * s, * channels = mash_get_channels (mash);

  stream->channels = cd_set_new (mash);

  for (s = channels; s; s = s->next) {
    mash_stream_add_channel (mash, stream, s->key);
  }

  mash_stream_optimise (mash, stream);
  return (MashBase *)stream;
}

static MashStream *
mash_stream_add_channel_unchecked (MASH * mash, MashStream * stream, int name,
				 MashChannel * channel)
{
  stream->channels = cd_set_insert (mash, stream->channels, name,
				    CD_POINTER(channel));
  return stream;
}

MashChannel *
mash_stream_add_channel (MASH * mash, MashStream * stream, int name)
{
  MashChannel * channel = mash_stream_find_channel (mash, stream, name);

  if (channel == MashNone) {
    channel = mash_channel_new (mash);
    mash_stream_add_channel_unchecked (mash, stream, name, channel);
  }

  return channel;
}

MashStream *
mash_stream_new (MASH * mash)
{
  MashStream * stream =
    (MashStream *)mash_base_new_subclass (mash, sizeof (struct _MashStream));

  mash_stream_init (mash, (MashBase *)stream);

  return stream;
}

MashStream *
mash_stream_new_contiguous (MASH * mash, MashCount length)
{
  MashStream * stream = mash_stream_new (mash);
  mash_stream_add_chunks (mash, stream, 0, length);
  return stream;
}

static MashBase *
mash_stream_clone (MASH * mash, MashBase * base)
{
  MashStream * stream = (MashStream *)base;
  MashStream * new_stream;
  CDSet * s;
  MashChannel * channel, * new_channel;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  new_stream = (MashStream *)mash_stream_new (mash);

  /* clone channels */
  for (s = stream->channels; s; s = s->next) {
    channel = (MashChannel *)s->data.s_pointer;
    new_channel = mash_channel_clone (mash, channel);
    mash_stream_add_channel_unchecked (mash, new_stream, s->key, new_channel);
  }
  return (MashBase *)new_stream;
}

static int
mash_stream_destroy (MASH * mash, MashBase * base)
{
  MashStream * stream = (MashStream *)base;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  cd_set_destroy_with (mash, stream->channels, (CDDestroyFunc)mash_destroy);

  mash_free (stream);
  return 0;
}

MashCount
mash_stream_nr_channels (MASH * mash, MashStream * stream)
{
  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  return (MashCount)cd_set_size (mash, stream->channels);
}

MashChannel *
mash_stream_find_channel (MASH * mash, MashStream * stream, int name)
{
  CDScalar k;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  k = cd_set_find (mash, stream->channels, name);
  return (MashChannel *)k.s_pointer;
}


MashStream *
mash_stream_remove_channel (MASH * mash, MashStream * stream, int name)
{
  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }
   
  stream->channels = cd_set_remove (mash, stream->channels, name);

  return stream;
}

MashStream *
mash_stream_add_chunks (MASH * mash, MashStream * stream, MashCount offset,
		      MashCount length)
{
  CDSet * s;
  MashChannel * channel;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return MashNone;
  }

  for (s = stream->channels; s; s = s->next) {
    channel = (MashChannel *)s->data.s_pointer;
    mash_channel_add_new_chunk (mash, channel, offset, length);
  }

  return stream;
}

MashCount
mash_stream_write0 (MASH * mash, MashStream * stream, MashCount count)
{
  CDSet * s;
  MashChannel * channel;
  MashCount offset;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  offset = mash_tell (mash, (MashBase *)stream);

  for (s = stream->channels; s; s = s->next) {
    channel = (MashChannel *)s->data.s_pointer;
    mash_channel_write0 (mash, channel, count);
  }
  
  mash_seek (mash, (MashBase *)stream, offset + count, SEEK_SET);

  mash_dprintf ("[mash_stream_write0] (%p) written %ld\n", stream, count);

  return count;
}

/*
 * mash_stream_process (mash, stream, count, input, output)
 *
 * MashProcessFunc for MashStream.
 * Ignores input. Copies data from 'stream' to 'output'.
 */
static MashCount
mash_stream_process (MASH * mash, MashBase * base, MashCount count,
		   MashStream * input, MashStream * output)
{
  MashStream * stream = (MashStream *)base;
  return mash_stream_write (mash, output, count, stream);
}

static MashCount
mash_stream_length (MASH * mash, MashBase * base)
{
  MashStream * stream = (MashStream *)base;
  MashCount length, maxlength = 0;
  CDSet * s;
  MashChannel * channel;

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  for (s = stream->channels; s; s = s->next) {
    channel = (MashChannel *)s->data.s_pointer;
    length = _mash_channel_length (mash, channel);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

static MashCount
mash_stream_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashStream * stream = (MashStream *)base;
  CDSet * s;
  MashChannel * channel;

  for (s = stream->channels; s; s = s->next) {
    channel = (MashChannel *)s->data.s_pointer;
    _mash_channel_seek (mash, channel, offset);
  }

  return offset;
}

static struct _MashMethods _mash_stream_methods = {
  mash_stream_clone,
  mash_stream_destroy,
  NULL, /* ready */
  NULL, /* prepare */
  mash_stream_process,
  mash_stream_length,
  mash_stream_seek,
};

static MashStream *
mash_stream_optimise (MASH * mash, MashStream * stream)
{
  _mash_set_methods (mash, stream, &_mash_stream_methods);
  return stream;
}


/*
 * mash_stream_chunkfuncify (stream, count, func, data)
 *
 */
MashCount
mash_stream_chunkfuncify (MASH * mash, MashStream * stream, MashCount count,
			MashChunkFunc func, void * data)
{
  MashCount n, minn = count, offset = mash_tell (mash, (MashBase *)stream);
  CDSet * s, * channels = mash_get_channels (mash);
  MashChannel * channel;

  for (s = channels; s; s = s->next) {
    mash_dprintf ("[mash_stream_chunkfuncify] thinking of channel %d\n", s->key);
  }

  if (stream == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  mash_dprintf ("[mash_stream_chunkfuncify] (%p, +%ld) @ %ld\n",
	      stream, count, offset);

  for (s = stream->channels; s; s = s->next) {
    if (cd_set_contains (mash, channels, s->key)) {
      channel = (MashChannel *)s->data.s_pointer;
      n = mash_channel_chunkfuncify (mash, channel, minn, func, s->key, data);
      minn = MIN (n, minn);
    } else {
      mash_dprintf ("[mash_stream_chunkfuncify] channel %d not funced\n", s->key);
    }
  }

  mash_seek (mash, (MashBase *)stream, offset + minn, SEEK_SET);

  return minn;
}

MashCount
mash_stream_chunkchunkfuncify (MASH * mash, MashStream * src, MashStream * dest,
			     MashCount count, MashChunkChunkFunc func,
			     void * data)
{
  MashCount n, minn = count;
  MashCount src_offset = mash_tell (mash, (MashBase *)src);
  MashCount dest_offset = mash_tell (mash, (MashBase *)dest);
  CDSet * s, * channels = mash_get_channels (mash);
  MashChannel * sch, * dch;

  if (dest == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  mash_dprintf ("[mash_stream_ccf...] (%p -> %p, +%ld), src @ %ld, dest @ %ld\n",
	      src, dest, count, src_offset, dest_offset);

  for (s = dest->channels; s; s = s->next) {
    if (cd_set_contains (mash, channels, s->key)) {
      dch = (MashChannel *)s->data.s_pointer;
      sch = mash_stream_find_channel (mash, src, s->key);
      if (sch != MashNone) {
	n = mash_channel_chunkchunkfuncify (mash, sch, dch, count, func, s->key,
					  data);
	if (n == -1) {
	  return -1;
	}

	minn = MIN (n, minn);
      }
    }
  }

  mash_seek (mash, (MashBase *)src, src_offset + minn, SEEK_SET);
  mash_seek (mash, (MashBase *)dest, dest_offset + minn, SEEK_SET);

  return minn;
}

MashCount
mash_stream_chunkchunkchunkfuncify (MASH * mash,
				  MashStream * src1, MashStream * src2,
				  MashStream * dest, MashCount count,
				  MashChunkChunkChunkFunc func, void * data)
{
  MashCount n, minn = count;
  MashCount src1_offset = mash_tell (mash, (MashBase *)src1);
  MashCount src2_offset = mash_tell (mash, (MashBase *)src2);
  MashCount dest_offset = mash_tell (mash, (MashBase *)dest);
  CDSet * s, * channels = mash_get_channels (mash);
  MashChannel * s1ch, * s2ch, * dch;

  if (dest == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  for (s = dest->channels; s; s = s->next) {
    if (cd_set_contains (mash, channels, s->key)) {
      dch = (MashChannel *)s->data.s_pointer;
      s1ch = mash_stream_find_channel (mash, src1, s->key);
      s2ch = mash_stream_find_channel (mash, src2, s->key);
      if (s1ch != MashNone && s2ch != MashNone) {
	n = mash_channel_chunkchunkchunkfuncify (mash, s1ch, s2ch, dch,
					       count, func, s->key, data);
	minn = MIN (n, minn);
      }
    }
  }

  mash_seek (mash, (MashBase *)src1, src1_offset + minn, SEEK_SET);
  mash_seek (mash, (MashBase *)src2, src2_offset + minn, SEEK_SET);
  mash_seek (mash, (MashBase *)dest, dest_offset + minn, SEEK_SET);

  return minn;
}

/*
 * mash_stream_gain (mash, src, dest, count, gain)
 */
MashCount
mash_stream_gain (MASH * mash, MashStream * stream, MashCount count, MashPCM gain)
{
  return mash_stream_chunkfuncify (mash, stream, count, _mash_chunk_gain, &gain);
}

/*
 * mash_stream_copy (mash, src, dest, count)
 *
 * Copy 'count' samples from 'src' to 'dest'
 */
MashCount
mash_stream_copy (MASH * mash, MashStream * src, MashStream * dest, MashCount count)
{
  mash_dprintf ("[mash_stream_copy] (%p -> %p, +%ld)\n", src, dest, count);

  return mash_stream_chunkchunkfuncify (mash, src, dest, count,
				      _mash_chunk_copy, NULL);
}

/*
 * mash_stream_write (mash, stream, count, data)
 *
 * Write 'count' samples of 'data' to the stream 'stream'.
 */
/* XXX: this is a wrapper to obsolete mash_stream_write above */
MashCount
mash_stream_write (MASH * mash, MashStream * stream, MashCount count,
		 MashStream * data)
{
  mash_dprintf ("[mash_stream_write] (%p -> %p, +%ld)\n", data, stream, count);

  if (data == MashNone)
    return mash_stream_write0 (mash, stream, count);

  return mash_stream_copy (mash, data, stream, count);
}

/*
 * mash_stream_mix (src, dest, count)
 *
 * Mix 'count' samples from 'src' into 'dest'.
 */
MashCount
mash_stream_mix (MASH * mash, MashStream * src, MashStream * dest, MashCount count)
{
  mash_dprintf ("[mash_stream_mix] (%p -> %p, +%ld)\n", src, dest, count);

  return mash_stream_chunkchunkfuncify (mash, src, dest, count,
				      _mash_chunk_add_inplace, NULL);
}

/*
 * mash_stream_mult (src, dest, count)
 *
 * Multiply 'count' samples of 'src' into 'dest'.
 */
MashCount
mash_stream_mult (MASH * mash, MashStream * src, MashStream * dest, MashCount count)
{
  mash_dprintf ("[mash_stream_mult] (%p -> %p, +%ld)\n", src, dest, count);

  return mash_stream_chunkchunkfuncify (mash, src, dest, count,
				      _mash_chunk_mult_inplace, NULL);
}

/*
 * mash_stream_fade (src, dest, count)
 *
 * Fade 'count' samples of 'dest' by values in 'src'.
 */
MashCount
mash_stream_fade (MASH * mash, MashStream * src, MashStream * dest,
		  MashCount count)
{
  mash_dprintf ("[mash_stream_fade] (%p -> %p, +%ld)\n", src, dest, count);

  return mash_stream_chunkchunkfuncify (mash, src, dest, count,
					_mash_chunk_fade_inplace, NULL);
}

/*
 * mash_stream_blend (mash, src, blend, dest, count)
 *
 * Blend 'count' samples of 'src' into 'dest' by amounts in 'blend'.
 */
MashCount
mash_stream_blend (MASH * mash, MashStream * src, MashStream * blend,
		   MashStream * dest, MashCount count)
{
  mash_dprintf ("[mash_stream_blend] (%p -> (%p) -> %p, +%ld)\n", src, blend,
	      dest, count);

  return mash_stream_chunkchunkchunkfuncify (mash, src, blend, dest, count,
					     _mash_chunk_blend_inplace, NULL);
}


/*
 * mash_streams_mix (streams, dest, count)
 *
 * Mix 'count' samples from all streams in list 'streams' together into
 * 'dest'.
 */
MashCount
mash_streams_mix (MASH * mash, CDList * streams, MashStream * dest,
		  MashCount count)
{
  CDList * sl;
  CDSet * s;
  MashChannel * sch, * dch;
  MashStream * stream;
  MashCount dest_start = mash_tell (mash, (MashBase *)dest);
  MashCount stream_start;

  if (dest == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOENTITY);
    return -1;
  }

  for (s = dest->channels; s; s = s->next) {
    dch = (MashChannel *)s->data.s_pointer;
    for (sl = streams; sl; sl = sl->next) {
      stream = (MashStream *)sl->data.s_pointer;
      stream_start = mash_tell (mash, (MashBase *)stream);
      sch = mash_stream_find_channel (mash, stream, s->key);
      if (sch != MashNone) {
        _mash_channel_seek (mash, dch, dest_start);
        mash_channel_mix (mash, sch, dch, count);
        mash_seek (mash, (MashBase *)stream, stream_start, SEEK_SET);
      }
    }
  }

  for (sl = streams; sl; sl = sl->next) {
    stream = (MashStream *)sl->data.s_pointer;
    stream_start = mash_tell (mash, (MashBase *)stream);
    mash_seek (mash, (MashBase *)stream, stream_start + count, SEEK_SET);
  }

  mash_seek (mash, (MashBase *)dest, dest_start+count, SEEK_SET);

  return count;
}

/*
 * mash_stream_interleave_2 (mash, stream, name1, name2, dest, count)
 *
 * Interleave 'count' frames of the channels named 'name1' and 'name2' in
 * 'stream', placing the resulting PCM data in the memory region pointed
 * to by 'dest'.
 */
MashCount
mash_stream_interleave_2 (MASH * mash, MashStream * stream,
			  int name1, int name2,
			  MashPCM * dest, MashCount count)
{
  MashChannel * channel1, * channel2;
  MashCount n = 0;

  channel1 = mash_stream_find_channel (mash, stream, name1);
  channel2 = mash_stream_find_channel (mash, stream, name2);

  if (channel1 != MashNone && channel2 != MashNone) {
    n = mash_channel_chunkchunkfuncify (mash, channel1, channel2,
					count, _mash_chunk_interleave_2,
					0, dest);
  }

  if (n > 0) mash_seek (mash, (MashBase *)stream, n, SEEK_CUR);

  return n;
}

/*
 * mash_stream_deinterleave_stereo (mash, stream, name1, name2, src, count)
 *
 * Deinterleave 'count' frames from the memory region pointed to by 'src',
 * placing the resulting channels into those named 'name1' and 'name2' in
 * 'stream'.
 */
MashCount
mash_stream_deinterleave_2 (MASH * mash, MashStream * stream,
			    int name1, int name2,
			    MashPCM * src, MashCount count)
{
  MashChannel * channel1, * channel2;
  MashCount n = 0;

  channel1 = mash_stream_find_channel (mash, stream, name1);
  channel2 = mash_stream_find_channel (mash, stream, name2);

  if (channel1 != MashNone && channel2 != MashNone) {
    n = mash_channel_chunkchunkfuncify (mash, channel1, channel2,
					count, _mash_chunk_deinterleave_2,
					0, src);
  }

  if (n > 0) mash_seek (mash, (MashBase *)stream, n, SEEK_CUR);

  return n;
}
