/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashTime: A generic time abstraction for sequencing information.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 */

#define __MASH__
#include "mash.h"

MashTime
mash_time_zero (MashTimeType type)
{
  return _mash_time_zero (type);
}

MashTime
mash_time_invalid (MashTimeType type)
{
  return _mash_time_invalid (type);
}

int
mash_time_is_invalid (MashTimeType type, MashTime time)
{
  return _mash_time_is_invalid (type, time);
}

MashTime
mash_time_add (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_add (type, t1, t2);
}

MashTime
mash_time_sub (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_sub (type, t1, t2);
}

MashTime
mash_time_min (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_min (type, t1, t2);
}

MashTime
mash_time_max (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_max (type, t1, t2);
}

int
mash_time_eq (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_eq (type, t1, t2);
}

int
mash_time_gt (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_gt (type, t1, t2);
}

int
mash_time_lt (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_lt (type, t1, t2);
}

int
mash_time_ge (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_ge (type, t1, t2);
}

int
mash_time_le (MashTimeType type, MashTime t1, MashTime t2)
{
  return _mash_time_le (type, t1, t2);
}


static float
mash_samples_to_seconds (MashCount samples, MashSamplerate samplerate)
{
  return ((float)samples / samplerate);
}

static int
mash_samples_to_beat24s (MashCount samples, MashSamplerate samplerate,
		       MashTempo tempo)
{
  return ((int)((float)samples * tempo * 24.0 / (samplerate * 60.0)));
}

static MashCount
mash_seconds_to_samples (float seconds, MashSamplerate samplerate)
{
  return (MashCount)(seconds * samplerate);
}

static int
mash_seconds_to_beat24s (float seconds, MashTempo tempo)
{
  return (int)(seconds * tempo * 24.0 / 60.0);
}

static MashCount
mash_beat24s_to_samples (int beat24s, MashSamplerate samplerate, MashTempo tempo)
{
  return (MashCount)(beat24s * samplerate * 60.0 / (tempo * 24.0));
}

static float
mash_beat24s_to_seconds (int beat24s, MashTempo tempo)
{
  return ((float)beat24s * 60.0 / (tempo * 24.0));
}


MashTime
mash_time_convert (MASH * mash, MashTime time, MashTimeType old_type,
		   MashTimeType new_type)
{
  MashSamplerate samplerate;
  MashTempo tempo;

  if (old_type == new_type) return time;

  samplerate = mash_get_samplerate (mash);
  tempo = mash_get_tempo (mash);

  switch (old_type) {
  case MASH_TIME_SAMPLES:
    switch (new_type) {
    case MASH_TIME_SECONDS:
      return (MashTime) mash_samples_to_seconds (time.samples, samplerate);
      break;
    case MASH_TIME_BEAT24S:
      return (MashTime) mash_samples_to_beat24s (time.samples, samplerate, tempo);
      break;
    default: break;
    }
    break;
  case MASH_TIME_SECONDS:
    switch (new_type) {
    case MASH_TIME_SAMPLES:
      return (MashTime) mash_seconds_to_samples (time.seconds, samplerate);
      break;
    case MASH_TIME_BEAT24S:
      return (MashTime) mash_seconds_to_beat24s (time.seconds, tempo);
      break;
    default: break;
    }
    break;
  case MASH_TIME_BEAT24S:
    switch (new_type) {
    case MASH_TIME_SAMPLES:
      return (MashTime) mash_beat24s_to_samples (time.beat24s, samplerate, tempo);
      break;
    case MASH_TIME_SECONDS:
      return (MashTime) mash_beat24s_to_seconds (time.beat24s, tempo);
      break;
    default: break;
    }
    break;
  default:
    /* Cannot convert invalid time type to anything */
    break;
  }

  return mash_time_invalid (new_type);
}
