/*
 * MASH -- An audio mixing and sequencing library.
 *
 * Copyright (C) 2001 Commonwealth Scientific and Industrial Research
 * Organisation (CSIRO), Australia.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * MashTrack: An MashLayer mixing abstraction.
 *
 * Conrad Parker <Conrad.Parker@CSIRO.AU>, August 2001
 *
 * Description
 * -----------
 *
 * A track is contained within a deck. A track contains a number of
 * layers which are mixed in series.
 *
 * Invariants
 * ----------
 *
 * A track must be contained within a deck.
 */

#define __MASH__
#include "mash.h"

/* Optimisation dependencies: optimise on change of nr. layers */
static MashTrack * mash_track_optimise (MASH * mash, MashTrack * track);

static void
mash_track_replace_mixstreams (MASH * mash, MashTrack * track)
{
  MashCount mixlength = _mash_base_get_mixlength (mash, track);

  if (track->_mixstream_a != MashNone)
    mash_destroy (mash, (MashBase *)track->_mixstream_a);
  if (track->_mixstream_b != MashNone)
    mash_destroy (mash, (MashBase *)track->_mixstream_b);

  track->_mixstream_a = mash_stream_new_contiguous (mash, mixlength);
  track->_mixstream_b = mash_stream_new_contiguous (mash, mixlength);
}

static MashBase *
mash_track_init (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  track->gain = 1.0;
  track->layers = cd_list_new (mash);
  track->_mixstream_a = track->_mixstream_b = MashNone;
  mash_track_replace_mixstreams (mash, track);
  mash_track_optimise (mash, track);
  return (MashBase *)track;
}

static MashTrack *
mash_track_new (MASH * mash)
{
  return (MashTrack *) mash_base_new_subclass (mash, sizeof (struct _MashTrack));
}

MashBase *
mash_track_clone (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  MashTrack * new_track = mash_track_new (mash);

  new_track->gain = track->gain;
  new_track->layers = cd_list_clone (mash, track->layers,
				     (CDCloneFunc)mash_layer_clone);
  mash_track_optimise (mash, track);

  return (MashBase *)new_track;
}

static int
mash_track_destroy (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  mash_destroy_list (mash, track->layers);
  mash_free (track);
  return 0;
}

static int
mash_track_ready (MASH * mash, MashBase * base)
{
  return (mash_base_encompasses_mixlength (mash, base) &&
	  mash_base_encompasses_channels (mash, base));
}

static MashBase *
mash_track_prepare (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  mash_track_replace_mixstreams (mash, track);
  return base;
}

MashTrack *
mash_add_track (MASH * mash, MashDeck * deck)
{
  MashTrack * track;

  track = mash_track_new (mash);
  track->deck = deck;
  mash_track_init (mash, (MashBase *)track);
  _mash_deck_add_track (mash, deck, track);

  return track;
}

static MashCount
mash_track_length (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  MashCount length, maxlength = 0;
  CDList * l;
  MashLayer * layer;

  for (l = track->layers; l; l = l->next) {
    layer = (MashLayer *)l->data.s_pointer;
    length = mash_length (mash, (MashBase *)layer);
    mash_dprintf ("[mash_track_length] found layer %p length %ld\n",
		layer, length);
    maxlength = MAX (maxlength, length);
  }

  return maxlength;
}

MashPCM
mash_track_set_gain (MASH * mash, MashTrack * track, MashPCM gain)
{
  MashPCM old = track->gain;
  track->gain = gain;
  return old;
}

MashPCM
mash_track_get_gain (MASH * mash, MashTrack * track)
{
  return track->gain;
}

void
mash_remove_track (MASH * mash, MashTrack * track)
{
  MashDeck * deck = track->deck;

  if (deck)
    _mash_deck_remove_track (mash, deck, track);
}

MashDeck *
mash_track_get_deck (MASH * mash, MashTrack * track)
{
  return track->deck;
}

/*
 * _mash_track_add_layer_above (mash, track, layer, above)
 *
 * Adds 'layer' above 'above'. If above is MashNone, adds 'layer' on top
 * of 'track'.
 */
MashLayer *
_mash_track_add_layer_above (MASH * mash, MashTrack * track, MashLayer * layer,
			   MashLayer * above)
{
  layer->track = track;
  if (above == MashNone)
    above = (MashLayer *)
      (cd_list_last (mash, track->layers, CD_TYPE_POINTER)).s_pointer;
  track->layers = cd_list_add_after (mash, track->layers, CD_TYPE_POINTER,
				     CD_POINTER(layer), CD_POINTER(above));
  mash_track_optimise (mash, track);
  return layer;
}

MashLayer *
_mash_track_remove_layer (MASH * mash, MashTrack * track, MashLayer * layer)
{
  track->layers = cd_list_remove (mash, track->layers, CD_TYPE_POINTER,
				  CD_POINTER(layer));
  mash_track_optimise (mash, track);
  return layer;
}

/*
 * gets layer above 'above'. If above is MashNone, returns topmost layer
 */
MashLayer *
_mash_track_get_layer_above (MASH * mash, MashTrack * track, MashLayer * above)
{
  CDList * above_item;
  MashLayer * layer;

  if (track->layers == NULL) return MashNone;

  if (above == MashNone) {
    layer = (MashLayer *)
      (cd_list_last (mash, track->layers, CD_TYPE_POINTER)).s_pointer;
  } else {
    above_item = cd_list_find (mash, track->layers, CD_TYPE_POINTER,
			       CD_POINTER(above));
    if (above_item == NULL) return MashNone;
    above_item = above_item->next;
    if (above_item == NULL) return MashNone;
    layer = (MashLayer *) above_item->data.s_pointer;
  }

  return layer;
}

/*
 * gets layer below 'below'. If below is MashNone, returns lowest layer
 */
MashLayer *
_mash_track_get_layer_below (MASH * mash, MashTrack * track, MashLayer * below)
{
  CDList * below_item;
  MashLayer * layer;

  if (track->layers == NULL) return MashNone;

  if (below == MashNone) {
    layer = (MashLayer *) track->layers->data.s_pointer;
  } else {
    below_item = cd_list_find (mash, track->layers, CD_TYPE_POINTER,
			       CD_POINTER(below));
    if (below_item == NULL) return MashNone;
    below_item = below_item->prev;
    if (below_item == NULL) return MashNone;
    layer = (MashLayer *) below_item->data.s_pointer;
  }

  return layer;
}


static MashCount
mash_track_process (MASH * mash, MashBase * base, MashCount count, MashStream * input,
                  MashStream * output)
{
  MashTrack * track = (MashTrack *)base;
  CDList * l;
  MashBase * layer;
  MashStream * si, * so, * swap_stream;
  MashCount remaining = count, processed = 0, n = 0;
  MashCount mixlength = _mash_base_get_mixlength (mash, track);

  mash_dprintf ("PROCESS TRACK (%p, +%ld, %p -> %p) @ %ld\n",
		track, count, input, output, mash_tell (mash, base));

  if (track->layers == MashNone) {
    mash_set_error (mash, MASH_ERROR_NOOP);
    return 0;
  }

  while (remaining > 0) {
    l = track->layers;
    si = input;
    so = track->_mixstream_a;
    n = MIN (remaining, mixlength);

    while (l) {
      layer = (MashBase *)l->data.s_pointer;

      if (si == input) {
	swap_stream = track->_mixstream_b;
      } else {
	swap_stream = si;
	mash_seek (mash, (MashBase *)si, 0, SEEK_SET);
      }
      
      if (l->next == MashNone) {
	so = output;
      } else {
	mash_seek (mash, (MashBase *)so, 0, SEEK_SET);
      }
      
      n = mash_process (mash, layer, n, si, so);
      
      l = l->next;

      /* swap si, st using swap_stream as evaluated above */
      si = so; so = swap_stream;
    }
    remaining -= n;
    processed += n;
  }

  mash_dprintf ("[mash_track_process] processed %ld\n", processed);

  return processed;
}

static MashCount
mash_track_twolayer_process (MASH * mash, MashBase * base, MashCount count,
			   MashStream * input, MashStream * output)
{
  MashTrack * track = (MashTrack *)base;
  CDList * l;
  MashBase * layer1, * layer2;
  MashCount remaining = count, processed = 0, n = 0;
  MashStream * mix = track->_mixstream_a;
  MashCount current_offset = mash_tell (mash, base);
  MashCount mixlength = _mash_base_get_mixlength (mash, track);

  mash_dprintf ("PROCESS TRACK [twolayer] (%p, +%ld, %p -> %p) @ %ld\n",
	      track, count, input, output, current_offset);

  l = track->layers;
  layer1 = (MashBase *)l->data.s_pointer;
  mash_seek (mash, (MashBase *)layer1, current_offset, SEEK_SET);
  
  l = l->next;
  layer2 = (MashBase *)l->data.s_pointer;
  mash_seek (mash, (MashBase *)layer2, current_offset, SEEK_SET);

  while (remaining > 0) {
    n = MIN (remaining, mixlength);

    mash_seek (mash, (MashBase *)mix, 0, SEEK_SET);
    n = mash_process (mash, layer1, n, input, mix);

    mash_seek (mash, (MashBase *)mix, 0, SEEK_SET);
    n = mash_process (mash, layer2, n, mix, output);

    remaining -= n;
    processed += n;
  }

  mash_dprintf ("*** PRE-SEEK: track @ %ld\n", mash_tell (mash, base));

  /*mash_seek (mash, base, current_offset + processed, SEEK_SET);*/

  mash_dprintf ("*** POST-SEEK: track @ %ld\n", mash_tell (mash, base));

  mash_dprintf ("[mash_track_twolayer_process] processed %ld\n", processed);

  return processed;
}

static MashCount
mash_track_onelayer_process (MASH * mash, MashBase * base, MashCount count,
			   MashStream * input, MashStream * output)
{
  MashTrack * track = (MashTrack *)base;
  CDList * l = track->layers;
  MashBase * layer = (MashBase *)l->data.s_pointer;
  MashCount n;

  mash_dprintf ("PROCESS TRACK [onelayer] (%p, +%ld, %p -> %p) @ %ld\n",
	      track, count, input, output, mash_tell (mash, base));

  n = mash_process (mash, layer, count, input, output);

  mash_dprintf ("[mash_track_onelayer_process] processed %ld\n", n);

  return n;
}

static MashCount
mash_track_seek (MASH * mash, MashBase * base, MashCount offset)
{
  MashTrack * track = (MashTrack *)base;
  CDList * l;
  MashLayer * layer;

  for (l = track->layers; l; l = l->next) {
    layer = (MashLayer *)l->data.s_pointer;
    mash_seek (mash, (MashBase *)layer, offset, SEEK_SET);
  }

  return offset;
}

static int
mash_track_flush (MASH * mash, MashBase * base)
{
  MashTrack * track = (MashTrack *)base;
  CDList * l;
  MashLayer * layer;

  for (l = track->layers; l; l = l->next) {
    layer = (MashLayer *)l->data.s_pointer;
    mash_flush (mash, (MashBase *)layer);
  }

  return 0;
}

static struct _MashMethods _mash_track_empty_methods = {
  mash_track_clone,   /* clone */
  mash_track_destroy, /* destroy */
  mash_track_ready,   /* ready */
  mash_track_prepare, /* prepare */
  mash_null_process,  /* process */
  mash_null_length,   /* length */
  mash_null_seek,     /* seek */
  NULL,             /* flush */
};

static struct _MashMethods _mash_track_methods = {
  mash_track_clone,   /* clone */
  mash_track_destroy, /* destroy */
  mash_track_ready,   /* ready */
  mash_track_prepare, /* prepare */
  mash_track_process, /* process */
  mash_track_length,  /* length */
  mash_track_seek,    /* seek */
  mash_track_flush,            /* flush */
};

static struct _MashMethods _mash_track_onelayer_methods = {
  mash_track_clone,            /* clone */
  mash_track_destroy,          /* destroy */
  mash_track_ready,            /* ready */
  mash_track_prepare,          /* prepare */
  mash_track_onelayer_process, /* process */
  mash_track_length,           /* length */
  mash_track_seek,             /* seek */
  mash_track_flush,            /* flush */
};

static struct _MashMethods _mash_track_twolayer_methods = {
  mash_track_clone,            /* clone */
  mash_track_destroy,          /* destroy */
  mash_track_ready,            /* ready */
  mash_track_prepare,          /* prepare */
  mash_track_twolayer_process, /* process */
  mash_track_length,           /* length */
  mash_track_seek,             /* seek */
  mash_track_flush,            /* flush */
};

static MashTrack *
mash_track_optimise (MASH * mash, MashTrack * track)
{
  MashCount nr_layers = cd_list_length (mash, track->layers);

  switch (nr_layers) {
  case 0:
    _mash_set_methods (mash, track, &_mash_track_empty_methods); break;
  case 1:
    _mash_set_methods (mash, track, &_mash_track_onelayer_methods); break;
  case 2:
    _mash_set_methods (mash, track, &_mash_track_twolayer_methods); break;
  default:
      _mash_set_methods (mash, track, &_mash_track_methods); break;
  }

  return track;
}
